(function ($) {
    "use strict";
    //region Imports
    /**
     * @name controls
     * @namespace use the global cm.controls namespace
     */
    var controls = $.namespace("cm.controls", true);
    //endregion

    //region ColorPicker
    /**
     * Static Singleton ColorPickerController
     * really setFieldValue was supposed to be ignored by a control if it was active as change then only happens by user action
     * and all contextual data should be stored in the context not in vars in the control
     * currently code allows a non active control to change orig for the active control
     * thus triggering a commit when no real change had occurred
     * if used from a button rather than an input a default color must be set on the button
     * otherwise it might blow up in commit value if nothing was ever clicked
     * buttons don't have a ctx
     * NOTE:
     * ctl.value is always hex, origColor is also always hex and is only used to check if 'value' has changed and also to replace invalid values
     * data color and dataset.color are based on color mode (hex or rgb)
     * if color mode is not set it is determined on first value or defaults to hex.
     * @returns {ColorPickerController}
     * @constructor
     */
    var ColorPickerController = function() {
        var self = this;
        //the picker UI
        var picker;
        //the current target element
        var activeCtl;

        var selectedCell = -1; //the index of the selected cell in the picker (if one is selected)

        var colors = ['#AC725E', '#D06B64', '#F83A22', '#FA573C', '#FF7537', '#FFAD46', '#42D692', '#16A765', '#7BD148', '#B3DC6C', '#FBE983', '#FAD165', '#92E1C0', '#9FE1E7', '#9FC6E7', '#4986E7', '#9A9CFF', '#B99AFF', '#C2C2C2', '#CABDBF', '#CCA6AC', '#F691B2', '#CD74E6', '#A47AE2'];

        var template =
            '<dialog id="colorpicker" class="colorpicker popover picker"  data-position="picker">' +
                '<header class="dialog-header" tabindex="0"></header>' +
                '<section class="dialog-content">' +
                    '<table class="cp table-condensed ignore">' +
                        '<tbody>[[COLORCELLS]]</tbody>' +
                    '</table>' +
                '</section>' +
                '<footer class="dialog-footer"><button class="btn-dialog" data-action="dialog-close">Done</button></footer>' +
            '</dialog>';

        /**
         *
         * @param id
         * @returns {string}
         */
        function getTemplatePart(id) {
            var html = '';
            switch(id) {
                case 'colorcells':
                    var k = 0;
                    for(var i = 0; i < 4; i++) {
                        html += '<tr>';
                        for(var j = 0; j < 6; j++) {
                            html += '<td class="cp-cell"><div class="cp-box" data-key="' + (colors[k]) + '"' + 'style="background-color:' + (colors[k]) + '" data-index="' + k + '"></div></td>';
                            k++;
                        }
                        html += '</tr>';
                    }
                    break;
                default:
                    break;
            }
            return html;
        }

        /**
         * get the index of a color in the colors array
         * @param {string} color
         * @returns {number}
         */
        function getColorIndex(color) {
            return colors.indexOf(color)
        }
        /**
         * Initialize the controller
         */
        function init() {
            //localize template
            template = template.replace('[[COLORCELLS]]', getTemplatePart('colorcells'));
            //create global picker
            picker = DialogManager.addDialog(template);
            //track resize + orientation changes in case we need to reposition the picker
            Views.notify.add(viewsNotifyHandler);
        }

        /**
         * Display the picker
         */
        function showPicker() {
            //show the picker dropdown
            picker.show(activeCtl);
        }

        /**
         * Close the picker
         */
        function hidePicker() {
            //hide UI
            if(picker.open) picker.close();
        }

        /**
         * Toggle the display state of the picker dropdown UI
         * @param {HTMLElement} el
         * @param {Boolean=} state
         */
        function showDropDown(el, state) {
            if(state === false || (state !== true && picker.open)) {
                hidePicker();
                if(el.tagName !== 'INPUT') setActiveControl(el, false, false);
            } else {
                el.focus();
                setActiveControl(el, true, true);
            }
        }

        /**
         * Reposition the dropdown on a window resize
         * @param type
         */
        function viewsNotifyHandler(type) {
            switch(type) {
                case 'resize':
                    hidePicker();
                    break;
                case 'scroll.start':
                    var target = arguments[1];
                    //close any open popup.... but only if the scroll is outside the popup - a bit like mdo...
                    if(activeCtl && !activeCtl.contains(target) && !e.defaultPrevented) {
                        hidePicker();
                    }
                    break;
            }
        }
        /**
         * Handle picker clicks
         * @param {Event} e
         * @param view
         * @param cur
         * @param prev1
         * @param prev2
         * @param prev3
         * @returns {boolean}
         */
        function clickHandler(e, view, cur, prev1, prev2, prev3) {
            //find what was clicked
            var target = e.target;
            if(target && target.tagName === 'DIV') {
                setCtlValue(target.dataset.key);
                setIconColor(activeCtl);
                commitValue();
                hidePicker();
            }
            return false;
        }
        /**
         * Commit the value if it is valid
         * Called from blur or keydown-return
         */
        function commitValue() {
            var el = activeCtl;
            var color;
            var origColor = el.dataset.origColor;
            if(el.tagName === 'INPUT') {
                color = el.value.toUpperCase();
                if( origColor !== color) {
                    if(setCtlValue(color)) {
                        setIconColor(el);
                        //fires the notify event and marks control and form dirty
                        Forms.commitValue(el);
                    }
                }
            } else {
                //probably a popup triggered from a button, el will be the button and has no ctx
                color = valueToHex(el.dataset.color, '#000', el.dataset.colorMode);
                if(origColor !== color) {
                    if(setCtlValue(color)) {
                        setIconColor(el);
                        //fires the notify event and marks control and form dirty
                        Views.fireActionEvent(el, 'change', false, false, { action: el.dataset.changeAction || 'colorchange', key: color, color: color, index: getColorIndex(color) });
                    }
                }
            }
        }

        /**
         * Used to set the activeCtl value and ensure that it begins with #
         * Called from commit (2) move (2) and click, the input is always hex or supposed to be and output is in color mode
         * @param {string} value hex value to be validated
         * @returns {boolean} returns true if valid
         */
        function setCtlValue(value) {
            var valid = true;
            if(value) {
                value = (((value.indexOf('#') !== 0) ? '#' : '') + value).toUpperCase();
                if(!isValidColor(value)) {
                    //if not valid revert to orig and flag as invalid
                    value = valueToHex(activeCtl.dataset.origColor, '',activeCtl.dataset.colorMode);
                    valid = false;
                }
            }
            //value is now vaid hex and uppercase, so now push the value into the ctl
            if(activeCtl.tagName === 'INPUT') {
                activeCtl.value = value;
            } else {
                //non-input stores dataset value in native type
                activeCtl.dataset.color = hexToValue(value, 0, activeCtl.dataset.colorMode);
            }
            return valid;
        }

        var rxColor = /(^#[0-9A-F]{6}$)|(^#[0-9A-F]{3}$)/i;
        /**
         * Validates the Hex color value
         * @param {string} value value to be validated
         * @returns {boolean} returns true if valid
         */
        function isValidColor(value) {
            return value ? rxColor.test(value) : true;
        }
        /**
         * Handler for Color Input blur event
         * @param {Event} e
         */
        function onInputBlur(e) {
            if(activeCtl) {
                if(Focus.isFocusOut(e, activeCtl, picker)) {
                    hidePicker();
                    commitValue();
                    if(!e.force || (e.force && e.willBlur)) {
                        activeCtl.dataset.origColor = '';
                        activeCtl = null;
                        selectedCell = -1;
                    }
                }
            }
        }

        /**
         * Handler for the Color Input focus event
         * @param {Event} e
         */
        function onInputFocus(e) {
            //if getting focus from our own picker do nothing...
            setActiveControl(e.target, false, true);
        }

        /**
         *
         * @param {HTMLElement} el
         * @param {boolean} show
         * @param {boolean} active
         */
        function setActiveControl(el, show, active) {
            if(active) {
                if(activeCtl !== el) {
                    //the new active control....
                    activeCtl = el;
                    if(el.tagName === 'INPUT') {
                        var data = Forms.getFormData(el);
                        //this hold the full field value so we can track if it changed
                        activeCtl.dataset.origColor = valueToHex(data[el.name], '', el.dataset.colorMode);
                        setIconColor(el);
                    } else {
                        //probably a button
                        activeCtl = el;
                        //this hold the full field value so we can track if it changed
                        activeCtl.dataset.origColor = valueToHex(el.dataset.color, '#000', el.dataset.colorMode);
                        //set icon color into the control
                        setIconColor(el);
                    }
                }
                if(show) showPicker(el);
            } else {
                activeCtl.dataset.origColor = '';
                activeCtl = null;
                selectedCell = -1;
            }
        }

        /**
         * Form Instance KeyBoard Event Handling
         * @param {KeyboardEvent} e
         */
        function onInputKeypress(e) {
            switch(e.keyCode || e.key || e.which) {
                case 32://SPACE
                case 13://ENTER
                    if(picker.open) {
                        e.preventDefault();
                        e.stopImmediatePropagation();
                    } else {
                        //commit the current value...
                    }
                    break;
            }
        }
        /**
         * Handler for keyup
         * @param {KeyboardEvent} e
         */
        function onInputKeyup(e) {
            switch(e.keyCode || e.key || e.which) {
                case 27://escape
                    if(picker.open) {
                        //if open then close else undo to starting value
                        hidePicker();
                        e.preventDefault();
                        e.stopImmediatePropagation();
                    }
                    break;
            }
        }
        /**
         * Handler for keydown
         * @param {KeyboardEvent} e
         */
        function onInputKeydown(e) {
            var keyCode = e.keyCode || e.key || e.which;
            if(picker.open) {
                //dropdown is active
                var block = true;
                switch(keyCode) {
                    case 32://SPACE
                    case 13://ENTER
                        //transfer the viewDate and commit
                        hidePicker();
                        commitValue();
                        break;
                    case 38://up -7
                        if(e.altKey) {
                            hidePicker();
                        } else {
                            moveCellY(-1);
                        }
                        break;
                    case 40://down +7
                        if(e.altKey) {
                            showPicker(e.target);
                        } else {
                            moveCellY(1);
                        }
                        break;
                    case 37://left -1
                        if(e.ctrlKey || e.metaKey) moveCellX(-1);
                        else block = false;
                        break;
                    case 39://right +1
                        if(e.ctrlKey || e.metaKey) moveCellX(1);
                        else block = false;
                        break;
                    default:
                        block = false;
                        break;
                }
                if(block) { e.preventDefault(); e.stopImmediatePropagation(); }
            } else {
                //dropdown is not active
                switch(keyCode) {
                    case 40:
                        if(e.altKey) {
                            showPicker(e.target);
                            e.preventDefault(); e.stopImmediatePropagation();
                        }
                        break;
                    default:
                        break;
                }
            }
        }
        init();
        /**
         * Handler for all the callbacks from the main input controller
         * @param {Event} e
         * @returns {boolean} returns true for the Form Controller's default processing
         */
        function onInputEvent(e) {
            switch(e.type) {
                case 'input': return true;
                case 'change': return false;
                case 'focus': return onInputFocus(e);
                case 'blur': return onInputBlur(e);
                case 'keyup': return onInputKeyup(e);
                case 'keydown': return onInputKeydown(e);
                case 'keypress': return onInputKeypress(e);
                default:
                    return true;
            }
        }

        /**
         * Init the input element
         * @param {HTMLElement} el
         * @param {HTMLFormElement} elForm
         */
        function initInputElement(el, elForm) {
        }
        /**
         * Update a form controls value
         * When is the element anything other than an input?
         * Maybe when it is a button.... (the button drives of the dataset and doesn't have a ctx....)
         * Apparently because the devs aren't setting the color mode we have to check for an rgb value being passed in
         * although surely this must cause a conversion on writing value as there is no check to see whether the value is required as rgb or not
         * @param {HTMLElement} el
         * @param {*} data the row data object -  the control 'value' is in a 'field' in the data
         */
        function setFieldValue(el, data) {
            //orig color is always stored as hex
            if(el.tagName === 'INPUT') {
                el.dataset.origColor = el.value = valueToHex(data[el.name], '', el.dataset.colorMode);
            } else {
                el.dataset.origColor = valueToHex(el.dataset.color, '#000', el.dataset.colorMode);
            }
            setIconColor(el);
        }
        /**
         * Sets the controls icon color.
         * If the control is active then it also highlights a color cell in the picker
         * if the value is one of the preselected colors
         * @param {HTMLElement} el
         */
        function setIconColor(el) {
            var icon, clr;
            //find the icon and get the ctl clr
            if(el.tagName === 'INPUT') {
                var els = el.parentNode.getElementsByTagName('icon');
                icon = (els.length > 0) ? els[0] : null;
                clr = (el.value || '').toUpperCase();
            } else {
                icon = el;
                clr = (el.dataset.color || '').toUpperCase();
            }
            //update the visuals
            if(icon) {
                if(clr) {
                    icon.style.backgroundColor = clr;
                    icon.classList.remove('i-list', 'i-fill');
                } else {
                    icon.style.backgroundColor = '';
                    icon.classList.add(icon.dataset.defaultIcon || 'i-list');
                }
            }
            //if active update the picker
            if(activeCtl === el) {
                //mark selection
                selectedCell = -1;
                var div = picker.querySelector('[data-key="' + clr + '"]');
                if(div) {
                    //remove the current selected
                    var cells = picker.getElementsByClassName('selected');
                    for(var i = 0; i < cells.length; i++) {
                        cells[i].classList.remove('selected');
                    }
                    //add the new selected
                    selectedCell = parseInt(div.getAttribute('data-index'));
                    div.classList.add('selected');
                }
            }
        }
        /**
         * Moves the selected cell horizontally when using the keyboard
         * @param {number} increment
         */
        function moveCellX(increment) {
            selectedCell += increment;
            if(selectedCell < 0) {
                selectedCell = 23;
            } else if(selectedCell >= 24) {
                selectedCell = 0;
            }
            setCtlValue(colors[selectedCell]);
            setIconColor(activeCtl);
        }

        /**
         * Moves the selected cell vertically when using the keyboard
         * @param {number} increment
         */
        function moveCellY(increment) {

            selectedCell += (increment * 6);
            if(selectedCell < 0) {
                //send it back to the bottom of the same column
                selectedCell += 24;
            } else if(selectedCell > 24) {
                //send it bck to the top of the same column
                selectedCell += -24;
            }
            setCtlValue(colors[selectedCell]);
            setIconColor(activeCtl);
        }

        /**
         *
         * @param hex
         * @param defaultValue
         * @param mode
         * @returns {*}
         */
        function hexToValue(hex, defaultValue, mode){
            mode = mode || 'hex';
            if(hex){
                return (mode==='rgb') ? hexToRgb(hex) : hex;
            }else{
                return defaultValue;
            }
        }

        /**
         *
         * @param {string|number} value
         * @param {string|number} defaultValue
         * @param {string} mode 'rgb' or 'hex'
         * @returns {string} valid hex color
         */
        function valueToHex(value, defaultValue, mode){
            mode = mode || 'hex';
            var hex = '';
            if(mode==='rgb'){
                hex = (value || value===0) ? '#'+rgbToHex(value) : defaultValue;
            }else{
                hex = value || defaultValue;
            }
            return hex.toUpperCase();
        }
        /**
         * Convert rgb (or part) to hex
         * @param rgb
         * @returns {string}
         */
        function rgbToHex (rgb) {
            return ('000000' + (Number(rgb).toString(16))).substr(-6);
        }
        // Expand shorthand form (e.g. "03F") to full form (e.g. "0033FF")
        var shorthandRegex = /^#?([a-f\d])([a-f\d])([a-f\d])$/i;
        function hexToRgb(hex) {
            // Expand shorthand form (e.g. "03F") to full form (e.g. "0033FF")
            hex = hex.replace(shorthandRegex, function(m, r, g, b) {return r + r + g + g + b + b;});

            var bigint = parseInt(hex, 16);
        }
        self.getColorIndex = getColorIndex;
        self.onInputEvent = onInputEvent;
        self.setFieldValue = setFieldValue;
        self.initInputElement = initInputElement;

        self.showDropDown = showDropDown;

        self.clickHandler = clickHandler;
        return self;
    };
    /**@type ColorPickerController*/
    window.ColorPicker = controls.ColorPickerController = new ColorPickerController();
    Widget.setDefaultOptions('cmcolor', { colorMode: 'hex' });
    //endregion


})(jQuery);