(function ($) {
    "use strict";
    //region Imports
    /**
     * @name controls
     * @namespace use the global cm.controls namespace
     */
    var controls = $.namespace("cm.controls", true);

    //endregion

    //region Dropdown

    /**
     * Controller for the Dropdown input control
     * Buttons don't get focus in Safari so this means the initial sequence of events differs
     * on chrome/windows its focus/click on safari/mac its click/focus
     * chrome on a return/spacebar key triggers a simulated 'click '
     * data for DD come from the module model - there is no built in RDC - and we only run after the module initialises its data
     * we dont cache any data
     * @returns {DropdownController}
     * @constructor
     */
    var DropdownController = function () {
        var self = this;
        //the context object for the active dropdown
        var ctx;
        //the dropdown dialog element
        var dd;
        //the dropdown dialog template
        var template = '' +
            '<dialog id="dropdown" class="select dropdown" data-position="dropdown"  tabindex="-1" data-action-topic="dialog:dropdown">' +
            '<header class="dialog-header">' +
                '<div class="input-control search flex">' +
                    '<input type="search" data-type="search" name="searchText" class="ignore clearable" incremental data-source-target="lookup" data-search-action="lookup-search" autocomplete="off" placeholder="Search..." title="Enter some text to search for or press the search button for the entire list">' +
                    '<icon class="search i-list helper" tabindex="-1"></icon>' +
                '</div>' +
            '</header>' +
            '<section class="dialog-content"></section>' +
            '<footer class="dialog-footer">' +
                '<button class="btn-dialog pad-r-m  margin-l-a" data-action="dialog-done">done</button>' +
            '</footer>' +
            '</dialog>';
        /**
         * Initialize the controller
         */
        function init() {
            //create global picker
            dd = DialogManager.addDialog(template, {owner: self});
        }
        init();
        /**
         * Init the input element
         * @param {HTMLElement} el
         * @param {HTMLFormElement} elForm
         * @memberOf DropdownController#
         */
        function initInputElement(el, elForm) {
            //just create the dropdown ctx object
            var controlCtx = Widget.getContext(el);
            if (controlCtx.iconField) {
                el.classList.add('uses-icons');
            }
            //always a button type input whether touch or not
            el.type = 'button';
        }
        /**
         * Show a dropdown
         * This is only called from a dropdown button click
         * It acts as a toggle
         * @param {HTMLElement} el the owning input element
         * @memberOf DropdownController#
         */
        function showDropDown(el) {
            var controlCtx = Widget.getContext(el);
            //if we already the active element then focus wont fire (I think)
            if (ctx === controlCtx) {
                // on safari we get click/focus but chrome it is focus/click
                console.log('show DD')
                //assume either kbd cmd or prg call or someone clicked the button
                showDropdownList(ctx.showDropDown, true);
            } else {
                console.log('show DD focus')
                // on safari we get click/focus but chrome it is focus/click
                // so simply put focus back on the input control but flag as DD needed and we will trigger there
                controlCtx.showDropDown = !dd.open;
                el.focus();
            }
        }
        /**
         * Show the dropdown element
         * @param {boolean} [state] the requested state
         * @param {boolean} [reload] do we need to rebuild the html list
         */
        function showDropdownList(state, reload) {
            //open the dropdown anchor to the editor and flag the editor as active
            //if state is undefined then toggle
            if (state === undefined) state = !dd.open;
            if (state) {
                //open it if currently closed - reload if required
                if (reload) {
                    //rebuild the list markup
                    loadData();
                }
                //make sure we have selected the current item
                setSelectedItem();
                if (!dd.open) dd.show(ctx.el);
            } else {
                //close it if currently open
                ctx.showDropDown = false;
                if (dd.open) dd.close();
            }

            ctx.showDropDown = undefined;
        }
        /**
         * Handler for list item click events
         * @param {Event} e
         * @param {View} view
         * @param {HTMLOListElement} ol
         * @param {HTMLLIElement} li
         * @returns {boolean}
         * @memberOf DropdownController#
         */
        function selectHandler(e, view, ol, li) {
            if (li) {
                // select the target item, check it, and refocus the proxy input control
                // the click will have caused a blur event so we need to refocus the control
                setSelectedItem(li, true);
            }
            return false;
        }
        /**
         * Handler for Dropdown Input blur event
         * If clicking to another tab or window two blur events fire - ignore the second one
         * @param e
         */
        function onInputBlur(e) {
            if (ctx) {
                if (Focus.isFocusOut(e, ctx.el, dd)) {
                    console.log('BLUR DD')
                    //we are really losing focus from the control
                    if (!e.force || (e.force && e.willBlur)) ctx.pendingBlur = true;
                    commitValue();
                }
            }
        }
        /**
         * Handler for the Dropdown Input focus event
         * If clicking in from another tab or window two focus events fire - ignore the second one
         * @param e
         */
        function onInputFocus(e) {
            if (!e.target.readOnly) {
                var controlCtx = Widget.getContext(e.target);
                if (controlCtx !== ctx) {
                    console.log('FOCUS DD')
                    //only if we are really becoming active
                    ctx = controlCtx;
                    // get the row so we can update from it
                    // this is usually the ds si but on a CF its a sub prop si.CF.values
                    ctx.row = Forms.getFormData(ctx.el);
                    //now load the dd list
                    loadData();
                    //and open if necessary
                    if (ctx.showDropDown) showDropDown(ctx.el);
                }
            }
        }

        /**
         * Handler for the input event
         * @param {Event} e
         */
        function onInputInput(e) {
        }

        /**
         *
         * @param e
         * @returns {boolean}
         */
        function onInputKeyup(e) {
            var killkey = false;
            // a key event..... we only trap key events to look for the esc key otherwise we use the search/change event
            var keyCode = e.keyCode || e.key || e.which;
            switch (keyCode) {
                case 13://ENTER
                case 27://ESC
                case 32://SPACEBAR
                    killkey = true;
                    break;
            }
            if (killkey) {
                e.preventDefault();
                e.stopPropagation();
            }

        }
        /**
         * Keydown event handler for the input control
         * @param {Event} e
         */
        function onInputKeydown(e) {
            if (ctx) {
                var killkey = false;
                var keyCode = e.keyCode || e.key || e.which;
                switch (keyCode) {
                    case 9://tab
                        //close
                        showDropdownList(false);
                        break;
                    case 38: // up
                        if (Key.ctrlKey) {
                            //toggle close
                            showDropdownList(false);
                        } else setSelectedItem('prev', true);
                        killkey = true;
                        break;
                    case 40: // down
                        if (Key.ctrlKey) {
                            //toggle open
                            showDropdownList(true);
                        } else setSelectedItem('next', true);
                        killkey = true;
                        break;
                    case 13:// return
                        showDropdownList(false);
                        commitValue();
                        ctx.el.focus();
                        killkey = true;
                        break;
                    case 32:// space
                        showDropdownList(false);
                        commitValue();
                        ctx.el.focus();
                        killkey = true;
                        break;
                    case 8://backspace
                        break;
                    case 27://escape
                        //if open then close else undo to starting value
                        showDropdownList(false);
                        killkey = true;
                        break;

                }
                if (killkey) {
                    e.preventDefault();
                    e.stopPropagation();
                }
            }
        }

        /**
         * Set the selected index and update the UI to reflect
         * note selected is NOT checked
         * @param {string|number|HTMLLIElement} [item] keyword prev, next or index value
         * @param {boolean} commit
         */
        function setSelectedItem(item, commit) {
            //and flag as selected - note selected is not the same as checked
            //note this does not commit or check a value, simply moves the 'selected' marker
            //the current item
            var el = dd.querySelector('li.selected');
            //unselect the current selected item
            if (el) el.classList.remove('selected');
            //get parent list
            var parentList = el ? el.parentElement : dd.querySelector('ol');
            //the new item - default first item, prev/next based on current or index value
            el = (item === undefined) ? dd.querySelector('LI[data-key="' + ctx.el.dataset.key + '"]') || dd.querySelector('LI') :
                item === 'prev' ? (el && el.previousElementSibling) || parentList.firstChild :
                    item === 'next' ? (el && el.nextElementSibling) || parentList.lastChild :
                        item instanceof HTMLElement ? item : dd.querySelector('LI[data-index="' + item + '"]');
            //select the new item
            if (el) el.classList.add('selected');
            //reposition the dd because behavior is to move the dd not the cursor
            dd.info.ref = el;
            DialogManager.positionDropdown(dd, ctx.el, dd.info);
            if(commit){
                commitValue();
                showDropdownList(false);
            }
            //refocus the input
            ctx.el.focus();
        }

        /**
         * Commit the current selected item update the input control value an d related properties
         */
        function commitValue() {
            if (ctx.cache) {
                var el = dd.querySelector('li.selected');
                if(el) {
                    var label = ctx.cache[+el.dataset.index].l;
                    var key = $.stringToValue(el.getAttribute('data-key'));
                    var changed = false;

                    if (ctx.row[ctx.el.name] !== label) {
                        //take current selected and 'commit' it
                        ctx.el.value = label;
                        if (ctx.keyField) ctx.row[ctx.keyField] = key;
                        if (ctx.iconField) ctx.row[ctx.iconField] = el.getAttribute('data-icon');
                        ctx.row[ctx.el.name] = ctx.el.value;
                        if (ctx.keyField) ctx.el.dataset.key = key;

                        //this will fire the change event - and only for the active control - ie not if we move off the row
                        Forms.commitValue(ctx.el);
                        changed = true;
                    }
                }else{
                    //focus gets in to the control but we dont do anything and then click away
                    //debugger
                }
            }
            //if a panel is open and you click on a dropdown on the base page the ctx is getting cleared out causing the dropdown to fail
            //when a item is selected. If the dropdown is active we shouldn't clear out the ctx
            var isDD = ctx.el.getAttribute('data-type') === 'dropdown' && ctx.el.classList.contains('active') ? true : false;
            //if this commit is part of a blur then complete the blur
            if (ctx.pendingBlur && !isDD) {
                ctx.cache = null;
                ctx.row = null;
                ctx.pendingBlur = false;
                ctx = null;
            }
            return changed;
        }

        /**
         * Load the dropdown list, retrieves the data and builds the markup
         */
        function loadData() {
            if (ctx.source) {
                var data = ctx.cache =  (ctx.source.indexOf('_vl') === 0) ? valueListCache[ctx.source].list : Views.findView(ctx.el).view.module.model[ctx.source] || [];
                if (data) {
                    loadListItems(data);
                }
            }
        }
        /**
         * Build the pick list markup and flag the checked and selected items
         * @param {Array} data the array of lookup values
         */
        function loadListItems(data) {
            if (data) {
                //set the pick list into the DD
                dd.childNodes[1].innerHTML = getDataMarkup(data);
            }
        }

        /**
         * Builds the dropdown list markup string
         * @param {Array} data
         * @returns {string}
         */
        function getDataMarkup(data) {
            var len = data.length;
            var row;
            var html = '';
            for (var i = 0; i < len; i++) {
                row = data[i];
                html += '<li class="data-row nofocus ' + (row.icon ? 'list-icon ' + row.icon : '') + '" data-key="' + row.k + '" data-icon="' + row.icon + '" data-index="' + i + '">';
                html += '<div class="hbox"><div class="margin-r-a">' + System.localize(row.l) + '</div></div>';
                html += '</li>';
            }
            html += '</ol>';
            return '<ol class="listview-dropdown" data-type="listview-dropdown">' + html + '</ol>';
        }
        /**
         * Update a form controls value
         * @param {HTMLInputElement} el
         * @param {*} data the row data object -  the control 'value' is in a 'field' in the data
         * @memberOf DropdownController#
         */
        function setFieldValue(el, data) {
            //do we need to update?
            if (ctx && ctx.el === el) {
                // if there is an active selectlist and the element is the one we are targeting then ignore it
                // we never update active controls here
            } else {
                var fv = data[el.name] || '';
                var kv = data[(el.dataset.keyField || 'key')] || '';
                if (fv !== el.value) {
                    el.value = fv;
                    el.dataset.key = kv;
                }
            }
        }
        /**
         * Handler for all the callbacks from the main input controller
         * @param {Event} e
         * @returns {boolean} returns true for the Forms Controller's default processing
         * @memberOf DropdownController#
         */
        function onInputEvent(e) {
            //no kbd text input or search in dd
            switch (e.type) {
                case 'input': return onInputInput(e);
                case 'change': return false;
                case 'focus': return onInputFocus(e);
                case 'blur': return onInputBlur(e);
                case 'keyup': return onInputKeyup(e);
                case 'keydown': return onInputKeydown(e);
                default:
                    return true;
            }
        }

        /**
         * topic handler for the pick list dialog
         * @param {string} topic
         * @param {Event} ae
         * @memberOf DropdownController#
         */
        function actionTopicHandler(topic, ae) {
            switch (ae.action) {
                case 'lookup-search':
                    search(ae.actionKey);
                    break;
                case 'dialog-done':
                    //mobile hit the done button
                    dd.close();
                    ctx.el.focus();
                    break;
                case 'dropdown.dialog-close':
                case 'dialog-close':
                    //the DD closed
                    if (ctx.showDropDown == false) {
                        //a commanded close ignore....
                        ctx.showDropDown = undefined;
                    } else {
                        //uncommanded close we need to force the commit and refocus
                        commitValue();
                        // but if you clicked from one DD to another then close is uncommanded
                        // but you want to focus the other control not this one ie pendingBlur was true...
                        if(ctx)ctx.el.focus();
                    }
                    break;

            }
        }
        //tmp cache for report options values
        var valueListCache = {};
        /**
         * Add a temp cached list
         * @param {string} values delimited list of values uses '|' as delimiter
         * @param {boolean} noKeys true if the values do not contain keys (custom fields)
         * @param {string} [delim] the list delimeter
         * @param {string} [listID] the list cache id
         * @memberOf DropdownController#
         */
        function addValueList(values, noKeys, delim, listID) {
            //check if list already exists
            for (var o in valueListCache) {
                if (valueListCache[o].values === values) {
                    return o;
                }
            }
            //not found so create a new one
            var list = values.split(delim || '|');
            if (values.indexOf(':') === -1 || noKeys) {
                for (var i = 0, len = list.length; i < len; i++) {
                    var v = list[i];
                    list[i] = { i: v, k: v, l: v };
                }
            } else {
                for (var i = 0, len = list.length; i < len; i++) {
                    var v = list[i].split(':');
                    list[i] = { i: v[0], k: $.stringToValue(v[1]), l: v[0] };
                }
            }

            listID = '_vl' + (listID || ($.guid++));
            valueListCache[listID] = { list: list, values: values };
            return listID;
        }


        self.addValueList = addValueList;
        self.onInputEvent = onInputEvent;
        self.setFieldValue = setFieldValue;
        self.initInputElement = initInputElement;

        self.showDropDown = showDropDown;
        self.selectHandler = selectHandler;
        self.actionTopicHandler = actionTopicHandler;
        return self;
    };
    //shortcut
    /** @type DropdownController*/
    window.Dropdown = controls.DropdownController = new DropdownController();
    //register defaults - the label/value goes into name so we dont specify it here
    Widget.setDefaultOptions("dropdown", { source: '', iconField: '', keyField: 'key' });
    //endregion


})(jQuery);