(function ($) {
    "use strict";
    //region Imports
    /**
     * @name controls
     * @namespace use the global cm.controls namespace
     */
    var controls = $.namespace("cm.controls", true);

    //endregion

    //region Time
    /**
     * Controller for the Time input control
     * A time control only exists if there is an associated date control i.e. it must be under DTM control.
     * set the type attribute to one of the DTM types i.e. time.Activity (see below)
     * any time value typed in to the input is accepted if the picker is not displayed
     * if anything is typed when the picker is displayed it should close the picker and then accept anything typed
     * @returns {TimePickerController}
     * @constructor
     */
    var TimePickerController = function(){
        var self = this;
        //the picker UI
        var picker;
        //the context for the current active control element
        var ctx;
        //array of time values
        var time;

        function init(){
            var html = getMarkup();
            picker = DialogManager.addDialog('<dialog id="timepicker" class="timepicker popover picker" data-position="picker" tabindex="-1"  data-action-topic="dialog:timepicker"><header class="dialog-header"><span class="date-time-header"></span></header><section class="dialog-content scrollable">' + html + '</section><footer class="dialog-footer"><button class="btn-dialog" data-action="dialog-clear">Clear</button><button class="btn-dialog margin-l-a" data-action="dialog-done">Done</button></footer></dialog>', {position: 'picker'});
            time = [];
            var d = new Date(0,0,1);
            var dt = '', t = '', fmt = '';
            for (var i=0; i < 96; i++) {
                d.setHours(Math.floor(i/4),(i%4)*15);
                dt = d.toISOString();
                fmt = formatValue('t', dt);
                t = dt.split('T')[1].split('.')[0];
                time[i] = {dt: t, t: t, fmt: fmt};
            }
        }

        /**
         * Returns markup string for the picker UI
         * @returns {string}
         */
        function getMarkup() {
            var html = '';
            html += '<table class="grid-timepicker" data-type="am">';
            html += '<caption>AM/PM</caption>';
            html += '<tr class="data-row"><td>12</td><td>01</td><td>02</td><td>03</td><td>04</td><td>05</td></tr>';
            html += '<tr class="data-row"><td>06</td><td>07</td><td>08</td><td>09</td><td>10</td><td>11</td></tr>';
            html += '</table>';

            html += '<table class="grid-timepicker" hidden data-type="pm">';
            html += '<caption>PM</caption>';
            html += '<tr class="data-row"><td>12</td><td>01</td><td>02</td><td>03</td><td>04</td><td>05</td></tr>';
            html += '<tr class="data-row"><td>06</td><td>07</td><td>08</td><td>09</td><td>10</td><td>11</td></tr>';
            html += '</table>';


            html += '<table class="grid-timepicker" data-type="mins">';
            html += '<tr class="data-row"><td>00</td><td>15</td><td>30</td><td>45</td></tr>';
            html += '</table>';

            html += '<table class="grid-timepicker" data-type="ampm">';
            html += '<tr class="data-row"><td>AM</td><td>PM</td></tr>';
            html += '</table>';

            return html;
        }
        /**
         * Internal routine to display the picker UI
         * @param {HTMLElement} target the anchor usually an input control but can be a trigger button
         * @param {boolean} [state]
         * @param {boolean} [focus]
         * @param {boolean} [commit]
         */
        function showDropdownList(target, state, focus, commit) {
            if (state) {
                //focus triggers immediately on calling it firing the blur then the focus except in IE11 where it happens on the next micro loop
                if (focus && document.activeElement !== target) {
                    console.log('CALLING FOCUS');
                    if ($.browser.isIE11) {
                        if (ctx) {
                            //do a blur
                            showDropdownList(ctx.el, false, false, true);
                            setActiveControl(ctx.el, false);
                        }
                        //do a focus
                        setActiveControl(target, true);
                        //then call the real focus
                        target.focus();
                    } else {
                        target.focus();
                    }
                }
                //make sure the display is up to date and we have set a default time
                setSelectedIndex(getCurrentTime(true), true, true);
                if (commit) commitValue();
                // now open the picker
                if (!picker.open) picker.show(ctx.el);
            } else {
                //focus triggers immediately on calling it firing the blur then the focus except in IE11 where it happens on the next micro loop
                if (picker.open) picker.close();
                if (focus && document.activeElement !== target) {
                    if ($.browser.isIE11) {
                        if (ctx) {
                            //do a blur
                            showDropdownList(ctx.el, false, false, true);
                            setActiveControl(ctx.el, false);
                        }
                        setActiveControl(target, true);
                        target.focus();
                    } else {
                        target.focus();
                    }
                }
                if (commit) commitValue();
            }
        }
        /**
         * Toggle the display state of the picker dropdown UI
         * Usually called from clicking the dd button in an input control
         * @param {HTMLElement} el the target usually an input control
         * @param {Boolean} [state]
         */
        function showDropDown(el, state){
            state = state || (state === undefined && picker.open && ctx.el === el) ? false : true;
            showDropdownList(el, state, state, false);
        }
        /**
         * Handle picker clicks
         * @param {Event} e
         * @param view
         * @param cur
         * @param prev1
         * @param prev2
         * @param prev3
         * @returns {boolean}
         */
        function clickHandler(e, view, cur, prev1, prev2, prev3) {
            //find what was clicked
            var tgt = e.target;

            if (tgt.tagName === 'TD') {
                var cell = tgt;
                var ci = cell.cellIndex;
                var ri = cell.parentElement.rowIndex;
                var table = cell.parentElement.parentElement.parentElement;
                var selected = table.querySelector('.selected');
                if (selected) selected.classList.remove('selected');
                cell.classList.add('selected');
                var type = table.dataset.type;
                var h = 0;
                var m = 0;
                //everything defaults to 0
                switch (type) {
                    case 'ampm':
                        //implies we are only using bank 1
                        var ampm = ci * 12;
                        //get hours
                        var hCell = cur.querySelector('[data-type=am] .selected');
                        var mCell = cur.querySelector('[data-type=mins] .selected');
                        m = mCell ? mCell.cellIndex * 15 : 0;
                        h = (hCell ? hCell.cellIndex + (hCell.parentElement.rowIndex * 6) : 0) + ampm;
                        picker.querySelector('.date-time-header').innerText = h + ':' + m;
                        break;
                    case 'am':
                        // could be bank 1 and 2 or bank1 and ampm
                        var ampmCell = cur.querySelector('[data-type=ampm] .selected');
                        var ampm = (ampmCell ? ampmCell.cellIndex : 0) * 12;
                        //calc hours
                        h = ((ri * 6) + ci) + ampm;
                        //get mins
                        var mCell = cur.querySelector('[data-type=mins] .selected');
                        m = mCell ? mCell.cellIndex * 15 : 0;
                        picker.querySelector('.date-time-header').innerText = h + ':' + m;
                        break;
                    case 'pm':
                        //implies we are using bank 1 and 2
                        //make sure bank1 gets deselected
                        var hCell = picker.querySelector('[data-type=am] .selected');
                        if (hCell) hCell.classList.remove('selected');
                        //calc hours
                        h = 12 + ((ri * 6) + ci);
                        //get mins
                        var mCell = cur.querySelector('[data-type=mins] .selected');
                        m = mCell ? mCell.cellIndex * 15 : 0;
                        picker.querySelector('.date-time-header').innerText = h + ':' + m;
                        break;
                    case 'mins':
                        var ampmCell = cur.querySelector('[data-type=ampm] .selected');
                        var ampm = (ampmCell ? ampmCell.cellIndex : 0) * 12;

                        var hCell = cur.querySelector('[data-type=am] .selected');
                        h = (hCell ? hCell.cellIndex + (hCell.parentElement.rowIndex * 6) : 0) + ampm;
                        if (!h && !ampm) {
                            hCell = cur.querySelector('[data-type=pm] .selected');
                            h = (hCell ? hCell.cellIndex : 0);
                        }
                        m = ci * 15;
                        break;
                    default:
                        break;

                }
                //this will update the header display
                setSelectedIndex(((h * 60) + m) / 15, true, true);
            }

        }

        /**
         * Move the current selection indicator
         * @param dir the direction to move -1 up 1 down
         */
        function moveSelect(dir) {
            //make sure there is something to key off
            var idx = (ctx.selectedIndex === -1) ? getCurrentTime(true) : ctx.selectedIndex;
            //now move it
            setSelectedIndex(idx + dir, true);
        }

        /**
         * Input values
         * 830, 8:30,19:30,1230,130,530,630,0700,700,100
         * if has ':' contains hh and mm
         * if no ':' is present then
         * if 2 chars then only hh
         * if 3 or 4 then hh plus mm
         * @param {boolean} returnIndex if true returns the index in the time array otherwise returns the ISO Time Value
         * @returns {number|string}
         */
        function getCurrentTime(returnIndex) {
            //need to select and scroll something....
            //either we have a value - or if no value and useDefault then use current time....
            var v = ctx.el.value;// || (useDefault ? DateTime.getDefaultDate(ctx.el, 'time') : '');
            if (v) {
                var amAdj = v.toUpperCase().indexOf('A') !== -1;
                var pmAdj = v.toUpperCase().indexOf('P') !== -1;
                var t = v.replace(/\D/g, '');
                var h = 0;
                var m = 0;
                var tlen = t.length;
                switch (tlen) {
                    case 1:
                        //hours only
                        h = parseInt(t);
                        m = 0;
                        break;
                    case 2:
                        //hours only
                        h = parseInt(t) % 24;
                        m = 0;
                        break;
                    case 3:
                        //mins in last 2
                        h = parseInt(t.substr(0, 1));
                        m = parseInt(t.substr(1, 2)) % 60;
                        break;
                    case 4:
                        //mins in last 2
                        h = parseInt(t.substr(0, 2)) % 24;
                        m = parseInt(t.substr(2, 2)) % 60;
                    default:
                        //mins in last 2
                        h = parseInt(t.substr(0, 2)) % 24;
                        m = parseInt(t.substr(2, 2)) % 60;
                        break;

                }
                if (amAdj) {
                    h -= h === 12 ? 12 : 0;
                }
                if (pmAdj) {
                    h += h === 12 ? 0 : h > 12 ? 0 : 12;
                }
                if (!pmAdj && !amAdj) {
                    //make a guess
                    h += h < 7 ? 12 : 0
                }

                return returnIndex ? h * 4 + Math.floor(m / 15) : (('00' + h).substr(-2)) + ':' + (('00' + m).substr(-2)) + ':00';
            } else return returnIndex ? -1 : '';
        }
        /**
         * Set the selected index and update the UI to reflect
         * @param {number} index
         * @param {boolean} [updateValue]
         * @param {boolean} [commit]
         */
        function setSelectedIndex(index, updateValue, commit) {
            if (index < 0 || index > 95) {
                ctx.selectedIndex = -1;
                ctx.fmttime = '';
                if (updateValue) ctx.el.value = ctx.fmttime;
                var nl = picker.querySelectorAll('.selected');
                for (var i = 0, len = nl.length; i < len; i++) {
                    nl[i].classList.remove('selected');
                }
            } else {
                //note this does not commit a value, simply moves the 'selected' marker
                ctx.selectedIndex = index % 96;
                var selTime = time[ctx.selectedIndex];
                ctx.fmttime = selTime.fmt;
                if (updateValue) ctx.el.value = ctx.fmttime;
                var nl = picker.querySelectorAll('.selected');
                for (var i = 0, len = nl.length; i < len; i++) {
                    nl[i].classList.remove('selected');
                }

                //and then update the current date display
                picker.querySelector('.date-time-header').innerText = selTime.fmt;
                var m = ctx.selectedIndex % 4;
                var h = (ctx.selectedIndex - m) / 4;
                var ampm = h >= 12 ? 1 : 0;
                if (h >= 12) {
                    //update both the 12 and 24 panels just in case.....
                    picker.querySelector('[data-type=am]').rows[(h >= 18 ? 1 : 0)].cells[h % 6].classList.add('selected');
                } else {
                    picker.querySelector('[data-type=am]').rows[(h >= 6 ? 1 : 0)].cells[h % 6].classList.add('selected');
                }
                picker.querySelector('[data-type=mins]').rows[0].cells[m].classList.add('selected');
                picker.querySelector('[data-type=ampm]').rows[0].cells[ampm].classList.add('selected');
            }
            commitValue();
        }

        /**
         * Commit the value if it is valid
         * Called from blur or keydown-return
         * All datetime object from server do not have ms or tz set so in order to properly detect change we have to match
         */
        function commitValue(){
            var t='';
            if (ctx.selectedIndex === -1) {
                // kbd driven - per gg tabbing into and out of a control should do nothing
                if (ctx.el.value) {
                    t = getCurrentTime(false);
                } else {
                    //no value
                    t = ''
                }
            }else{
                //click driven - use selected item
                t = time[ctx.selectedIndex].t;
            }
            //if the current time differs from the orig datetime (last commited) then we need to handle it
            if (ctx.time !== t) {
                ctx.time = t;
                //set value into the form data-this will update the row data and may change other fields
                //this returns the full field value
                var dt = DateTime.commitInputField(ctx.el, 'time', ctx.dtm, t, Forms.getFormData(ctx.el.form));
                if (ctx.orig !== dt) {
                    ctx.orig = dt;
                    //update the control
                    ctx.el.value = ctx.fmttime = formatValue('t', dt);
                    //fires the notify event and marks control and form dirty
                    Forms.commitValue(ctx.el);
                } else {
                    //it is possible we changed the value to null but DTM reset it becuase of required back to old value
                    //the actual T time value
                    ctx.time = ctx.orig ? ctx.orig.split('T')[1] : '';
                    //formatted display time
                    ctx.fmttime = ctx.orig ? formatValue('t', ctx.orig) : '';
                    //update the control
                    ctx.el.value = ctx.fmttime = formatValue('t', dt);
                    //set the current selectedIndex
                    ctx.selectedIndex = getCurrentTime(true);
                }
            }
        }
        /**
         * Handler for keyup
         * @param {KeyboardEvent} e
         */
        function onInputKeyup(e){
            switch(e.keyCode||e.key|| e.which){
                case 27://escape
                    if(picker.open){
                        //if open then close else undo to starting value
                        showDropdownList(ctx.el, false, true, false);
                        e.preventDefault();
                        e.stopImmediatePropagation();
                    }
                    break;
                case 13://ENTER
                    if (e.target === picker) {
                        //commit the current picker value
                        showDropdownList(ctx.el, false, true, true);
                    }
                    break;
            }
        }
        /**
         * Handler for keydown
         * @param {KeyboardEvent} e
         */
        function onInputKeydown(e){
            var keyCode = e.keyCode||e.key|| e.which;
            //block the keystroke
            var block = true;
            switch(keyCode){
                case 9://TAB
                    if (e.target === picker) {
                        showDropdownList(ctx.el, false, true, true);
                    }
                    break;
                case 13://ENTER
                    if (e.target === picker) {
                        block = false;
                        break;
                    }
                case 32://SPACE
                    //the dropdown maybe open or not if it is close but dont save
                    showDropdownList(ctx.el, false, true, true);
                    break;
                case 38://up -7
                    if(e.altKey){
                        showDropdownList(ctx.el, false, true, false);
                    }else moveSelect(-1);
                    break;
                case 40://down +7
                    if(e.altKey){
                        showDropdownList(ctx.el, true, true, false);
                    }else moveSelect(1);
                    break;
                case 37://left -1
                case 39://right +1
                case 48://0
                case 49://1
                case 50://2
                case 51://3
                case 52://4
                case 53://5
                case 54://6
                case 55://7
                case 56://8
                case 57://9
                case 65://A
                case 77://M
                case 80://P
                case 186://:
                case 59://: on FF Mac

                //10 key number pad
                case 96://0
                case 97://1
                case 98://2
                case 99://3
                case 100://4
                case 101://5
                case 102://6
                case 103://7
                case 104://8
                case 105://9

                case KeyCode.BACKSPACE:
                case KeyCode.DELETE:
                    //close the dropdown and let the key thru we are in key mode not mouse
                    ctx.selectedIndex = -1;
                    showDropdownList(ctx.el, false, true, false);
                    block = false;
                    break;
                default:
                    block = true;
                    break;
            }
            if(block){
                e.preventDefault();
                e.stopImmediatePropagation();
            }
        }
        /**
         * Handler for the Date Input focus event
         * @param e
         */
        function onInputFocus(e){
            console.log('FOCUSIN')
            //focus can be transferred from the dd or the dd button or the dd or another control
            //we only care if the active control changes
            setActiveControl(e.target, true);
        }
        /**
         * Handler for Date Input blur event
         * @param e
         */
        function onInputBlur(e){
            //focus may be only going to the picker or the toggle button
            if (Focus.isFocusOut(e, ctx.el, picker)) {
                console.log('FOCUSOUT')
                showDropdownList(e.target, false, false, true);
                if (!e.force || (e.force && e.willBlur)) setActiveControl(ctx.el, false);
            }
        }

        /**
         * orig the ISO DateTime field value
         * time the time part of the ISO value
         * fmttime the time part locally formatted
         * @param el
         * @param active
         */
        function setActiveControl(el, active) {
            if (active) {
                if (el.tagName === 'INPUT') {
                    var controlCtx = Widget.getContext(el);
                    if (ctx !== controlCtx) {
                        //setup the new context
                        ctx = controlCtx;
                        //get the row data
                        ctx.data = Forms.getFormData(el);
                        //the real field value
                        ctx.orig = ctx.data[el.name] || '';
                        //the actual T time value
                        ctx.time = ctx.orig ? ctx.orig.split('T')[1] : '';
                        //formatted display time
                        ctx.fmttime = ctx.orig ? formatValue('t', ctx.orig) : '';
                        //set the current selectedIndex
                        ctx.selectedIndex = -1;
                        //if newly active then hide the dropdown - if we clicked the helper button this will still open on the next action
                        showDropdownList(ctx.el, false, false, false);
                    }
                } else {
                    //a trigger button - currently doesnt exist
                    var controlCtx = {el: el};
                    if (ctx.el !== controlCtx.el) {
                        ctx = controlCtx;
                        showDropdownList(ctx.el, false, false, false);
                    }
                }
            } else {
                ctx.data = null;
                ctx.orig = ctx.fmttime = ctx.time = '';
                ctx.selectedIndex = -1;
                ctx = null;
            }
        }
        /**
         * Handler for all the callbacks from the main input controller
         * @param {Event} e
         * @returns {boolean} returns true for the Forms Controller's default processing
         */
        function onInputEvent(e){
            switch(e.type){
                case 'input':return true;
                case 'change':return false;
                case 'focus':return onInputFocus(e);
                case 'blur':return onInputBlur(e);
                case 'keyup':return onInputKeyup(e);
                case 'keydown':return onInputKeydown(e);
                default:
                    return true;
            }
        }

        /**
         * Init the input element
         * @param {HTMLElement} el
         * @param {HTMLFormElement} elForm
         */
        function initInputElement(el, elForm) {
            var ctl = Widget.getContext(el);
            el.maxLength = 8;
            //in mobile the control doesn't directly take input but becomes a button and a fake  input is shown in the dropddown
            if (Device.status.touchActive) el.type = 'button';
        }
        /**
         * Update a form controls value
         * The control displays the fomratted value
         * The data stores the field value
         * Only the DTM can/knows how to write to the field
         * @param {HTMLInputElement} el
         * @param {*} data the row data object -  the control 'value' is in a 'field' in the data
         */
        function setFieldValue(el, data){
            var fv = formatValue('t', data[el.name]);
            if(fv!==el.value)el.value = fv;
        }

        /**
         * topic handler for the pick list dialog
         * @param {string} topic
         * @param {Event} ae
         */
        function actionTopicHandler(topic, ae) {
            switch (ae.action) {
                case 'dialog-done':
                    showDropdownList(ctx.el, false, true, true);
                    break;
                case 'dialog-clear':
                    //close the dropdown and let the key thru
                    if (picker.open) picker.close();
                    ctx.selectedIndex = -1;
                    ctx.el.value = '';
                    break;

            }
        }

        init();

        self.onInputEvent = onInputEvent;
        self.setFieldValue=setFieldValue;
        self.initInputElement = initInputElement;

        self.actionTopicHandler = actionTopicHandler;

        self.clickHandler = clickHandler;

        self.showDropDown=showDropDown;
        return self;
    };
    //shortcut
    /**@type TimePickerController*/
    window.TimePicker = controls.TimePickerController =  new TimePickerController();
    Widget.setDefaultOptions("time", {dtm:''});
    Widget.setDefaultOptions("cmtime", {dtm:''});
    //endregion


})(jQuery);