(function ($) {
    "use strict";
    //region Imports
    /**
     * @name controls
     * @namespace use the global cm.controls namespace
     */
    var controls = $.namespace("cm.controls", true);

    //endregion

    //region Form
    /**
     * The static singleton form controller
     * All forms are backed by this controller
     * @class
     * @name FormController
     */
    var FormController = function() {
        var self = this;
        //region GlobalHandlers
        /**
         * Global handler for the Form reset/cancel event
         * We block this and use our own internal action
         * @param {Event} e
         */
        function resetHandler(e) {
            e.preventDefault();
        }
        /**
         * Global handler for the Form submit event
         * We block this and use our own internal action
         * @param {Event} e
         */
        function submitHandler(e) {
            //we always block the actual form submission
            e.preventDefault();
        }

        /**
         * Global handler for the createElement callback
         * @param {string} type
         */
        function viewsNotifyHandler(type) {
            switch(type) {
                case 'createElement':
                    var el = arguments[1];
                    //FYI we trigger on potentially disconnected nodes
                    //find all bindable forms
                    var forms = el.getElementsByTagName('form');
                    for(var i = 0; i < forms.length; i++) {
                        initForm(forms[i], true);
                    }
                    //any standalone input controls
                    var ctls = el.querySelectorAll('.cm-standalone > input');
                    for(var i = 0; i < ctls.length; i++) {
                        initInputElement(ctls[i], null);
                    }
                    //if this is a list inside a fieldset check for bindable inputs - lists must process second to prevent double processing of controls
                    if(el instanceof HTMLOListElement && el.parentElement instanceof HTMLFieldSetElement) {
                        //find all controls and init
                        var ctls = el.querySelectorAll('input,textarea');
                        for(var i = 0; i < ctls.length; i++) initInputElement(ctls[i], null);
                    }
                    break;
                case 'removeElement':
                    var el = arguments[1];
                    if(el.tagName === 'FORM') {
                        initForm(el, false);
                    }
                    break;
                case 'activateElement':
                    var el = arguments[1];
                    //find all bindable forms
                    var forms = el.getElementsByTagName('form');
                    for(var i = 0; i < forms.length; i++) {
                        initForm(forms[i], true);
                    }
                    break;
            }
        }

        //endregion

        //region initform
        /**
         * Init a form
         * A form will automatically begin editing the current selected item when it is instantiated
         * It can end editing at any time by some explicit or implicit action
         * @param {HTMLFormElement} el
         * @param {boolean} state true for creating flase for destroying
         */
        function initForm(el, state) {
            if(el.classList.contains('ignore')) {
                //so that devs can screw the system up.....
            } else {
                if(state) {
                    //get the context
                    var form = Widget.getContext(el);
                    //ignore if already initialised
                    if(!form.init) {
                        //the view element
                        var vel = Views.findView(el);
                        //the view object
                        var view = vel.view;
                        //set init false - force to false from undef
                        form.init = false;
                        form.initialising = true;
                        //time of last save
                        form.lastSave = 0;
                        //set the generated uid....
                        el.id = form.uid;
                        //ok store a ref to the form on the View
                        view.forms[form.id] = el;
                        //set the variant class
                        if(el.dataset.type) el.classList.add(el.dataset.type);
                        //if no source is specified create a default one based on the form id
                        if(!form.source) form.source = form.id;
                        //setup the error panel
                        switch(form.formErrors) {
                            case 'none'://no error panel exists or should be created
                                break;
                            case 'custom'://an error panel exists somewhere within the >FORM<
                                //so we assume there is an error panel somewhere it is located in a notify panel
                                form.errorPanel = form.el.querySelector('notify');
                                break;
                            case 'grid':
                                //underneath the form
                                form.errorPanel = form.el.nextElementSibling;
                                break;
                            case 'panel':
                                // this is the standard - we add the error block
                                // support for entity errors - used in setting screen where a single row is split over several forms - each form is regarded as an entity
                                // in FF if you set the form attribute then form.errors is udf until later
                                // IE11 doesn't support name attribute on a fieldset
                                el.insertAdjacentHTML('afterbegin', '<notify data-error-entity="' + form.formErrorEntity + '"></notify>');
                                //get ref to panel - use tgt as we might have switched it
                                form.errorPanel = el.firstChild;
                                break;
                            case 'panel:last':
                                el.insertAdjacentHTML('beforeend', '<notify data-error-entity="' + form.formErrorEntity + '"></notify>');
                                //get ref to panel - use tgt as we might have switched it
                                form.errorPanel = el.lastChild;
                                break;
                            default://external - somewhere in the >VIEW< using an #id
                                //an error panel exists somewhere within the view - formErrors is its #id - we check our local panel view and if not found we check the page view
                                form.errorPanel = vel.querySelector('#' + form.formErrors.replace('custom:', '')) || view.el.querySelector('#' + form.formErrors.replace('custom:', ''));
                                break;
                        }
                        // before we process the individual input elements we have to process anything that generates content
                        // an event to allow the module to generate dynamic form content initTemplate
                        //TODO: deprecate this - should use fieldset tags instead
                        var detail = {};
                        Views.fireActionEvent(el, 'initTemplate', false, false, detail);
                        if(detail.tpl) {
                            var html = Template.renderTemplate(Template.buildSubTemplate(detail.tpl), [{}]);
                            var tgt = (el.firstChild.classList.contains('scrollable')) ? el.firstChild : el;
                            tgt.insertAdjacentHTML('beforeend', html);
                            //in case there were any lists or grids or bindings in that markup
                            Views.elementCreated(tgt);
                        }

                        //make sure the CF associative array  exists - note there could be more than one FS so the other FS might have created it
                        form.CF = {};
                        //find the dynamic fieldsets
                        var fs = el.querySelectorAll('fieldset[data-mode]');
                        for(var i = 0, len = fs.length; i < len; i++) {
                            Fieldset.initFieldset(fs[i]);
                        }
                        //store the fs array
                        form.fs = fs;

                        // are we a sticky button.... sticky buttons float in a sticky panel outside the main section
                        // we have changed our id so we need to update any references back to us
                        var sc = vel.querySelectorAll('button[form=' + form.id + ']');
                        if(sc.length) {
                            vel.dataset.sticky = 'true';
                            //the new live instance id
                            var formInstanceID = el.id;
                            for(var i = 0, len = sc.length; i < len; i++) {
                                sc[i].setAttribute('form', formInstanceID);
                                //point the data attribute to the original form id
                                sc[i].dataset.form = form.id;
                            }
                        }

                        //find ctl-groups and wrap the children
                        var groups = el.querySelectorAll('.ctl-group');
                        if(groups.length) {
                            var docfrag, grp, div;
                            for(var i = 0, ilen = groups.length; i < ilen; i++) {
                                grp = groups[i];

                                docfrag = document.createDocumentFragment();
                                if(docfrag.append) {
                                    docfrag.append.apply(docfrag, grp.children);
                                } else {
                                    var children = grp.children;
                                    for(var j = 0, jlen = children.length; j < jlen; j++) {
                                        docfrag.appendChild(children[0]);
                                    }
                                }

                                div = document.createElement('div');
                                div.classList.add('ctl-group-inner');
                                div.appendChild(docfrag);
                                grp.appendChild(div);
                            }
                        }
                        // make sure that any dynamic controls have been injected before the main loop
                        // fieldsets tend to inject on data rather than initialisation
                        for(var i = 0, ctl; (ctl = el.elements[i]) ; i++) {
                            if(inputTypes[ctl.tagName]) {
                                initInputElement(ctl, el);
                            }
                        }

                        //external controls - these need initialising - buttons obviously don't
                        var xc = vel.querySelectorAll('input[form=' + form.id + ']');
                        if(xc.length) {
                            for(var i = 0, len = xc.length; i < len; i++) {
                                xc[i].setAttribute('form', el.id);
                                initInputElement(xc[i], el);
                            }
                        }
                        // get embed mode and embed parent
                        // an ol that is not a vs or a accordiona, a grid, a slideout panel, main page, a dialog
                        var pel = el.closest('OL,DATAGRID,PANEL,PAGE,DIALOG,CMVIEWSTACK');
                        form.embedMode = pel.tagName === 'OL' ? 'list' : pel.tagName === 'DATAGRID' ? 'grid' : pel.tagName === 'PANEL' ? 'panel' : pel.tagName === 'DIALOG' ? 'dialog' : pel.tagName === 'CMVIEWSTACK' ?  'viewstack' : '';
                        form.owner = pel;

                        //process prompts....
                        if(form.prompt) {
                            form.prompts = {};
                            var d = form.prompt.split(',');
                            var p, n, t;
                            for(var i = 0, len = d.length; i < len; i++) {
                                p = d[i].split(':');
                                n = p[0].toUpperCase();
                                form.prompts[n] = p[1] || Notify.getPrompt(n);
                            }
                        } else {
                            form.prompts = (form.embedMode === 'list' || form.embedMode === 'grid') ? pel.ctx.prompts : {};
                        }
                        //is this form an embedded form
                        el.classList.toggle('form-embed',form.embedMode === 'list' || form.embedMode === 'grid') ;
                        //create the tab group
                        Focus.tabGroup(el, true);
                        // now open the form -note so far we haven't set any 'data' - just initialised the controls
                        setEditMode(el, form.initMode);
                        //non passive handler to block number input scrolling
                        el.addEventListener('wheel',mouseWheelHandler,VMouse.getListenerOptions(false, false));

                        // init done
                        form.initialising = false;
                        form.init = true;
                        form.el.classList.add('init');
                    } else {
                        //form already been initialized.....
                        //do we ever get here - well when we call init on a widget panel and also the widget itself
                    }
                } else {
                    //close the form
                    setEditMode(el, 'destroy');
                }
            }
        }

        //endregion

        //region values
        /**
         *
         * @param {HTMLElement|Array} el
         * @returns {string}
         * @memberOf FormController#
         */
        function getFieldName(el){
            if(Array.isArray(el)) el = el[0];
            var field = el.name;
            return (field && field[0]==='_') ? field.substr(1) : field;
        }
        //just in case no form or binding is specified a control can fallback to this object
        var defaultFormData = {};
        /**
         * Return form data
         * @param {HTMLFormElement} el
         * @param {boolean} ds if true returns the ds instead of the row
         * @param {boolean} reload reload data from source
         * @param {boolean} commit force active element to commit
         * @returns {*}
         * @memberOf FormController#
         */
        function getFormData(el, ds, reload, commit) {
            if(el) {
                if(el instanceof HTMLFormElement) {
                    if(commit) {
                        commitActiveElement(el);
                    }
                    var form = Widget.getContext(el);
                    if(reload) {
                        return (el.dataset.editKey) ? form.ds.findByKey(el.dataset.editKey) : form.ds.si;
                    } else return ds ? form.ds : form.formdata || form.ds.si;
                } else {
                    //we need to check if this element is a CF INPUT and if so we need to get the CF Values rather than the row data
                    var form = Widget.getContext(el.form);
                    var formdata = form.formdata || form.ds.si;
                    var field = getFieldName(el);
                    return (form.CF.hasCF && form.CF[field]) ? formdata[form.CF[field]].values : formdata;
                }
            } else return defaultFormData;
        }

        /**
         * Set form data and update the controls
         * If the data is the same as the existing then we just update the controls
         * If its a new object we reset the errors etc
         * An issue is that if we get a new row (or a source change) and we already have focus the control may not update
         * @param {HTMLFormElement} el
         * @param {*} data the row data object
         * @memberOf FormController#
         */
        function setFormData(el, data) {
            var form = Widget.getContext(el);
            //get form elements - form els appear as array indexed by number and named props - radios will be in a RadioNodeList
            var els = el.elements;
            //get the data
            var fd = form.formdata;
            //is it a new data object
            var dataChange = (fd !== data);
            //if the activelement and form then set the focusctl 
            var focusCtl;
            if(dataChange) {
                //console.log('NEW DATA')

                /*
                  On the viewport of the schedule, when clicking the arrows to move from one task to another, the blur event is firing which in turn
                  is firing the field change event which the dev is handling to do something. The expectation is that when the field value changes then 
                  another value needs to change.  However, this doesn't apply when the si changes.  In order to handle the si changing we needed to move 
                  the code from for loop below to here.  See issue 442973 for more details
                */
                if(document.activeElement && document.activeElement.form === el){
                    focusCtl = document.activeElement;
                    if (NoChangeTypes[getInputType(document.activeElement)]){
                        document.activeElement.blur();
                    }
                }

                //the form data object changed - a new row or new source!!
                //store the data - default in an empty object so we dont blow on accessing data
                fd = form.formdata = data || {};
                // clear all form errors - clear all field errors
                window.ValidationManager.clearFormErrors(el);
                window.ValidationManager.showFormErrors(el, '', fd.errors);
                //create a backup copy of the source data before we edit it...
                form.old = JSON.stringify(fd, formDataReplacer);
                //setup/init dynamic fields
                for(var i = 0, len = form.fs.length; i < len; i++) {
                    Fieldset.setFormData(form.fs[i], fd);
                }
                if(form.fileSource && form.filePath){
                    form.ds.model[form.fileSource].source = data[form.filePath];
                }
            }
            //check the dynamic fields first in case they inject anything
            for(var i = 0, len = form.fs.length; i < len; i++) {
                Fieldset.setFieldValue(form.fs[i], fd);
            }
            //now update the control values themselves (els is a live object)
            if(els) {
                //not all fields are bound so we cant loop over the data - we loop over the els
                //we ignore named items as they duplicate indexed items (note radios show once as a named array and then individually as indexed items)

                var field, ctl;
                for(var i = 0; i < els.length; i++) {
                    ctl = els[i];
                    field = getFieldName(ctl);
                    if(isNaN(i)) {
                        //ignore named items.... for now .... use force if it is a new data object... if CF pick up the correct sub data values to pass on
                    } else{
                        setFieldValue(ctl, form.CF.hasCF && form.CF[field] ? fd[form.CF[field]].values : fd, dataChange);
                    }
                }

                // set the focus back into the activeelement only if the focusctl is set
                if(focusCtl){
                    focusCtl.focus();
                }
            }
            //if the data object changed and we have an active element then need to force load the data this is done by the control itself in the setFieldValue
        }

        /**
         *
         * @param {Array} els array of elements to update
         * @param {HTMLFormElement|Object} el the form element or the form context
         * @memberOf FormController#
         */
        function updateFieldValues(els, el){
            var form = (el instanceof HTMLFormElement) ? Widget.getContext(el) : el;
            //now update the control values themselves (els is a live object)
            if(els) {
                //not all fields are bound so we cant loop over the data - we loop over the els
                //we ignore named items as they duplicate indexed items (note radios show once as a named array and then individually as indexed items)
                var fd = form.formdata;
                var field, ctl;
                for(var i = 0; i < els.length; i++) {
                    ctl = els[i];
                    field = getFieldName(ctl);
                    if(isNaN(i)) {
                        //ignore named items.... for now .... use force if it is a new data object... if CF pick up the correct sub data values to pass on
                    } else{
                        setFieldValue(ctl, form.CF.hasCF && form.CF[field] ? fd[form.CF[field]].values : fd, true);
                    }
                }
            }
        }
        /**
         * Helper for creating data backup - so we can restore later
         * excludes any 'children' array
         * @param key
         * @param value
         * @returns {undefined}
         */function formDataReplacer(key, value) {
            return key === 'children' ? undefined : value;
        }
        //endregion

        //region formedit
        /**
         * Callback from a DS when a value has changed
         * It could be a simple field change or a completely new row
         * @param {HTMLFormElement} form the target form
         * @param {*} item the item that changed
         * @param {DataSource} list the DS source the form is bound to
         * @param {string} type the data change type
         * @param {string} field the data change field
         */
        function updateFormElement(form, item, list, type, field) {
            switch(type) {
                case 'error':
                    //only care about errors on current row were editing...
                    if(form.ctx.formdata === item) window.ValidationManager.showFormErrors(form, field, item.errors);
                    break;
                case 'select'://select a new row
                    if(form.ctx.init) {
                        //only place we set active/inactive from is here
                        setEditMode(form, item.empty ? 'inactive' : 'active');
                    }
                case 'render'://or could render on a new ds
                case 'item'://or some update????
                case 'field'://should be no change in the row
                    if(!form.ctx.inactive) {
                        //we usually update to the si becuase we are set on the si but allow for none si editing....
                        setFormData(form, getFormData(form, false, true));
                    }
                    break;
                case 'state'://rowstate changed - on our row? - fix it up
                    if(form.ctx.formdata === item && !item.rowstate) {
                        //if the row is clear and we are in grid mode then we can close whatever we are in......
                        if(form.getAttribute('data-commit-action') === 'close') Views.close(Views.findView(form));
                    }
                    break;
                case 'select-before':
                    //autosave on selection change if we are bound to that item and it is dirty
                    if(form.ctx.formdata === item && form.ctx.autoSave && isFormDirty(form)) {
                        //need to force a save - field contains the details
                        autosave(form, field);
                    }
                    break;
                default:
                    break;
            }
        }
        /**
         * Switch edit mode.
         * 'saving'://after firing the save event if the submitMode is saving
         * 'cancel'://after pressing a cancel button
         * 'close'://after firing the save event if the submitMode is close
         * 'editing'://called from the elementCreateHandler
         * 'destroy'://called from elementRemoveHandler
         * At start Binds DS, stores current data state, finds focusable
         * At end wipes out all existing data and clears all flags - puts the form beyond use?
         * Devs are using saved, saved:close, continue
         * DataGrid uses endEdit
         * form.mode editing/close/destroy - currently only used by lookup to see if we are editing or not
         * setEditMode, setFormState, isDirty, itemChanges, updateFormElement use inactive
         * setEditMode,itemChanges set it isDirty and updateElment block on it
         * @param {HTMLFormElement} el
         * @param {string=} action one of continue/cancel/close - if undefined the forms is being destroyed
         * @memberOf FormController#
         */
        function setEditMode(el, action) {
            if(el) {
                var form = Widget.getContext(el);
                var view = Views.findView(el);
                switch(action) {
                    case 'endEdit':
                    case 'save':
                        save(el, 'none');
                        break;
                    case 'inactive':
                        //special mode where it just stops reacting to data changes
                        form.inactive = true;
                        el.classList.toggle('active', false);
                        el.classList.toggle('inactive', true);
                        break;
                    case 'active':
                        //special mode were it resumes previous mode and start listening for data changes
                        //now any data changes will update controls
                        form.inactive = false;
                        el.classList.toggle('active', true);
                        el.classList.toggle('inactive', false);
                        //flag as begin changes
                        itemChanges(form, form.ds, 'begin');
                        //now set the data - this updates the input controls
                        setFormData(el, getFormData(el, false, true));
                        break;
                    case 'saving'://after firing the save event if the submitMode is saving
                        //generally called from save - 'save' button - puts form in a save state whilst saving
                        form.mode = 'saving';
                        break;
                    case 'searching':
                    case 'saving:editing'://alias for resume - we are currently saving but we are continuing to edit
                        form.mode = 'editing';
                        form.inactive = false;
                        el.classList.toggle('active', true);
                        el.classList.toggle('inactive', false);
                        //flag as begin changes
                        itemChanges(form, form.ds, 'begin');
                        break;
                    case 'saved:errors'://alias for resume
                        form.mode = 'editing';
                        setEditMode(el, 'active');
                        break;
                    case 'saved:close'://alias for done
                        DataManager.commitModifiedItems(form.ds, form.formdata);
                        form.mode = 'close';
                        Views.dismiss(form.dismiss, el, true);
                        break;
                    case 'saved'://alias for continue but commits the current data
                        DataManager.commitModifiedItems(form.ds, form.formdata);
                    case 'continue':
                        //generally called by dev after returning from a server 'save' call to signify the form can continue to be used
                        //because we are continuing with the same data some things won't auto reset
                        //backup the 'new' old data
                        form.old = JSON.stringify(form.formdata);
                        //we should also clear out the old data - should we reload defaults..
                        setEditMode(form.el, 'editing');
                        break;
                    case 'cancel:close'://the caller will handle the close - in response to a no on a confirm save
                        form.mode = 'close';
                    case 'cancel'://after pressing a cancel button
                        // a 'cancel' implies undoing any changes......whereas just a 'close' doesn't!!!!!!!!!!
                        // like a save without a save
                        // dismiss rules for 'auto' - if on page or panel - close it but can't close a base page
                        //tell the ds were done - if it was an insert it can end with the data being removed
                        itemChanges(form, form.ds, 'cancel');
                        //what next? if we have a dismiss mode then the form is going to close so ntg but if its none then we will continue editing
                        if(form.mode === 'close') {
                            //already in close mode, the caller is handling it so we do nothing
                        } else {
                            if(form.dismiss === 'none') {//only used by deliverables AFAIK
                                //no dismiss just push into the cancelMode
                                setEditMode(form.el, form.cancelMode);
                            } else {///auto - so figure it out
                                //close it and let Views sort out what to do next
                                form.mode = form.cancelMode;
                                Views.dismiss(form.dismiss, el, true);
                            }
                        }
                        break;
                    case 'delete':
                        // the edit row was deleted - this happens after the delete
                        if(form.embedMode) {
                            form.mode = 'close';
                            Views.dismiss(form.dismiss, el, true);
                        } else {
                            form.mode = 'close';
                            Views.dismiss(form.dismiss, el, true);
                        }
                        break;
                    case 'edit-next':
                        // after firing the save event if the savedMode/submitMode is edit-next in a grid or list....
                        // its a forced close, go to next row and insert if required and then edit in current active column
                        // dismiss controls what gets closed....
                        form.mode = 'edit-next';
                        Views.dismiss(form.dismiss, el, true);
                        break;
                    case 'close':
                        // after firing the save event if the savedMode/submitMode is close....
                        // its a forced close as we are cleared to cancel or close so no need to ask about dirty....
                        // dismiss controls what gets closed....
                        // the form is finally nuked from the remove element handler so we dont do anything here
                        form.mode = 'close';
                        Views.dismiss(form.dismiss, el, true);
                        break;
                    case 'editing'://called from the elementCreateHandler to kiick things off
                        //the final part of init but I guess there is no harm in deferring
                        //find the ds then bind in the change handler
                        if(!form.ds) form.ds = view.view.module.model.get(form.source);
                        //bind the change handler to the source now - inactive will block any updating
                        if(!form.changeHandler) {
                            form.ds.change.add(form.changeHandler = function(item, list, type, field) {
                                updateFormElement(el, item, list, type, field);
                            });
                        }

                        //the real 'start to edit' phase
                        //set our 'mode'
                        form.mode = 'editing';
                        //flag embed containers as editing
                        if(form.embedMode === 'list') {
                            form.el.closest('LISTVIEW').ctx.isEditing = true;
                        }
                        if(form.embedMode === 'grid') {
                            form.el.closest('DATAGRID').ctx.isEditing = true;
                        }
                        //if there is a ds but the si is empty skip it and wait
                        form.inactive = form.ds.si.empty;
                        el.classList.toggle('active', !form.ds.si.empty);
                        el.classList.toggle('inactive', form.ds.si.empty);

                        // push form into its initial state - this will also focus something if we are active
                        // don't reset state back to original state on a 'continue'
                        setFormState(form.el, form.state, 'initial');
                        //update the display group
                        setDisplayGroupState(form.el, '', false);

                        if(!form.inactive) {
                            //notify we are going to change data
                            itemChanges(form, form.ds, 'begin');
                            //now set the data - this updates the input controls
                            setFormData(el, getFormData(el, false, true));
                        }

                        break;
                    case 'destroy'://only called from elementRemoveHandler - expects 'close' has been called first
                        // you cant comeback from destroy!!!
                        // generally a view will delete all form references when it closes
                        // but if this was a form in a repeater the view may not be closing its just the form that is going
                        // if we are closing due to form save then we will run the end edit but
                        // if closing due to panel/page close we dont
                        if(!form.inactive  ) {
                            if(view.view.type !== 'page' && view.view.type !== 'system') {
                                itemChanges(form, form.ds, 'end-force');
                            }
                            //now any data changes will not update controls
                            form.inactive = true;
                            el.classList.toggle('active', false);
                            el.classList.toggle('inactive', true);
                        }
                        // get rid of the ref to the form on the View
                        delete view.view.forms[el.id];
                        //generally called either from destructor or post a server 'save' call to shut the form
                        form.mode = 'destroy';
                        //unbind the change handler from the source now
                        view.view.module.model.get(form.source).change.remove(form.changeHandler);
                        //drop the old data
                        form.ds = null;
                        form.old = null;
                        form.formdata = null;
                        form.changeHandler = null;
                        form.fs = null;
                        form.owner = null;
                        el.classList.remove('form-embed');
                        //remove listener
                        el.removeEventListener('wheel',mouseWheelHandler);
                        //remove key bindings
                        Focus.tabGroup(el, false);
                        break;
                }
            }
        }
        /**
         * Begins or ends making changes to an item
         * This is called 'before' saving or editing starts
         * all controls will have committed and validations will have been called
         * this is just a call to tell any listener to get out of edit mode
         * definitive what happens next
         * @param {Object} form
         * @param {DataSource|DataItem} ds
         * @param {string} state one of 'end' 'begin' 'cancel' 'delete' 'invalid'
         */
        function itemChanges(form, ds, state) {
            var item = form.formdata || ds.si;
            var dsOpt = ds.options;
            var dsMode = ds.mode;//flat,tree,group,grouptree
            switch(state) {
                case 'delete':
                    // delete row in ds - reverts to initial values but normally removal is deferred,
                    // this will auto refresh anyway
                    // 'I' will trigger cancel and will remove and have an option (hideDeletes) for immediate removal
                    form.inactive = true;
                    form.el.classList.toggle('active', false);
                    form.el.classList.toggle('inactive', true);
                    //now revert the data
                    restoreFormData(form.formdata, form.old);
                    //set deleted
                    ds.itemUpdated('D', item, true);
                    //notify delete
                    ds.fire('changes-delete');
                    //notify end
                    ds.fire('changes-end');
                    // no reason to fire anything as its not actually been removed - unless it was 'I' -
                    // there is a standard refresh coming which will redo the UI just in case its a remove immediate or 'I'
                    break;
                case 'invalid':
                    // 'IN' can't be invalid as there was no user input so only 'I' or 'U'
                    // we don't rollback but we don't save either but we do commit ('update') locally so...
                    // totals might need updating, group/tree summary rows might change, sorts and filters need applying
                    // if editing changed a group or key field in a tree or group ds then we need to refresh the whole thing
                    form.inactive = true;
                    form.el.classList.toggle('active', false);
                    form.el.classList.toggle('inactive', true);
                    //notify end
                    ds.fire('changes-end');
                    //refresh to update summaries and groups/tree
                    ds.refresh();
                    break;
                case 'end-cancel':
                case 'cancel':
                    //reverts everything back to what it was but as we don't push item changes until now nothing to undo
                    form.inactive = true;
                    form.el.classList.toggle('active', false);
                    form.el.classList.toggle('inactive', true);
                    //now revert the data
                    restoreFormData(form.formdata, form.old);
                    //notify cancel - !!intentional drop through
                    ds.fire('changes-cancel');
                    //if still insert mode then remove it and redo everything......
                    if(form.autoDelete && item.rowstate === 'IN') {
                        //only used by grid/list at moment....
                        //inserted row that has never been modified this will auto refresh
                        ds.itemUpdated('D', item, true);
                        //notify of delete
                        ds.fire('changes-cancel-delete');
                    }
                    //notify end
                    ds.fire('changes-end');
                    //standard refresh will patch up UI in a moment
                    // totals and summaries need updating, sorts and filters need applying
                    // if editing changed a group or key field in a tree or group ds then we need to refresh the whole thing
                    // server will update later
                    ds.refresh();
                    break;
                case 'end-submit':
                case 'end-save':
                    //end with a save or submit
                    //just ends and commits and refresh everything - will call server and update again on return
                    form.inactive = true;
                    form.el.classList.toggle('active', false);
                    form.el.classList.toggle('inactive', true);
                    //fire save
                    if(state === 'end-save' && !dsOpt.batchMode) {
                        //need to update any rollups so individual row data is correct on the save (? maybe)
                        if(dsOpt.rollup) DataManager.rollupData(ds, dsOpt.rollup, dsOpt.summary, true, dsOpt.mode);
                        //now trigger GMI with a save
                        ds.getModifiedItems(true);
                    }
                    //notify end
                    ds.fire('changes-end');
                    // totals and summaries need updating, sorts and filters need applying
                    // if editing changed a group or key field in a tree or group ds then we need to refresh the whole thing
                    // server will update later
                    ds.refresh();
                    break;
                case 'end-force':
                    alert('what screen and what action caused this - please call rob now')
                    break;
                case 'begin':
                default:
                    // fire an event to flag into edit mode
                    // note the event assumes the si is being edited but it may not be
                    ds.fire('changes-begin');
                    break;
            }
        }

        /**
         * restore the prev version of a data object - compilcated because of children, CF or HDDLists
         * @param formdata
         * @param old
         * @returns {*}
         */
        function restoreFormData(formdata, old) {
            //check only if the row was dirty do we need to clean up ...
            if(formdata.rowstate) {
                var data = JSON.parse(old);
                for(var prop in formdata) {
                    if(prop !== 'children' && !data.hasOwnProperty(prop)) delete formdata[prop];
                }
                //now restore the old props
                Object.assign(formdata, data);
            }
            //return for chaining
            return formdata;
        }
        /**
         * Check whether a form and any related DataSources are dirty
         * @param {HTMLFormElement} el
         * @param {boolean} removeNewRow
         * @param {boolean} noCommit
         * @returns {boolean}
         * @memberOf FormController#
         */
        function isFormDirty(el, removeNewRow, noCommit) {
            el = (el instanceof HTMLFormElement) ? el : el.form;
            var form = el.ctx;
            if(form && !form.inactive) {
                //make sure the active element is committed before checking whether we are dirty or not
                if(!noCommit) commitActiveElement(el);
                //quick check if the current row already dirty - ignore none edited inserted rows
                //check current edit row first plus any child ds with dirty related to parent
                if(form.formdata && form.formdata.rowstate) {
                    if(form.formdata.rowstate === 'IN') {
                        if(removeNewRow) {
                            itemChanges(form, form.ds, 'cancel');
                        }
                        return false;
                    } else return true;
                } else {
                    //check any related ds
                    var rds = el.dataset.related;
                    if(rds) {
                        rds = rds.split(',');
                        var model = Views.findView(el).view.module.model;
                        for(var i = 0, len = rds.length; i < len; i++) {
                            if(model.get(rds[i]).isDirty()) {
                                //if a related ds is dirty force this row to also be dirty
                                form.ds.itemUpdated('U', form.formdata);
                                return true;
                            }
                        }
                    }
                }
            }
            return false;
        }
        /**
         * find the first dirty form in a view
         * dirtyPrompt(default), dirtySave, dirty
         * @param {HTMLElement} view
         * @param {boolean} findAll
         * @param {boolean} queryCloseMode
         * @param {boolean} removeNewRow
         * @returns {HTMLFormElement}
         * @memberOf FormController#
         */
        function findDirtyForm(view, findAll, queryCloseMode, removeNewRow) {
            // check whether there are any 'edited' forms that will block the close...
            var forms = view.getElementsByTagName('form');
            // handle dirty forms...
            // do we allow auto close if (1) no input or (2) dirty or if not we need to ask...
            // auto-close is either dirty/dirtyPrompt/dirtySave/none
            // can close is treated like an implicit cancel...
            var dirty=[];
            var autoClose={};
            for(var i = 0, len = forms.length; i < len; i++) {
                var el = forms[i];
                if(isFormDirty(el, removeNewRow) && (el.ctx.autoClose !== 'dirty')) {
                    //if its currently dirty
                    dirty.push(el);
                    autoClose[el.ctx.autoClose]=true;
                }
            }
            return dirty.length ? (queryCloseMode ? autoClose : (findAll ? dirty : dirty[0])) : null;
        }
        /**
         * find the first dirty grid in a view
         * queryclose, prompt(Default), query, dirty
         * @param {HTMLElement} view
         * @returns {HTMLTableElement}
         * @memberOf FormController#
         */
        function findDirtyGrid(view, findAll, queryCloseMode) {
            //grids can have unsaved error rows
            var grids = view.querySelectorAll('.datagrid');
            var dirty=[];
            var autoClose={};
            for(var i = 0, len = grids.length; i < len; i++) {
                var el = grids[i];
                var dsn = el.dataset.source;
                if(dsn) {
                    var ds = view.view.module.model.get(dsn);
                    if((ds.hasErrors || ds.isDirty()) && (el.ctx.layout.autoclose !== 'dirty')) {
                        //if its currently dirty
                        dirty.push(el);
                        autoClose[el.ctx.layout.autoclose]=true;
                    }
                }
            }
            return dirty.length ? (queryCloseMode ? autoClose : (findAll ? dirty : dirty[0])) : null;
        }

        //endregion

        //region other stuff


        /**
         * Gets a forms current state
         * @param {HTMLElement} el form
         * @memberOf FormController#
         */
        function getFormState(el) {
            return Widget.getContext(el).state;
        }
        /**
         * Change a forms current state
         * Disables' hidden inputs (so they don't validate) and sets focus to the first control or one marked autofocus
         * When a form is initialised at the same time as the view then offset parent will be undefined
         * if inactive we dont reload or focus
         * @param {HTMLElement} el form
         * @param {string} formState new form state, the form state is an optional three part string state:submitAction:focusField
         * @param {string} force if set the state is updated even if already in that state 'initial' or 'reset' - never used when really changing state
         * @memberOf FormController#
         */
        function setFormState(el, formState, force) {
            var form = Widget.getContext(el);
            var s = (formState || form.state).split(':');
            var state = s[0] || (form.state.indexOf(':') === 0 ? '' : form.state);
            var submitAction = s[1] || (form.submitAction);
            var focusField = s[2] || (form.el.getAttribute('autofocus'));

            if(force || form.state !== state) {
                //set new state - state must exist, you cant have an null state
                form.state = state;
                form.submitAction = submitAction;
                form.focusField = focusField;
                //now make new state changes
                if(form.state) {
                    //find and render any form-state fieldsets
                    var els = form.el.querySelectorAll('fieldset[data-mode=state]');
                    for(var i = 0; i < els.length; i++) window.Fieldset.initFieldset(els[i]);

                    //fix up the form state buttons that are being used like a tab so they are toggled active
                    var els = form.el.querySelectorAll('[data-form-state]');
                    for(var i = 0, len = els.length; i < len; i++) {
                        var item = els[i];
                        item.classList.toggle('active', item.classList.toggle('checked', item.getAttribute('data-form-state') === state));
                    }

                    //update all the new controls - but not needed when forcing to initial as caller will do this
                    if(force !== 'initial' && !form.inactive) setFormData(form.el, getFormData(form.el, false, true));

                    //push the state back into the binding variable
                    if(form.el.hasAttribute('data-bind-state')) {
                        BindingManager.updateNodeBindings(Views.findView(form.el), form.el, true);
                    }
                    //notify of change....
                    Views.fireActionEvent(form.el, 'stateChange', false, false, { state: form.state, submitAction: form.submitAction, focusField: form.focusField });
                    //set the focus control..... but only if its post load
                    if(document.documentElement.contains(form.el)) {
                        Focus.focusLater(Focus.findFirstFocusable(focusField ? form.el[focusField] || form.el : form.el));
                    }
                } else {
                    //no state used by this form so just refresh the controls - but not needed when forcing to initial as caller will do this
                    if(force !== 'initial' && !form.inactive) setFormData(form.el, getFormData(form.el, false, true));
                    //and set the focus control..... but only if its post load
                    if(!form.inactive && document.documentElement.contains(form.el)) {
                        Focus.focusLater(Focus.findFirstFocusable(focusField ? form.el[focusField] || form.el : form.el), true);
                    }
                }
            } else {
                //already in the correct state and not forcing so nothing to do but set focus anyway
                //set the focus control..... but only if its post load
                //IE11 offsetParent is not set even though the element is in the DOM, baseURI is also not set so we use html.contains instead
                if(!form.inactive && document.documentElement.contains(form.el)) {
                    Focus.focusLater(Focus.findFirstFocusable(focusField ? form.el[focusField] || form.el : form.el));
                }
            }
        }

        /**
         *
         * @param {HTMLFormElement} el
         * @param {string} name name of field to update
         * @param {boolean} [visible] ignored if not set otherwise changes the control
         * @param {boolean} [editable] ignored if not set otherwise changes the control
         * @param {boolean} [required] ignored if not set otherwise changes the control
         * @memberOf FormController#
         */
        function updateControlState(el, name, visible, editable, required){
            var ctl = el.elements[name];
            if(ctl && ctl.length===undefined) {
                if (visible === true || visible === false) {
                    ctl.parentElement.classList.toggle('bind-hidden', !visible);
                    Forms.setDisabled(el, !visible, 'vis');
                }
                if (editable === true || editable === false) ctl.readonly = editable;
                if (required === true || required === false) ctl.required = required;
            }
        }
        /**
         * Set the display state of a group of controls
         * There 3 use cases for a 'more' toggle button either
         * in a form or
         * in a view or
         * in a div using d-b-v that contains form controls
         * @param {HTMLFormElement|HTMLElement} el a form or view element or a div if grp is auto
         * @param {string} grp
         * @param {boolean} state false is hidden true is visible
         * @returns {boolean} the new state (for chainging)
         * @memberof FormController#
         */
        function setDisplayGroupState(el, grp, state) {
            if(grp) {
                //now make new state changes - item flagged or an input, textarea, contenteditable that are child of a flagged item
                var els = '';
                if(grp === 'auto') {
                    //from a data-bind-vis call that wants its child inputs disabled - it has the data-display-goup set to auto
                    els = $.NodeListToArray(el.querySelectorAll('input,textarea,.contenteditable,fieldset,.fieldset.required'));
                    els.unshift(el);
                } else {
                    //not dbv - this is from a toggle button passing in a form or view - get all eligible elements
                    els = $.NodeListToArray(el.querySelectorAll('.' + grp + ', .' + grp + ' input, .' + grp + ' textarea' + ', .' + grp + ' .contenteditable'));
                }
                for(var i = 0, len = els.length; i < len; i++) {
                    var ctl = els[i];
                    if(inputTypes[ctl.tagName]) {
                        //flag the wrapper so we can hide the input + helpers
                        ctl.closest('.input-control').classList.toggle('grp-hidden', !state);
                        //disable to prevent validation
                        ctl.classList.toggle('grp-disabled', !state);
                        checkDisabled(ctl);
                    } else {
                        //flag the wrapper so we can hide the input + helpers
                        ctl.classList.toggle('grp-hidden', !state);
                        //handle multiple radio/checkbox with required flag
                        if(ctl.classList.contains('fieldset')&&ctl.classList.contains('required'))ctl.classList.toggle('disabled', !state)
                    }
                }
            } else {
                // setting initial state - grp is only ever null when called by a form during form init
                // we want to set all the various grps in that form
                var els = el.querySelectorAll('[data-action=more]:not(.showmoreless)');
                var elSML = el.querySelectorAll('[data-action=more],[data-bind-toggle].showmoreless');
                for(var i = 0, len = els.length; i < len; i++) {
                    setDisplayGroupState(el, els[i].getAttribute('data-target'), state);
                    //fix up the group button
                    //el.querySelector('[data-target=' + grp + ']').value = state ? 1 : 0;
                    els[i].value = state ? 1 : 0;
                }
                if(elSML.length > 0) {
                    //set state
                    Buttons.setState(elSML, Views.findView(el).view.name, 0, el);
                }
            }
            return state;
        }

        //endregion
        function checkDisabled(el, isControl) {
            //should it really be disabled or not - check for the existence of any flag or simply the readOnly attribute
            var disabled = el.disabled = el.readOnly || el.classList.contains('vis-disabled') || el.classList.contains('grp-disabled') || el.classList.contains('edit-disabled') || el.classList.contains('action-disabled');
            //if its a control we need to also handle the wrapper
            if(isControl !== false) el.closest('.input-control').classList.toggle('disabled', disabled);
            if(el.editor) {
                el.editor.contentEditable = !disabled;
            }
            //return the current real state
            return disabled;
        }
        /**
         * Disables all the input controls in a form
         * It will also disable and hide any child element with the hideDisabled class
         * browser will validate disbaled but not readonly - becuase readonly is display only and couldnt have changed
         * so we have explicit flags for each of the mode of disabling a control and disable if any are true and enable if all are clear.
         * @param {HTMLElement} el
         * @param {boolean} state
         * @param {string} mode vis edit grp action
         * @param {string} editable  a csv string of elements that are not editable
         * @param {string} required a csv string of elements that are required
         * @memberOf FormController#
         */
        function setDisabled(el, state, mode, editable, required) {
            var disabled = !!state;
            switch(mode) {
                case 'grid'://grid form editor on open - passes editable and required csv strings
                    mode = 'edit-disabled';
                    //note these are the controls to be disabled....
                    if(required) required += ',';
                    if(editable) editable += ',';
                    //if you only passed some view we look for the first form.....
                    if(!(el instanceof HTMLFormElement)) el = el.getElementsByTagName('FORM')[0];
                    //make sure that any controls get injected before the main loop
                    for(var i = 0, ctl; (ctl = el.elements[i]) ; i++) {
                        if(inputTypes[ctl.tagName]) {
                            if(editable.indexOf(ctl.name + ',') === -1) {
                                ctl.classList.toggle(mode, disabled);
                                checkDisabled(ctl);
                            }
                            if(required.indexOf(ctl.name + ',') !== -1) ctl.required = true;
                        }
                    }
                    break;
                case 'action'://the 'save all' or 'prev' or 'next' buttons being db disabled
                    mode = mode + '-disabled';
                    //a single element
                    el.classList.toggle(mode, disabled);
                    checkDisabled(el, false);
                    break;
                case 'grp'://hide on setting group state
                /////////////case 'state'://hide on changing form state - form state no longer used like this - we used fieldset to add or remove controls
                case 'vis'://hide on a binding
                    mode = mode + '-disabled';
                    if(inputTypes[el.tagName]) {
                        //a single bound element
                        el.classList.toggle(mode, disabled);
                        checkDisabled(el);
                    } else if(el.tagName === 'BUTTON') {
                        //a single element
                        el.classList.toggle(mode, disabled);
                        checkDisabled(el, false);
                    }else if(el.classList.contains('fieldset')){
                        el.classList.toggle(mode, disabled);
                        el.classList.toggle('disabled', disabled);
                    } else {
                        //should be a FORM or FIELDSET or a DIV - i.e. a container - disable inputs but what about buttons well they can use data-bind-visible... hideReadonly
                        var els = el.querySelectorAll('input,textarea');
                        for(var i = 0, len = els.length; i < len; i++) {
                            var ctl = els[i];
                            el.classList.toggle(mode, disabled);
                            checkDisabled(ctl);
                        }
                        //other elements
                        var els = el.querySelectorAll('.hideDisabled');
                        for(var i = 0, len = els.length; i < len; i++) {
                            var ctl = els[i];
                            ctl.classList.toggle('disabled', disabled);
                            ctl.disabled = disabled;
                        }
                        var els = el.querySelectorAll('.showDisabled');
                        for(var i = 0, len = els.length; i < len; i++) {
                            var ctl = els[i];
                            ctl.classList.toggle('disabled', disabled);
                            ctl.disabled = !disabled;
                        }
                    }

                    break;
                case 'edit'://db editable or db enabled or db disabled
                default:
                    mode = 'edit-disabled';
                    if(inputTypes[el.tagName]) {
                        //a single element
                        el.classList.toggle(mode, disabled);
                        checkDisabled(el);
                    } else {
                        //should be a FORM or FIELDSET or a DIV - i.e. a container - disable inputs but what about buttons well they can use data-bind-visible... hideReadonly
                        var els = el.querySelectorAll('input,textarea');
                        for(var i = 0, len = els.length; i < len; i++) {
                            var ctl = els[i];
                            ctl.classList.toggle(mode, disabled);
                            checkDisabled(ctl);
                        }
                        //other elements
                        var els = el.querySelectorAll('.hideDisabled');
                        for(var i = 0, len = els.length; i < len; i++) {
                            var ctl = els[i];
                            ctl.classList.toggle('disabled', disabled);
                            ctl.disabled = disabled;
                        }
                        var els = el.querySelectorAll('.showDisabled');
                        for(var i = 0, len = els.length; i < len; i++) {
                            var ctl = els[i];
                            ctl.classList.toggle('disabled', disabled);
                            ctl.disabled = !disabled;
                        }
                    }
                    break;
            }
        }

        /**
         * A MDO causes an autoclose on embeded forms - determine what to do.
         * @param {Event} e the mouse or touch event
         * @param {HTMLElement} target object clicked on
         * @param {HTMLElement} vel view element
         * @memberOf FormController#
         */
        function autocloseHandler(e, target, vel) {
            // do we have an embedded form - all implicit closes of embedded forms are 'saves'
            // if there are any custom dialogs open then they must have failed validation and then we don't want to auto close a form
            if(!DialogManager.hasOpenDialog()) {
                var activeForms = vel.getElementsByClassName('active form-embed');
                //ok were good to go on closing down a grid or list embeded form
                var form = activeForms.length ? activeForms[activeForms.length - 1] : null;
                if(form) {
                    var ctx = form.ctx;
                    var owner = ctx.owner;
                    if(owner === target || owner.contains(target)) {
                        //in our grid/list
                        if(owner.tagName === 'OL') {
                            //in a list item or a grid row - don't close if clicked on a row (let the std click handler deal with it)
                            //no further processing is needed as we wait for the click to happen
                        } else if (owner.tagName === 'DATAGRID') {
                            // hit testing in the grid just got complicated...
                            // if we clicked the hScroller then its ok
                            // if we clicked the vScroller then its ok
                            // if we clicked the toolbar its ok
                            // if we clicked a row then its not ok
                            // if we clicked the footer its not ok
                            // if we clicked the header its not ok
                            // we should put focus back into the last (current) active control
                            var container = owner.ctx.gridSection;
                            if (container === e.target || container.contains(e.target) || e.target.closest('grid-scroller')) {
                                //in a grid row or scroller or divider or gr or gs
                                // - don't close if clicked on a row (let the std click handler deal with it)
                                // - but do close if clicked into blank area need to force close....
                                // no further processing is needed as we wait for the click to happen
                                if(target.tagName==='GRID-ROWS' || target.tagName==='GRID-SECTION'){
                                    setEditMode(form, 'endEdit');
                                }
                            } else {
                                // else in the list grid (maybe a header or footer or some empty space)
                                setEditMode(form, 'endEdit');
                            }
                        }
                    } else {
                        //clicked outside the list/grid but in another panel or the one we are in?
                        var view = Views.findView(target);
                        if(view.view.el.contains(form)) {
                            //we are clicking in the same page....
                            if(view.tagName === 'PANEL') {
                                //in a side panel
                                if(view.contains(form)) {
                                    // the same panel we are in - close
                                    setEditMode(form, 'endEdit');
                                } else {
                                    // in another panel belonging to our page
                                    // - DONT close -
                                    // we assume the grid opended the panel and will wait for it to close to refocus the grid editor
                                    form.classList.add('waiting', 'view-'+view.id);
                                    view.view.onClose[view.id] = focusActiveForm;
                                }
                            } else {
                                //in the main page - close it
                                setEditMode(form, 'endEdit');
                            }
                        } else {
                            // not in the same page must be on another page - close
                            setEditMode(form, 'endEdit');
                        }
                    }
                }
            }
        }

        /**
         * On closing a panel find the active form to focus
         * @param {string} pid panel id
         */
        function focusActiveForm(pid) {
            //ok were good to go on closing down a panel so refocus the grid editor
            var form = document.querySelector('form.waiting.view-'+pid);
            if(form) {
                var ctx = form.ctx;
                var owner = ctx.owner;
                form.classList.remove('waiting', 'view-'+pid);

                //focus the form if it hasnt already done so
                if (!form.contains(document.activeElement))Focus.findFirstFocusable(form).focus();

            }
        }
        /**
         * autosave on datasource navigation
         * @param el
         * @param detail
         */
        function autosave(el, detail) {
            var form = Widget.getContext(el);
            if(form.autoPrompt) {
                detail.cancel = true;//prevent the default behavior of the caller
                Notify.confirm('SAVE', true).always(
                    function(result) {
                        if(result === true) {
                            //wants to save first
                            save(el, form.submitAction, form.autoSubmitMode, false);
                        } else if(result === false) {
                            //no save no move back but must undo changes and undirty the form
                            restoreFormData(detail.item, form.old);
                            form.ds.si = detail.nextItem;
                        } else {
                            //no save but move back
                        }
                    }
                );
            } else {
                //default action is save but can be overridden and set to none
                save(el, form.submitAction, form.autoSubmitMode, false);
            }
        }

        /**
         *
         * @param tab
         * @param view
         * @memberOf FormController#
         */
        function saveAll(tab, view){
            var activeForm = tab.querySelector('form');
            var forms = tab.parentElement.querySelectorAll('form#'+activeForm.id+',form:not(#'+activeForm.id+')');
            //first try the active tab then in sequence
            save($.NodeListToArray(forms),'save', 'viewstack');
        }

        /**
         * force autofilled controls to commit values
         * @param el
         */
        function autofillForm(el){
            if(el.hasAttribute('autofill')) {
                //force commit any autofill ctl
                var els = el.elements;
                for(var i = 0, len = els.length; i < len; i++) {
                    var ctl = els[i];
                    if(ctl.hasAttribute('autofill') && !ctl.disabled) {
                        commitValue(ctl);
                    }
                }
            }
        }
        /**
         * Submit/Save a form
         * @param {HTMLFormElement} el
         * @param {string} [submitAction] the action to use on the save event (used by buttons)
         * @param {string} [submitMode] the mode to use whilst saving
         * @param {boolean} [force] force the form to save even if not dirty
         * @return {boolean} returns true if save successful
         * @memberOf FormController#
         */
        function save(el, submitAction, submitMode, force) {
            if(el) {
                var forms;
                var viewstackForms = {}; //will hold an array of forms
                if(Array.isArray(el)){
                    forms = el;
                    el = forms[0];
                    if(!submitMode){

                        //When user clicks on the X button and their are multiple forms in a viewstack we need
                        //to save the forms all together. This might not apply to future use cases but as of today
                        //any form in a viewstack will be treated as one.  Any form that fails will cause all forms in the
                        //viewstack to fail.

                        //if no special submitMode then just call save on each one
                        for (var i = 0, len = forms.length; i < len; i++) {
                            var frm = forms[i];
                            var vs = frm.closest('cmviewstack');
                            if(vs){
                                if(!viewstackForms.hasOwnProperty(vs.id)) viewstackForms[vs.id] = []; //instantiate the array
                                viewstackForms[vs.id].push(frm);
                            }else{
                                save(frm, submitAction, submitMode, force);
                            }                                                                                       
                        }

                        //check for viewstack forms and save
                        var vstacks = Object.getOwnPropertyNames(viewstackForms);
                        if(vstacks.length > 0){                            
                            Object.keys(vstacks).forEach(function(key) {
                                // skip this if undefined
                                if(vstacks[key] === undefined) {
                                    return;
                                }                                
                                save(viewstackForms[vstacks[key]],'save', 'viewstack');                                
                            })
                        }
                        return true;
                    }
                }
                var view = Views.findView(el);
                var form = Widget.getContext(el);
                if((Date.now() - form.lastSave) < 1000) {
                    //saving is currently blocked to prevent rapid saving
                } else {
                    //first make sure any active control commits...
                    if(form.el.contains(document.activeElement) && (document.activeElement.tagName !== 'BUTTON')) {
                        //focus should almost always be in a button on submit but for some reason currently is in an input control
                        commitActiveElement(el);
                    }
                    // check to make sure the form mode is saving.
                    // Enter key is being pressed more than once causing the form to save multiple times causing duplicates.
                    if(form.owner && !form.owner.classList.contains('popup') && form.mode === form.submitMode) {
                        //do nothing
                    } else {
                        autofillForm(form.el);
                        //fire an event off via the ds so that any control can commit data changes
                        form.ds.fire('changes-commit');
                        //now check to see whether we need to cancel the save
                        if(form.autoDelete && form.formdata.rowstate === 'IN') {
                            //only used by grid/list at moment....
                            //inserted row that has never been modified
                            itemChanges(form, form.ds, 'end-cancel');
                            //if embeded close otherwise nothing
                            setEditMode(form.el, 'delete');
                            return true;
                        } else {
                            //if force is true then force the row to be dirty....
                            if(force) form.ds.itemUpdated('U', form.formdata);
                            //ok first see if there has been any change to the data that would cause a save....
                            if(isFormDirty(form.el, false, true)) {
                                submitAction = submitAction || form.submitAction;
                                submitMode = submitMode || (submitAction === 'none' ? form.savedMode : form.submitMode);
                                //clear all invalid markers and remove error displays - we will retest and mark later - then call custom validations....
                                var vr = ValidationManager.validateForm(forms || el, submitAction);
                                if(!vr.hasErrors) {
                                    //passed validation
                                    if(submitAction==='none') {
                                        //no submit action - we use 'data' mode and save instead so just return........
                                        itemChanges(form, form.ds, 'end-save');
                                        //if embed then close if on a panel then close if on a page do nothing
                                        setEditMode(form.el, submitMode);
                                        //were done - should push into the committed mode
                                        return true;
                                    }else {
                                        switch (submitMode) {
                                            case 'viewstack':
                                                //assumes it is event driven and action is always save........
                                                Views.fireActionEvent(el, submitAction, false, true, {
                                                    formValues: form.formdata,
                                                    formState: form.state
                                                });
                                                itemChanges(form, form.ds, 'end-submit');
                                                //if embed then close if on a panel then close if on a page do nothing
                                                setEditMode(form.el, submitMode);
                                                //were done - should push into the committed mode
                                                return true;
                                                break;
                                            case 'closeDialog':
                                            case 'closePanel':
                                            case 'closePage':
                                                Views.fireActionEvent(view, submitAction, false, true, {
                                                    formValues: form.formdata,
                                                    formState: form.state,
                                                    forms: forms
                                                });
                                                return true;
                                                break;
                                            default:
                                                //use 'submit' mode -
                                                //in 'submit' mode temporarily block further saves but renable after a short gap
                                                form.lastSave = Date.now();
                                                //ok so the form is in a state to actually be submitted.....
                                                //the module will handle prompting the user then setting Busy and calling the server and will return later with success or failure
                                                //once it returns from the server the form either needs cleaning or it is left in it's current state but marked clean....
                                                //the details object has the formValues and the formState - the save event is ALWAYS fired on the FORM
                                                //publish the event so it gets picked up by the page module - and do the default action
                                                var saving = Views.fireActionEvent(el, submitAction, false, true, { formValues: form.formdata, formState: form.state});
                                                //ok so we have let the module know to save so now what do we do....
                                                if(saving) {
                                                    //we notify that editing is done, set the form as inactive and set the form into the devs chosen mode
                                                    itemChanges(form, form.ds, 'end-submit');
                                                    //we look at submitMode and let that figure it out if no submit mode then close it..
                                                    setEditMode(form.el, submitMode || 'close');
                                                }
                                                return true;
                                                break;
                                        }
                                    }
                                } else {
                                    //failed validation
                                    if(submitAction === 'none') {
                                        // validation failed - we wont save, error messages should be already displayed
                                        // in grid/list embed mode we still close anyway - but panel and page mode stay open
                                        //  1) in list 2) in grid 3) in panel 4) in page
                                        //  1) or 2) call dismiss 3) or 4) do nothing except refocus....
                                        if(form.embedMode === 'grid' || form.embedMode === 'list') {
                                            //really we should end itemChanges I think....
                                            itemChanges(form, form.ds, 'invalid');
                                            //if embed then close if on a panel then close if on a page do nothing
                                            setEditMode(form.el, submitMode);
                                        } else {
                                            //still active, item changes are continuing, just refocus the first invalid field
                                            var ael = el.querySelector('input:invalid');
                                            if(ael) Focus.focusLater(ael);//return focus but doing so could trigger a dropdown
                                        }
                                    } else {
                                        // validation failed - we wont save, error messages should be already displayed
                                        // panel and page stay open
                                        // in panel or in page
                                        // do nothing except refocus....
                                        if(submitMode==='viewstack'){                                            
                                            //need to activate the correct tab..
                                            var tab = vr.form.closest('tab');
                                            ListViews.showTab(Views.findView(tab), tab);                                            
                                        }
                                        
                                        var ael = el.querySelector('input:invalid');
                                        if(ael) Focus.focusLater(ael);//return focus but doing so could trigger a dropdown
                                    }
                                    return false;
                                }
                            } else {
                                // not dirty so commit false and close...... IN will have already been dealt with
                                // set focus back.....
                                // 1) in list 2) in grid 3) in panel 4) in page
                                // 1) or 2) or 3) call dismiss 4) do nothing....
                                if(form.embedMode === 'grid' || form.embedMode === 'list') {
                                    //really we should end itemChanges I think....
                                    itemChanges(form, form.ds, 'cancel');
                                    setEditMode(form.el, submitMode||'close');
                                } else {
                                    // if(ael)ael.focus();//return focus but doing so could trigger a dropdown
                                    // form.ds.commitRowUpdate(form.autoDelete);
                                }
                                return true;
                            }
                        }
                    }
                }
            }
        }
        /**
         * Delete the associated row from the datasource
         * only called from the click event delete handler
         * @param {HTMLFormElement} el
         * @param {boolean} force if true dont check prompt
         * @memberOf FormController#
         */
        function formDelete(el, force) {
            var form = Widget.getContext(el);
            var msg = form.prompts['DELETE'];
            if(msg) {
                Notify.confirm(msg).always(function(result) {
                    if(result === false) {//no
                        Focus.focusLater(Focus.findFirstFocusable(form));
                    } else {
                        formDelete(el, true);
                    }
                });
            } else {
                // if we are embedded in a grid or list then deleting the form also deletes the parent row
                // so we cant just dismiss and then rerender the parent row
                // we need to change the data and force the parent to rebuild.....
                //if we are not embedded then we cant dismiss we need to reload the next row if its an DSArry or load an empty row if a DI
                //stop listening to data events
                itemChanges(form, form.ds, 'delete');
                //what we need to do now - close?
                setEditMode(form.el, (form.embedMode) ? 'close' : 'continue');
            }
        }
        /**
         * Cancel form input
         * Could be a form in a page or panel or popup or embeded in a list or grid
         * Could be updating or inserting
         * Cancel always implies undoing the current changes
         * But does it remove an inserted record and does it close the view it is in?
         * @param {HTMLFormElement|Array} el
         * @param {boolean} [dismiss] if true the form is cancelled and dismissed. Only set in prompted closes.
         * @memberOf FormController#
         */
        function cancel(el, dismiss) {
            if(Array.isArray(el)){
                var forms = el;
                for (var i = 0, len = forms.length; i < len; i++) {
                    var form = forms[i];
                    cancel(form,dismiss);
                }
            }else {
                //first call the module and check whether it is ok to cancel
                if (Views.fireActionEvent(el, 'cancel', false, true, {})) {
                    //this is really cancel and close... we currently don't have a reset
                    setEditMode(el, dismiss ? 'cancel:close' : 'cancel');
                }
            }
        }
        /**
         * Global handler for the Form Input Element 'change' event
         * This event fires when a control loses focus and its value has changed since it got focus (in theory).
         * Except it doesn't fire if the change is programmatic and it for FF/C/S it doesn't fire on blur on checkbox/radio it fires on click instead.
         * This event is executed before the code specified by 'blur' when the control is also losing the focus.
         * So lookup/dropdown/selectlist/date/time should all ignore this event and handle commitValue in the blur instead.
         * text/search/textarea/number/checkbox/radio all use this
         * This is fired by textarea and inputs but not contenteditable
         * @param {Event} e
         */
        function changeHandler(e) {
            if(inputTypes[e.target.tagName]) {
                if(e.target.hasAttribute('debounce')){
                    debounceInputHandler.reset();
                }
                //specials ignore the change event (return false) and process by explicitly calling commit
                //defaults use the change event to update the data...
                if (dispatchInputEvent(e)) {
                    commitValue(e.target);
                }
            }
        }

        //these types ignore the change event and trigger on blur, click or input
        var NoChangeTypes = { file: true, button: true, submit: true, reset: true, menu: true, texteditor: true, article: true, 'text-editor': true, tageditor: true, 'tag-editor': true, selectlist: true, dropdown: true, lookup: true, cmtime: true, time: true, cmdate: true, date: true, checkbox: true, radio: true };
        /**
         * Get the control type
         * @param el
         * @returns {string}
         */
        function getInputType(el) {
            //inputs have to use data-type because we have to force type on some controls back to 'text'
            return el.tagName === 'INPUT' ? el.dataset.type : el.tagName.toLowerCase();
        }


        /**
         * Commit the value in the active element without removing focus
         * Called when we save a form to force the data from the last element to be committed
         * @param {HTMLFormElement} form
         * @param {HTMLElement} el override the active element
         * @param {boolean} willBlur
         * @memberOf FormController#
         */
        function commitActiveElement(form, el, willBlur) {
            //IE11 bug forces check for nullness
            el = el || document.activeElement || document.body;
            if(inputTypes[el.tagName]) {
                if(!form) form = el.form;
                //there are a couple of instances of unbound controls used in some widgets
                if(form && form === el.form) {
                    if(!form.ctx.inactive) {
                        var lv = lastValue;
                        if(NoChangeTypes[getInputType(el)]) {
                            console.log('COMMIT ACTIVE ELEMENT')
                            // specials ignore the change event (return false) and instead trigger there internal commit on the blur event
                            blurHandler({ type: 'blur', target: el, relatedTarget: el, currentTarget: el, force: true, willBlur:true });
                            //if the save fails then this will cause a problem as likely in the blur we destroyed everything
                            //really we need to refocus the control
                        } else {
                            // defaults use the change event to update the data and processing is all done in the main commit
                            // lastValue is the value on focus in
                            if(lastValue !== el.value) {
                                commitValue(el);
                                lv = el.value;
                            }
                        }
                        //since we are not actually removing the focus we need to restore lastValue
                        lastValue = lv;
                    }
                }
            }
        }
        /**
         * Commit a controls current value
         * called from onchange for std controls or explicitly by some special controls.
         * @param {HTMLElement} el
         * @memberOf FormController#
         */
        function commitValue(el) {
            if(inputTypes[el.tagName]) {
                if(el.classList.contains('cm-standalone')||el.classList.contains('ignore')) {
                    //standalones don't commit value
                } else {
                    var field = getFieldName(el);
                    if(field && !el.disabled) {
                        var elForm = el.form;
                        //CF's pull data from a sub data object but flag dirty on the main data
                        var data = getFormData(el);//this is usually the form data row but if this is a CF INPUT it is the CF values object
                        var row = getFormData(elForm);//this is always the data row
                        var form = Widget.getContext(elForm);
                        //set value into the form data
                        var type = getInputType(el);
                        var changed = true;
                        //for checkboxes and radios it is up to the UI to set the range of possible values
                        //and the form handler to update the value with the user action
                        //special handling for multiple selects
                        switch(type) {
                            //the control has no field value
                            case 'file':
                            case 'button':
                            case 'submit':
                            case 'reset':
                            case 'menu':
                                changed = false;
                                break;
                            //because the control writes the value to the data itself when it loses focus
                            case 'texteditor'://the proxied input
                            case 'article'://the rte tag
                            case 'text-editor'://the rte tag
                            // because these controls write the value to the data itself when it loses focus - they write multiple values
                            case 'dropdown':
                            case 'lookup':
                            case 'selectlist':
                            case 'tag-editor':
                            case 'tageditor':
                            //because these controls write the value to the data itself when it loses focus - they write multiple values
                            case 'cmtime':
                            case 'time':
                            case 'cmdate':
                                //changed = false;
                                break;
                            case 'color':
                            case 'cmcolor':
                                var v = (el.dataset.colorMode === 'rgb') ? parseInt(el.value.substr(1), 16) || 0 : el.value;
                                if(data[field] === v) {
                                    changed = false;
                                } else {
                                    data[field] = v;
                                }
                                break;
                            case 'date':
                                if(u.obb(el.dataset.dtm === 'datetime.CF')) { // CustomField date - stored without time in US format
                                    data[field] = el.value;
                                }
                                break;
                            case 'checkbox':
                                //check if we are an array...
                                var els = elForm[el.name];
                                var v = '';
                                //even if flagged as multiple there may only be one, (i.e. list driven attachements)
                                if(el.multiple && el !== els) {//els
                                    //we are an array of checkboxes - so we have to concat the values
                                    for(var i = 0, len = els.length; i < len; i++) {
                                        v += els[i].checked ? ',' + els[i].value : '';
                                    }
                                    v = v.substr(1);
                                } else {
                                    //std single checkbox
                                    v = $.stringToValue(el.checked ? el.value : el.getAttribute('alt'));
                                }
                                if(data[field] === v) {
                                    changed = false;
                                } else {
                                    data[field] = v;
                                }
                                break;
                            case 'radio':
                                //we skip radio that aren't checked-assumption is there is always one that is
                                var v = '';
                                if(el.checked) {
                                    v = $.stringToValue(el.value);
                                    if(data[field] === v) {
                                        changed = false;
                                    } else {
                                        data[field] = v;
                                    }
                                } else changed = false;
                                break;
                            case 'number':
                                var n = parseFloat(el.value);

                                if(isNaN(n)) {
                                    n = null;
                                    el.value = '';
                                } else {
                                    var p = el.dataset.precision || 2;
                                    n = Math.roundTo(n, p);
                                    el.value = n.toFixed(p);
                                }

                                if(el.dataset.displayType && el.dataset.valueType){
                                    n = NumberUtils.convertNumeric(n, el.dataset.valueType, el.dataset.displayType);
                                }

                                if(data[field] === n) {
                                    changed = false;
                                } else {
                                    if(el.dataset.dtm) {
                                        //duration handling - the DTM sets the value into the field
                                        DateTime.commitInputField(el, 'duration', el.getAttribute('data-dtm'), n, data);
                                    } else {
                                        data[field] = n;
                                    }
                                }
                                break;
                            case 'textarea':
                                var v = el.value;
                                v =  (v===undefined || v===null) ? '' : v;
                                if(data[field] === v) {
                                    changed = false;
                                } else {
                                    data[field] = v;
                                }
                                break;
                            default://text/search
                                var v = '';
                                v = $.stringToValue(el.value);
                                if(data[field] === v) {
                                    changed = false;
                                } else {
                                    data[field] = v;
                                }
                                break;
                        }
                        if(changed) {
                            //is the form currently in a saving mode if so put it back into editing mode
                            if(form.mode === form.submitMode) setEditMode(form.el, 'saving:editing');
                            // to get here we are either called explicitly by a 'special' or by a native 'change' from a 'default' control - so it should always represent a real 'change'
                            // checkbox and radios don't fire an 'input' event,
                            // lookup/dropdown/buttons also don't trigger change event's
                            // date+time don't trigger change event's if the value was 'picked' as opposed to typed
                            // fire the field change event
                            form.ds.itemUpdated('U', row);
                            form.ds.fire('field', field);
                        }
                    }
                }
            }
        }
        /**
         * Global handler for the Form Input Element 'input' event
         * Note: this is the DOM Event not jQ Event
         * This fires on all controls except radio and checkboxes on any user driven change including key/cut/paste
         * @param {Event} e
         */
        function inputHandler(e) {
            if(inputTypes[e.target.tagName]) {
                var target = e.target;
                if($.browser.isIE11) e = $.copyMerge(e, { type: 'input' });

                if(e.target.hasAttribute('debounce')){
                    debounceInputHandler(e);
                }else {
                    if (dispatchInputEvent(e)) {
                        if (target.tagName === 'TEXTAREA') {
                            //autosize helper - not needed for rte as it handles this by itself
                            target.nextElementSibling.innerText = (target.value || '') + " ";
                        }
                        //need this because the new UI has different visuals if the control has content....
                        if (target.tagName !== 'ARTICLE') System.setImmediate(toggleContentFlag, e);
                    }
                }
            }
        }
        //on mobile lookup/dropdown are converted to input type button...
        var nos = { button: true, checkbox: true, radio: true, number: true, time: true, date: true, range: true };
        var inputTypes = { INPUT: true, TEXTAREA: true, 'ARTICLE': true, 'TAG-EDITOR': true };
        //cache the last focused input value
        var lastValue;
        var debounceInputHandler = $.debounce(debouncedInputHandler,500);
        function debouncedInputHandler(e){
            if (dispatchInputEvent(e)) {
                var target = e.target;
                if (target.tagName === 'TEXTAREA') {
                    //autosize helper - not needed for rte as it handles this by itself
                    target.nextElementSibling.innerText = (target.value || '') + " ";
                }
                //need this because the new UI has different visuals if the control has content....
                if (target.tagName !== 'ARTICLE') System.setImmediate(toggleContentFlag, e);
            }
        }

        /**
         * Global focus in handler
         * Note on mac buttons do not ever get focus neither can they be tabbed to
         * Note this is a DOM Event not jQEvent
         * @param {Event} e
         */
        function focusHandler(e) {
            //TODO:get rid of the isContentEditable as it forces a style recalc switch to using a class
            if(inputTypes[e.target.tagName]) {
                if(!Focus.focusLock) {
                    var control = e.target.closest('.input-control') || e.target;
                    //flag it as focused do this first so can be overridden in handler
                    control.classList.add('focus');
                    //a form input controls has got focus - what do we want to do
                    //IE11 bug relatedTarget not set on blur/focus but is set on focusin/focusout
                    if ($.browser.isIE11) e = $.copyMerge(e, {type: 'focus'});
                    //can get here programmatic focus, tab focus or mouse focus
                    //if in gridedit and mouse focus and in basic editor we need to check whether we need to 'scrollInToView'
                    if (e.target.form && e.target.form.classList.contains('gridedit')) {
                        DataGrid.editRow(e.target.closest('datagrid'), 'cell', e.target);
                    }

                    dispatchInputEvent(e);

                    if (e.target.classList.contains('selectRequired')) {
                        Focus.selectContent(e.target);
                    }

                    //control processing completed - does it still have focus
                    if (control.classList.contains('focus')) {
                        //reset lastValue
                        lastValue = e.target.value;
                        if (e.target.hasAttribute('debounce')) {
                            debounceInputHandler.reset();
                        }
                    }
                }
            }
            console.log(e.type, e.target, e.relatedTarget)
        }
        /**
         * Global focus out handler
         * Note this is a DOM Event not jQEvent
         * TODO: get rid of number inputs..... to block scrolling
         * @param {Event} e
         */
        function blurHandler(e) {
            if(inputTypes[e.target.tagName]) {
                if(!Focus.focusLock) {
                    var control = e.target.closest('.input-control') || e.target;
                    //remove focus flag do this first so can be overridden in handlers
                    control.classList.remove('focus');
                    //for default controls we do the real processing in the change event....
                    //IE11 bug relatedTarget not set on blur/focus but is set on focusin/focusout
                    if ($.browser.isIE11) e = $.copyMerge(e, {type: 'blur'});

                    dispatchInputEvent(e);

                    if (!control.classList.contains('focus')) {
                        //reset lastValue
                        lastValue = undefined;
                        if (e.target.hasAttribute('debounce')) {
                            debounceInputHandler.reset();
                        }
                    }
                }
            }
        }

        /**
         * Initialise an Input element
         * @param {HTMLElement} el
         * @param {HTMLFormElement} elForm
         * @memberOf FormController#
         */
        function initInputElement(el, elForm) {
            //have to pull from attribute
            var type = getInputType(el);
            if(!elForm) elForm = el.form;
            switch(type) {
                case 'cmtime':
                case 'time':
                    controls.TimePickerController.initInputElement(el, elForm);
                    break;
                case 'cmdate':
                case 'date':
                    controls.DatePickerController.initInputElement(el, elForm);
                    break;
                case 'picture':
                    controls.PictureEditorController.initInputElement(el, elForm);
                    break;
                case 'texteditor':
                    controls.TextEditorController.initInputElement(el, elForm);
                    break;
                case 'search':
                    controls.SearchController.initInputElement(el, elForm);
                    break;
                case 'lookup':
                    controls.LookupController.initInputElement(el, elForm);
                    break;
                case 'dropdown':
                    controls.DropdownController.initInputElement(el, elForm);
                    break;
                case 'cmcolor':
                    controls.ColorPickerController.initInputElement(el, elForm);
                    break;
                case 'selectlist':
                    controls.SelectListController.initInputElement(el, elForm);
                    break;
                case 'checkbox':
                    //we have to check whether this is a single value field or a multi value field
                    var els = elForm[el.name];
                    if(els && els.length) {
                        //there are multiple ctl but are they multi valued
                        var multiple = (els[0].value !== els[1].value);
                        for(var i = 0, len = els.length; i < len; i++) {
                            els[i].multiple = multiple;
                        }
                    }
                    break;
                case 'range':
                    break;
                default:
                    break;
            }
        }
        /**
         * Handler for all Keyboard Input
         * @param {KeyboardEvent} e
         */
        function keyHandler(e) {
            var key = e.keyCode || e.key || e.which;
            if(key === KeyCode.TAB) {
                //let tab key handler process it....
            } else {
                if(inputTypes[e.target.tagName]) {
                    // note keypress is now deprecated and in any case only fired after the key was committed
                    // the order is keydown/keypress/input/keyup - if tabbing the keyup will fire on the new input - keyup fires after blur/focus
                    // keydown is before the character is inserted (charcode is 0), keypress fires after the insertion, keyup fires when released (charcode is 0)
                    // tab, backspace, delete do not generate keypress events - if u prevent default on the keydown there will be no keypress and the char is no inserted but there will be a keyup
                    // if focus moves in keydown then keypress and keup wont fire on the old target, keyup always fires on the new target, if new target is null nothing happens
                    // so if u r debugging and trap keydown - no keypress/keyup or trap keypress no keyup
                    // ie9 doesnt fire the input event on kbd removal/cut
                    // on Android most keys only generate a keycode of '229' - backspace only fires on an empty input - 13 is always reliable
                    switch(e.type) {
                        case 'keydown':
                            dispatchInputEvent(e);
                            break;
                        case 'keyup':
                            // 'page' uses keyup - escape
                            // 'grid' uses keyup - escape - also uses kbd nav and enter to start editor
                            // search uses keys
                            // enter in an input control causes form to submit - textarea/texteditor use as crlf
                            dispatchInputEvent(e);
                            var formEl = e.target.form ? e.target.form : document.forms[e.target.getAttribute('form')];
                            if(formEl) {
                                //now our form level handling....
                                if(e.keyCode === KeyCode.ENTER && e.target.tagName === 'INPUT' && !e.defaultPrevented) {
                                    e.stopImmediatePropagation();
                                    e.preventDefault();
                                    var form = Widget.getContext(formEl);
                                    if (form.embedMode === 'grid' || form.embedMode === 'list') {
                                        save(formEl, '', 'edit-next', true);
                                    }else{
                                        save(formEl, '', '', true);
                                    }
                                }
                                if(e.keyCode === KeyCode.ESCAPE && e.target.tagName === 'INPUT' && !e.defaultPrevented) {
                                    var form = Widget.getContext(formEl);
                                    if(form.embedMode === 'grid' || form.embedMode === 'list') {
                                        e.stopImmediatePropagation();
                                        e.preventDefault();
                                        save(formEl, '', '', true);
                                    }
                                }
                            }
                            break;
                        case 'keypress':
                        default:
                            break;
                    }
                } else if(e.target.tagName === 'DIALOG') {
                    //is this actually needed
                    switch(e.target.id) {
                        case 'timepicker':
                            dispatchInputEvent(e, 'time');
                            break;
                    }
                }
            }
        }
        /**
         * Handler for search event
         * @param {Event} e
         */
        function searchHandler(e) {
            if(e.target.tagName === 'INPUT') {
                dispatchInputEvent(e);
            }
        }

        /**
         * Handler for all range event;
         * @param {Event} e
         */
        function rangeHandler(e){
            if(e.target.type==='range'){
                switch(e.type){
                    case 'range-change-end':
                    case 'range-change-start':{
                        Views.fireActionEvent(e.target,e.type,false,false, {value:+e.target.value, rangeValue:+e.target.dataset.rangeValue});
                        break;}
                    case 'blur':{
                        updateRange(e.target, false);
                        break;}
                    case 'focus':{
                        updateRange(e.target, true);
                    break;}
                    case 'change':{
                        // they can have different controllers
                        // mediaplayer and pictureviewer
                        //IE11 fires change event when dragging the thumb everyone else fires input and change only on end
                        if($.browser.isIE11){
                            updateRange(e.target, true);
                            Views.fireActionEvent(e.target,'range-change',false,false, {value:+e.target.value, rangeValue:+e.target.dataset.rangeValue});
                        }else{
                            updateRange(e.target, false);
                            Views.fireActionEvent(e.target,'range-change',false,false, {value:+e.target.value, rangeValue:+e.target.dataset.rangeValue});
                        }
                        break;}
                    case 'input':{
                        // IE11 doesnt fire this event...
                        // they can have different controllers
                        // mediaplayer and pictureviewer
                        updateRange(e.target, true);
                        Views.fireActionEvent(e.target,'range-change',false,false, {value:+e.target.value, rangeValue:+e.target.dataset.rangeValue});
                        break;}
                    case 'keydown':
                    case 'keyup':
                    default:
                        break;
                }
            }else{
                //number input
                switch(e.type) {
                    case 'keyup':
                    case 'keydown':
                    case 'focus':
                        break;
                    case 'blur':
                    case 'change':
                    case 'input':
                        //from the range number input
                        var range = e.target.parentElement.querySelector('input[type="range"]');
                        var scale = range.dataset.scale;
                        var value = parseFloat(e.target.value);
                        value = ((scale === 'zoom') ? (value / 100) < 1 ? 1 - (1 / (value / 100)) : (value / 100) - 1 : value).toFixed(1);
                        if(+range.value !== +value) {
                            range.value = value;
                            range.dispatchEvent(new CustomEvent('change', {bubbles: true}));
                        }
                    default:
                        break;
                }
            }
        }
        /**
         * Update a range control on mouse/touch move this shouldn't be here!!!
         * (-) (input) (TT) (+) (ce)
         * @param {HTMLElement} el
         * @param {boolean} active
         */
        function updateRange(el, active) {
            var value = el.value;
            var lastValue = el.dataset.lastValue;
            var type = el.dataset.scale;
            var v = +value;
            var elTT = el.nextSibling;
            var elCE = el.parentElement.querySelector('input[type="number"]');
            if(value!==lastValue){
                var displayValue='';
                var rangeValue;
                var rv = 0;
                switch(type){
                    case 'time':{
                        rv = rangeValue = v;
                        var m = Math.floor(v / 60);
                        var s = (v - (m * 60)).toFixed(2);
                        displayValue = (('00' + m).substr(-2)) + ':' + ('00' + s).substr(-5);
                    break;}
                    case 'zoom':{
                        rv = rangeValue = (v < 0 ? (1 / (-v + 1)) : (v + 1));
                        rv = Math.round(rv * 100);//round to int
                        displayValue = formatValue('p0', rv / 100);
                    break;}
                    default:{
                        rv = rangeValue = v;
                        displayValue = formatValue('p', rv);
                    break;}
                }
                el.dataset.rangeValue = rangeValue;
                if(elCE)elCE.value = rv;
                if($.support.browser.isIE11){
                    //to force the UI to update
                    if(elTT)elTT.setAttribute( "data-value", displayValue );
                }else{
                    if(elTT)elTT.dataset.value = displayValue;
                }
                var min = +el.min;
                var max = +el.max;
                var pct = (v - min) / (max - min);
                var xPos = el.offsetLeft + (el.offsetWidth * pct);
                if(elTT)elTT.style.left = xPos +'px';
            }
            el.dataset.lastValue = active ? value : '';
        }
        /**
         * Handler for some form of input event, toggle the flag to indicate whether the input has content or not
         * We need to track this because the new UI Design has style difference whether there is content or not in a control
         * @param {Event} e
         */
        function toggleContentFlag(e) {
            //we have to do this because the new UI design has some variance dependent on whether an Input has content or not
            // so we need to be able to indicate to the css that the control hascontent....
            e.target.classList.toggle("hascontent", e.target.value);
        }

        /**
         * Update a form controls value - if an input is standalone it can never get here as there is no form
         * @param {(HTMLInputElement|HTMLTextAreaElement)} el
         * @param {*} data the row data object -  the control 'value' is in a 'field' in the data
         * @param {boolean} force force the control to update even if it has focus (from setFormData)
         */
        function setFieldValue(el, data, force) {
            var field = el.name;
            //we never update the active control
            if(field && ((document.activeElement !== el) || force)) {
                var type = getInputType(el);
                var value = data[field];
                var pel = el.closest('.input-control')||el;
                //set the empty flag -
                //seems to be used in read only mode to hide label + value on '', unchecked or 0 - not documented
                var isEmpty = !value;
                //for checkboxes and radios it is up to the UI to set the range of possible values
                //and the form handler to update the value with the user action
                //if its the active element it ignores this....
                switch(type) {
                    case 'fieldset':
                    case 'file':
                    case 'button':
                    case 'submit':
                    case 'reset':
                    case 'menu':
                        break;
                    case 'lookup':
                        controls.LookupController.setFieldValue(el, data);
                        break;
                    case 'dropdown':
                        controls.DropdownController.setFieldValue(el, data);
                        break;
                    case 'picture':
                        controls.PictureEditorController.setFieldValue(el, data);
                        break;
                    case 'article':
                    case 'text-editor':
                    case 'texteditor':
                        controls.TextEditorController.setFieldValue(el, data);
                        break;
                    case 'tag-editor':
                    case 'tageditor':
                    case 'selectlist':
                        SelectList.setFieldValue(el, data);
                        break;
                    case 'cmtime':
                    case 'time':
                        TimePicker.setFieldValue(el, data);
                        break;
                    case 'cmdate':
                    case 'date':
                        DatePicker.setFieldValue(el, data);
                        break;
                    case 'cmcolor':
                    case 'color':
                        ColorPicker.setFieldValue(el, data);
                        break;
                    case 'checkbox':
                        var checked;
                        //are we in an array....
                        if (el.multiple) {
                            //making sure the indexof includes the whole value and not partial.  For example ANABCBS as also checking ABCBS.
                            //Appending the first ',' cause the indexof to look at the whole value not find a partial match.
                            checked = ((',' + value + ',').indexOf(',' + el.value + ',') !== -1);
                        } else {
                            //set checked if value matches value - most checkboxes use 0/1
                            checked = ($.stringToValue(el.value) === $.stringToValue(value));
                        }
                        if(el.checked !== checked) {
                            el.checked = checked;
                            pel.classList.toggle('checked', checked);
                        }
                        break;
                    case 'radio':
                        //set checked if value matches value - most checkboxes use 0/1
                        var checked = ($.stringToValue(el.value) === ($.stringToValue(value)) || 0);
                        if(el.checked !== checked) {
                            el.checked = checked;
                            pel.classList.toggle('checked', el.checked);
                        }
                        break;
                    case 'textarea':
                        //so that 0 doesnt get ignored
                        value = (value===undefined || value===null) ? '' : value;
                        if(el.value !== value) {
                            el.value = value;
                            //drives the autosizing
                            el.nextElementSibling.innerText = value + " ";
                        }
                        break;
                    case 'number':
                        var n = +value;
                        if(el.dataset.displayType && el.dataset.valueType){
                            n = NumberUtils.convertNumeric(n, el.dataset.displayType, el.dataset.valueType);
                        }
                        var fv = '';
                        if(isNaN(n) || (!value && value !== 0)) {
                            fv = '';
                        } else if(n === 0 && el.dataset.clearZero) {
                            fv = '';
                            isEmpty = false;
                        } else {
                            var p = el.dataset.precision || 2;
                            fv = Math.roundTo(n, p).toFixed(p);
                        }
                        if(fv !== el.value) el.value = fv;
                        break;
                    default:
                        //set default and convert to string before setting into control....
                        value = (value === undefined || value === null) ? '' : '' + value;
                        if(el.value !== value) el.value = value;
                        break;
                }
                //set empty flag if it changed
                var wasEmpty = pel.classList.contains('empty');
                if(isEmpty && !wasEmpty) {
                    pel.classList.add('empty');
                } else if(!isEmpty && wasEmpty) {
                    pel.classList.remove('empty');
                }
            }
        }

        /**
         * Forward events to the subclassed control
         * input, change, focus, blur, keydown, keyup
         * @param {Event} e
         * @param {string} type if set overrides the event target type
         * @returns {boolean} returns false to prevent Forms Controller default processing
         */
        function dispatchInputEvent(e, type) {
            var result = true;
            //must read the attribute not the property as these are really invalid values
            switch(type || e.target.dataset.customType || getInputType(e.target)) {
                case 'cmtime':
                case 'time':
                    result = controls.TimePickerController.onInputEvent(e);
                    break;
                case 'cmdate':
                case 'date':
                    result = controls.DatePickerController.onInputEvent(e);
                    break;
                case 'lookup':
                    result = controls.LookupController.onInputEvent(e);
                    break;
                case 'dropdown':
                    result = controls.DropdownController.onInputEvent(e);
                    break;
                case 'tag-editor':
                case 'tageditor':
                case 'selectlist':
                    result = controls.SelectListController.onInputEvent(e);
                    break;
                case 'search':
                    result = controls.SearchController.onInputEvent(e);
                    break;
                case 'color':
                case 'cmcolor':
                    result = controls.ColorPickerController.onInputEvent(e);
                    break;
                case 'picture':
                    result = controls.PictureEditorController.onInputEvent(e);
                    break;
                case 'article':
                case 'text-editor':
                case 'texteditor':
                    result = controls.TextEditorController.onInputEvent(e);
                    break;
                case 'range':
                    result = rangeHandler(e);
                    //if(result===undefined) result = customEventDefaults[e.type];
                    break;
                case 'custom':
                    // custom controls are generally not bound to a form and are standalone
                    // intended to form part of some larger custom component
                    // they will contain an action attribute to call a controller or module
                    // they won't auto commit
                    result = Views.fireActionEvent(e.target, e.type, false, false);
                    //if(result===undefined) result = customEventDefaults[e.type];
                    break;
                default:
                    return true;
                    break;
            }
            if(result === false){
                if($.browser.isIE11){
                    e.defaultPrevented = true
                }else{
                    e.preventDefault();
                }
            }
            //to reject requires explicit false or call preventDefault
            return !(result === false || e.defaultPrevented);
        }
        //var customEventDefaults = {input:false, change:false, focus:true, blur:true, keyup:true, keydown:true};
        /**
         * Handle the window unload event
         * @param {Event} e
         */
        function unloadHandler(e) {
            destroy();
        }

        /**
         * Prepare to be removed
         */
        function destroy() {
            //listen for element creation events
            Views.notify.remove(viewsNotifyHandler);
            //form events
            document.removeEventListener("reset", resetHandler);
            document.removeEventListener('submit', submitHandler);
            //control events
            document.body.removeEventListener("change", changeHandler);
            document.body.removeEventListener(($.browser.isIE11) ? 'textinput' : 'input', inputHandler);
            document.body.removeEventListener(($.browser.isIE11) ? "focusin" : 'focus', focusHandler, true);
            document.body.removeEventListener(($.browser.isIE11) ? "focusout" : 'blur', blurHandler, true);
            document.body.removeEventListener('keydown', keyHandler);
            document.body.removeEventListener('keyup', keyHandler);
            window.removeEventListener('unload', unloadHandler);
            console.log('FORMS DONE')
        }

        /**
         * find the target from for a given element
         * @param {HTMLElement} el
         * @param {HTMLElement} view
         * @returns {HTMLFormElement}
         * @memberOf FormController#
         */
        function getTargetForm(el, view) {
            if(!view) view = Views.findView(el);
            //find the target form - either it has a form prop  or it has an attribute or data attribute or get the first form in the view
            //TODO: fix the attr form it will point at the live id not the old ref....
            return el.form || view.view.forms[(el.dataset.form || el.getAttribute('form'))] || (inputTypes[document.activeElement] ? (document.activeElement.form || document.activeElement.closest('form')) : view.querySelector('form')) || view.querySelector('form');

        }

        /**
         *
         * @param {string} fieldName
         * @param {HTMLFormElement} elForm
         * @param {boolean} [focus]
         * @returns HTMLElement
         * @memberOf FormController#
         */
        function findControl(fieldName, elForm, focus){
            var el = elForm.querySelector('[name="'+fieldName+'"]') || elForm[fieldName];

            if(el && focus)el.focus();

            return el;
        }
        /**
         * Block mousewheel number scrolling
         * Need to attach low down directly to a Form in order not to impact scrolling in grid/list controls
         * @param {Event} e
         */
        function mouseWheelHandler(e){
            // Disable mousewheel scrolling on number inputs only when focused (so can still scroll the parent)
            if(e.target.type==='number' && e.target === document.activeElement){
                e.preventDefault();
            }
        }


        /**
         * Init the controller
         * IE11 doesnt set the related target in focus/blur events os we use focusin and focusout
         * IE11 treats a placeholder like real text and fires an input event on focus/blur when the placeholder is
         * changed also in no input fires on contenteditable so we use textinput instead
         *
         */
        function init() {
            //listen for element creation events
            Views.notify.add(viewsNotifyHandler);
            //form events
            document.addEventListener('reset', resetHandler);
            document.addEventListener('submit', submitHandler);
            //control events
            document.body.addEventListener('change', changeHandler);
            document.body.addEventListener(($.browser.isIE11) ? 'textinput' : 'input', inputHandler);
            document.body.addEventListener(($.browser.isIE11) ? 'focusin' : 'focus', focusHandler, true);
            document.body.addEventListener(($.browser.isIE11) ? 'focusout' : 'blur', blurHandler, true);
            document.body.addEventListener('keydown', keyHandler);
            document.body.addEventListener('keyup', keyHandler);

            window.addEventListener('unload', unloadHandler);
        }

        /**
         *
         * @param {Object} msg
         * @memberOf FormController#
         */
        function onMessage(msg){
            switch(msg.type){
                case 'range-change-end':
                case 'range-change-start':{
                    rangeHandler(msg);
                break;}
            }

        }
        //PUBLIC
        self.initInputElement = initInputElement;

        self.setFormState = setFormState;
        self.getFormState = getFormState;
        self.setDisplayGroupState = setDisplayGroupState;

        self.getFormData = getFormData;
        self.setFormData = setFormData;
        self.isFormDirty = isFormDirty;
        self.findDirtyForm = findDirtyForm;
        self.findDirtyGrid = findDirtyGrid;
        self.setEditMode = setEditMode;
        self.setDisabled = setDisabled;
        self.save = save;
        self.saveAll = saveAll;
        self.cancel = cancel;
        self.formDelete = formDelete;
        self.commitValue = commitValue;
        self.commitActiveElement = commitActiveElement;
        self.autocloseHandler = autocloseHandler;
        self.getTargetForm = getTargetForm;
        self.updateFieldValues = updateFieldValues;
        self.updateControlState = updateControlState;
        self.getFieldName = getFieldName;
        self.findControl = findControl;
        self.onMessage = onMessage;
        init();
        return self;
    };
    //shortcut
    /**@type FormController */
    window.Forms = new FormController();
    controls.FormController = window.Forms;
        //register defaults
    /**
     * @typedef {Object} FormOptions
     * @property {string}   source The DataSource/DataItem the form is bound to.
     *
     * @property {string}   initMode The initial mode the form is set to after initialized.
     *
     * @property {boolean}  autoEdit switch to active edit mode if ds selects a none empty and back to inactive on a save.
     * @property {string}   autoClose The mode for implicit closes (by clicking the close button) dirty/dirtyPrompt/none.
     * @property {boolean}  autoDelete inserted but clean rows are deleted upon save.
     *
     * @property {boolean}  autoSave If true will automatically save on ds navigation.
     * @property {boolean}  autoPrompt If true will prompt on autoSave.
     * @property {string}   autoSubmitMode The mode during auto save.
     *
     * @property {string}   submitAction The action topic for the save event.
     * @property {string}   submitMode The mode during save.
     *
     * @property {string}   savedMode mode on save complete.
     * @property {string}   cancelMode mode on cancel complete.
     *
     * @property {string}   dismiss Dismiss mode. what to dismiss - a page, panel, list or grid or auto. auto means let the system figure it out - which means list/grid inside list or grid, panel inside panel and none inisde a page. Saved, Cancel, or Close will call dismiss
     *
     * @property {string}   fileSource The DataSource for file upload
     * @property {string}   state current form 'state'.
     * @property {string}   fieldErrors where to display field errors none notify field.
     * @property {string}   formErrors where to display form errors none notify panel last custom.
     * @property {boolean}  waiting if true saving is currently blocked
     */
    Widget.setDefaultOptions("form", { source: '', initMode: 'editing', autoEdit: false, autoDelete: false, submitAction: 'save', submitMode: 'saving', savedMode: 'close', cancelMode: 'close', dismiss: 'auto', autoClose: 'dirtyPrompt', autoSave: false, autoPrompt: '', autoSubmitMode: '', fileSource: '', filePath: '', state: '', formErrors: 'panel', fieldErrors: 'field', formErrorEntity: 'row', inactive: true, waiting: false });
    Widget.setDefaultOptions("form.search", { formErrors: 'none', autoClose: 'dirty', submitMode: 'searching' });
    Widget.setDefaultOptions("form.listedit", { autoDelete: true, submitAction: 'none', submitMode: '', savedMode: 'close', dismiss: 'listitem', autoSave: true, autoPrompt: false, autoSubmitMode: 'auto' });
    Widget.setDefaultOptions("form.gridedit", { autoDelete: true, submitAction: 'none', submitMode: '', savedMode: 'close', dismiss: 'griditem', autoSave: true, autoPrompt: false, autoSubmitMode: 'auto' });
    //endregion


})(jQuery);