(function ($) {
    "use strict";
    //region Imports

    //endregion
//region Structures

    /**
     * @typedef {Object} VirtualTarget
     * @property {boolean} trackingGesture Whether an implicit touch gesture is currently being tracked.
     * @property {boolean} trackingPointer Whether a pointer is currently being tracked.
     * @property {boolean} trackingClick Whether a click is currently being tracked.
     * @property {boolean} trackingDrag Whether a drag action is currently being tracked.
     * @property {boolean} isDown whether touch or pointer is down
     * @property {number} lastClickTime Timestamp for when when last click occurred.
     * @property {HTMLElement} target The element being tracked for a click.
     * @property {HTMLElement} currentTarget.
     * @property {number} touchID1 ID of the first touch, retrieved from Touch.identifier.
     * @property {number} touchID2 ID of the last touch, retrieved from Touch.identifier.
     * @property {Function} gestureHandler A custom gesture handler.
     * @property {Object} gestureContext A custom context object for gestures.
     * @property {Object} gestureConfig A custom context object for gestures.
     * @property {string} pointerType The type of event we are tracking.
     * @property {Object} p1 the first touch we are tracking.
     * @property {Object} p2 the last touch we are tracking.
     * @property {DragConfig} dragConfig the drag configuartion object
     * @property {boolean} draggable whether the tracked target is currently being dragged
     * @property {boolean} forceGesture if true the next touch or mousedown will be forced to a gesture
     * @property {boolean} pinching whether the current target is being pinched gesture related
     * @property {number} scale the current scale gesture related
     * @property {number} lastScale the previous scale gesture related
     * @property {Point} midpoint the midpoint between p1 and p2 gesture related
     * @property {number} distance the current distance between p1 and p2 gesture related
     * @property {number} initialDistance the distance between p1 and p2 when the gesture started gesture related
     * @property {number} xInc the distance that p2 has moved in the x plane since the last event
     * @property {number} yInc the distance that p2 has moved in the y plane since the last event
     * @property {number} xOffset the initial x offset of the mouse point from the target
     * @property {number} yOffset the initial y offset of the mouse point from the target
     * @property {string} mode touch or pointer
     * @property {boolean} waiting we are waiting on a timer
     * @property {boolean} fastClickMode if true we are in a fast click mode
     * @property {boolean} active is this vt currently active
     * @property {HTMLElement} activeElement
     */

    /**
     * @typedef {Object} DragConfig
     * @property {string} dragMode the preset 'type' of a DragConfig drag configuration object
     * @property {string} location the valid drop locations relative to the drop target - combination of 'before' 'on' 'after' .
     * @property {string} allowed the allowed types of drop operation 'none'/'copy'/'move'/'link'.
     * @property {string} performed the actual type of drop operation to be performed 'none'/'copy'/'move'/'link'.
     * @property {Array} dropzones an array of id's or HTMLElement's which contain dropTargets, these are usually UL or TABLE elements
     * @property {HTMLElement} dragContainer The root container of the draggable element usually a UL or TABLE.
     * @property {HTMLElement} dragTarget The draggable target HTMLElement usually a LI or TR.
     * @property {HTMLElement} dropContainer The root container of the drop target usually a UL or TABLE element
     * @property {HTMLElement} dropTarget The drop target HTMLElement usually a LI or TR.
     * @property {Model} model The model containing the dropSource and dragSource.
     * @property {DataSource} dragSource the DataSource.
     * @property {string|number} dragKey the key of the item being dragged.
     * @property {Object} dragItem the item being dragged.
     * @property {DataSource} dropSource
     * @property {string|number} dropKey
     * @property {Object} dropItem
     * @property {DataTransfer} dataTransfer The DT object.
     * @property {string} dropLocation the actual location relative to the dropTarget where the drop occurred
     * @property {string} action the current action being performed (dragstart/enter/over/leave/drop/end)
     * @property {Object} dropItemTypes associative array of dropTarget tagNames indexed by dropContainer tagName
     * @property {boolean} modifyDOM if true the DOM is updated on the drop, if false it is assumed the data change will propagate a DOM change
     * @property {string} showDragImage none target default custom
     * @property {HTMLElement} dragImage
     * @property {number} offsetX
     * @property {number} offsetY
     * @property {boolean} showDropMarker if true highlight dropzone
     * @property {function} getLocation function return valid drop locations
     * @property {function} canDrop function return if a potential dropTarget is valid
     * @property {Object} options dev set context options
     * @Property {boolean} dragMoveOnAppend if true a drag move 'on' will appendChild  else prependChild
     */

//endregion

//region MouseController
    /**
     * VMouseController
     * Abstracts touch and mouse pointer events
     * Gestures:
     * One Finger:Pan:finger down move up - no events until end then onscroll
     * One Finger:Hold:finger down down - no mouse events but on Android Hold fires mousedown after 650ms
     * One Finger:DTap(Zoom):finger down up down up - no mouse events - but not supported on Android and WP
     * One Finger:Tap:finger down up - mouse fires after touch is complete - mouse over move down up click!!!
     * Two Fingers:Pinch(Zoom)
     * Two Fingers:Pan: fingers down move (mousewheel) up (onscroll)
     *
     * Can use prevent default to stop scroll
     *
     * FastClick Support (Touch only prevent the 300ms wait for taphold)
     * Should only be supported in a Touch enviroment
     * <meta name="viewport" content="width=device-width, user-scalable=no">
     * plus css on body
     * -ms-touch-action: manipulation;
     * touch-action: manipulation;
     *
     * to activate gestures either the target must have the class gesture
     * or we wait for 'tap hold' before querying the element
     * we optimize for any BUTTON which isn't flagged for a tap-hold on touch devices and fire the internal 'click' on touchstart
     * we similarly optimize for all mouse devices and trigger click on the mousedown
     *
     * iOS Clickable
     * on iOS an element must be 'clickable' to receive click events which is defined as:
     * The target element of the event is 'typically clickable' that is either
     * a button ('button'), a link ('a'), a form field ('input','textarea','label') or image ('img', 'area').
     * If not 'typically clickable' then the following rules apply:
     * The target element, or any of its ancestors up to but not including the <body>, has an explicit event handler set for any of the mouse events.
     * This event handler may be an empty function or onclick="void(0)".
     * The target element, or any of its ancestors up to and including the document has a cursor: pointer CSS declarations.
     * NB if you override the css in children you might make the element non-clickable
     *
     * Button Click Focus
     * Windows: Yes - FF, Chrome, IE11, Edge
     * Mac: Yes - Chrome No - Safari, FF
     * iOS: No - Safari, Chrome
     * Android: Yes - Chrome
     *
     * To block focus moving to button, fire click in mousedown/touchstart and then preventDefault
     * iOS has a problem in firing click on the wrong element if touch inadvertently moves between start and end
     * particularly on scrolling it will fire click on the start element not the stop element
     * iOS has a problem requiring double-tap to fire click if the element has a hover effect
     * Fix by firing click on start and preventDefault
     * On Mac even though focus is not set on button it will remove it from current target
     * normal order
     * down/focus/up/click
     * preventDefault on down
     * down/up/click
     *
     * Hover DoubleClick
     * iOS has a problem if an element has a hover effect it requires a double click to fire the click event
     *
     * TrustedClick
     * iOS the TouchStart event is not 'isTrusted' only click is.
     * So events that require 'isTrusted' (fileOpen) must use click
     *
     *
     *
     * NB: event.timeStamp cannot be reliably used at the moment.
     * Chrome shipped a breaking change making the timeStamp an incompatible DOMHiResTimeStamp,
     * FireFox are giving up and leaving the value as 0
     * Apple will use event.time instead for the DOMHiResTimeStamp version
     * So we will for now only use the Date.now
     *
     * Event order
     * Although the specific ordering of touch and mouse events is implementation-defined,
     * the standard indicates the following order is typical: for single input:
     *
     * touchstart
     * Zero or more touchmove events, depending on movement of the finger(s)
     * touchend
     * mousemove
     * mousedown
     * mouseup
     * click
     * If the touchstart, touchmove or touchend event is canceled during an interaction,
     * no mouse or click events will be fired, and the resulting sequence of events would just be:
     *
     * touchstart
     * Zero or more touchmove events, depending on movement of the finger(s)
     * touchend
     */
    /**
     * We are essentially tracking clicks or gestures
     * See Handlers section
     * @returns {VMouseController}
     * @constructor
     */
    var VMouseController = function () {
        var self = this;

        var eventTypes = [
            'dragstart',
            'dragend',
            'drag',
            'dragenter',
            'dragover',
            'dragleave',
            'drop',
            'resizestart',
            'resizemove',
            'resizeend',
            'gesturestart',
            'gesturemove',
            'gestureend',
            'tap',
            'doubletap'
        ];
        var mouseEventTypes = ['click', 'dblclick', 'mousedown', 'mouseenter', 'mouseleave', 'mousemove', 'mouseout', 'mouseover', 'mouseup'];

        var POINTER_TYPE_TOUCH = "touch";//coarse
        var POINTER_TYPE_PEN = "pen";//fine
        var POINTER_TYPE_MOUSE = "mouse";//fine
        var actions = {
            drag: {cursor:'move', moveListener: 'dragMove'},
            resizex: {cursor:'e-resize',moveListener: 'resizeMove'},
            resizey: {cursor:'s-resize',moveListener: 'resizeMove' },
            resizexy: {cursor:'se-resize',moveListener: 'resizeMove'},
            gesture: {cursor:'',moveListener: 'gestureMove'}
        };
        var holdTimer;
        /**@var {number} - move touch pointer more than this and a click will be cancelled.*/
        var touchBoundary = 20;
        /**@var {number} - move mouse pointer more than this and a click will be cancelled.*/
        var mouseBoundary = 10;
        /**@var {number} - Swipe displacement must be more than this.*/
        var swipeDistanceThreshold = 20;//75;
        /**@var {number} - Block event that occur within this time, helper for ghost clicks on iOS*/
        var fastClickThreshold = 300;

        /**@type VirtualTarget */
        var vtRoot;
        var isMac = false;
        /**timestamp of last user touch/mouse activity*/
        var lastEvent = 0;

        /**
         * The internal context class for a virtual mouse target
         * @returns {VirtualTarget}
         */
        function CreateVirtualTarget() {
            return {
                trackingClick: false,
                trackingPointer: false,
                trackingGesture: false,
                trackingDrag: false,
                target: null,
                currentTarget: null,
                pointerType: $.support.touch ? 'touch' : 'mouse',
                p1: {x: 0, y: 0, timeStamp: 0},
                p2: {x: 0, y: 0, timeStamp: 0}
            }
        }
        /**
         * The internal context class for a drag source
         * @returns {DragConfig}
         */
        function CreateDragConfig() {
            return {
                location: 'before-after',
                allowed: 'move',
                performed: 'move',
                dropItemTypes: {DATAGRID: 'grid-section > grid-row.main', TABLE: 'TR', TASKBOARD: 'TASKCARD,TASKBIN', LISTVIEW: 'LI,UL', UL:'LI,UL'},
                dragItemTypes: {DATAGRID: 'grid-section > grid-row.main', TABLE: 'TR', TASKBOARD: 'TASKCARD,TASKBIN', LISTVIEW: 'LI,UL', UL:'LI,UL'},
                canDrop:checkCanDrop
            };
        }


//region FastClick

        //region Setup
        /**
         * Handler for all VMouse events
         * Note starting with Chrome 49 and Firefox 54,
         * Event.timeStamp is a high-resolution monotonic time (DOMHighResTimeStamp) instead of epoch time (DOMTimeStamp).
         * @param {Event} e
         */
        function handleEvent(e){
            var now = Date.now();
            try {
                //tracking inactivity
                if (window.System) window.System.data.lastTick = now;
                //tracking the control key
                Key.setControlKey(isMac ? e.metaKey : e.ctrlKey);
                //ignore events for text nodes, we only process elements and left button actions
                if(e.target.nodeType===1 && !e.button) {
                    //this part is for only catching events bubbling to the root
                    if (e.currentTarget === document || e.currentTarget === document.body || e.currentTarget === window) {
                        if (captureElement) {
                            //capturing events from outisde the browser
                            handleCapturedEvent(e);
                        } else {
                            switch (e.type) {
                                case 'mousedown':
                                    //if touch active and more than 300ms then hybrid switching to pointer mouse......
                                    if(Device.status.touchActive){
                                        if((now - lastEvent) > 400){
                                            console.info('VMOUSE: SWITCH TOUCH TO MOUSE '+now)
                                            //allow switch....
                                            vtRoot.pointerType = Device.setActivePointer(POINTER_TYPE_MOUSE);
                                            vtRoot.isDown = true;
                                            onMouse(e);
                                        }else{
                                            // synthetic mouse after previous touch
                                            // anything that fires less than 300ms after last touch event is considered synthetic
                                            // ignore touch based synthetic mouse events
                                            // should cancel the event - but can't do this because of WalkMe software
                                        }
                                    }else{
                                        vtRoot.pointerType = Device.setActivePointer(POINTER_TYPE_MOUSE);
                                        vtRoot.isDown = true;
                                        onMouse(e);
                                    }
                                    break;
                                case 'mousemove':
                                case 'mouseup':
                                    if(Device.status.touchActive) {
                                        // synthetic mouse after previous touch
                                        // anything that fires less than 300ms after last touch event is considered synthetic
                                        // ignore touch based synthetic mouse events
                                        // should cancel the event - but can't do this because of WalkMe software
                                    }else{
                                        onMouse(e);
                                    }
                                    break;
                                case 'click':
                                    //click happens after mouse or touch - we stick with whatever type it was
                                    onClick(e);
                                    break;
                                case 'touchstart':
                                    lastEvent = now;
                                    vtRoot.pointerType = Device.setActivePointer(POINTER_TYPE_TOUCH);
                                    vtRoot.isDown = true;
                                    onTouchStart(e);
                                    break;
                                case 'touchmove':
                                    lastEvent = now;
                                    vtRoot.pointerType = Device.setActivePointer(POINTER_TYPE_TOUCH);
                                    onTouchMove(e);
                                    break;
                                case 'touchend':
                                    lastEvent = now;
                                    vtRoot.pointerType = Device.setActivePointer(POINTER_TYPE_TOUCH);
                                    onTouchEnd(e);
                                    break;
                                case 'touchcancel':
                                    lastEvent = now;
                                    vtRoot.pointerType = Device.setActivePointer(POINTER_TYPE_TOUCH);
                                    onTouchCancel(e);
                                    break;
                                case 'dragstart':
                                    vtRoot.pointerType = Device.setActivePointer(POINTER_TYPE_MOUSE);
                                    vtRoot.p1.pageX = e.pageX;
                                    vtRoot.p1.pageY = e.pageY;
                                    vtRoot.p1.timeStamp = now;
                                case 'dragend':
                                case 'drag':
                                case 'dragenter':
                                case 'dragover':
                                case 'dragleave':
                                case 'drop':
                                    vtRoot.p2.timeStamp = now;
                                    vtRoot.p2.pageX = e.pageX;
                                    vtRoot.p2.pageY = e.pageY;
                                    onMouse(e);
                                    break;
                                case 'wheel':
                                    mouseWheelHandler(e);
                                    break;
                                case 'unload':
                                    unloadHandler(e);
                                    break;
                                case 'pointerdown':
                                    if((e.target.tagName==='INPUT') && e.target.type==='range'){
                                        Forms.onMessage({type:'range-change-start', target:e.target});
                                    }
                                    break
                                case 'pointerup':
                                    if((e.target.tagName==='INPUT') && e.target.type==='range'){
                                        Forms.onMessage({type:'range-change-end', target:e.target});
                                    }
                                    break;
                                default:
                                    break;
                            }
                        }
                    } else {
                        //event captured for a lower level element
                    }
                }
            } catch (ex) {
                Loggr.error(ex);
            }
        }
        //event options
        var optPassive = $.support.passiveeventlisteners ? {passive: true, capture: false} : false;
        var optPassiveCapture = $.support.passiveeventlisteners ? {passive: true, capture: true} : false;
        var optNonPassive = $.support.passiveeventlisteners ? {passive: false, capture: false} : false;
        var optNonPassiveCapture = $.support.passiveeventlisteners ? {passive: false, capture: true} : true;

        /**
         *
         * @param {boolean} passive
         * @param {boolean} capture
         * @returns {*}
         * @memberof VMouseController#
         */
        function getListenerOptions(passive, capture){
            if(passive){
                if(capture) {
                    return optPassiveCapture;
                }else{
                    return optPassive;
                }
            }else{
                if(capture) {
                    return optNonPassiveCapture;
                }else{
                    return optNonPassive;
                }
            }
        }
        /**
         * Attach all event handlers
         */
        function attachHandlers(){
            var vt = vtRoot = CreateVirtualTarget();
            // Mouse and Touch - always both as we now have hybrids
            document.body.addEventListener('mousedown', self, optNonPassive);
            document.body.addEventListener('mousemove', self, optNonPassive);
            document.body.addEventListener('mouseup', self, optNonPassive);

            //its ok to have non-passive here
            document.body.addEventListener('click', self, optNonPassiveCapture);

            if($.support.touch) {
                //must be passive
                document.body.addEventListener('touchstart', self, optPassive);
                document.body.addEventListener('touchmove', self, optPassive);
                document.body.addEventListener('touchend', self, optPassive);
                document.body.addEventListener('touchcancel', self, optPassive);
            }

            document.addEventListener('dragstart', self, optNonPassive);
            document.addEventListener('dragend', self, optNonPassive);//converted to dragcancel
            document.addEventListener('drag', self, optNonPassive);
            document.addEventListener('dragenter', self, optNonPassive);
            document.addEventListener('dragover', self, optNonPassive);
            document.addEventListener('dragleave', self, optNonPassive);
            document.addEventListener('drop', self, optNonPassive);//split into drop and dropcomplete

            //track mouse wheel events - force popups to close and disable hovers
            window.addEventListener('wheel', self, optPassive);
            //track the unload so we can cleanup for DevTools
            window.addEventListener('unload', self, optNonPassive);

            document.body.addEventListener('pointerdown', self, optPassive);
            document.body.addEventListener('pointerup', self, optPassive);
        }

        /**
         * remove all event handlers
         */
        function detachHandlers(){
            document.body.removeEventListener('mousedown', self, false);
            document.body.removeEventListener('mousemove', self, false);
            document.body.removeEventListener('mouseup', self, false);

            document.body.removeEventListener('click', self, true);

            document.body.removeEventListener('touchstart', self, false);
            document.body.removeEventListener('touchmove', self, false);
            document.body.removeEventListener('touchend', self, false);
            document.body.removeEventListener('touchcancel', self, false);

            document.removeEventListener('dragstart',self);
            document.removeEventListener('dragend',self);//converted to dragcancel
            document.removeEventListener('drag',self);
            document.removeEventListener('dragenter',self);
            document.removeEventListener('dragover',self);
            document.removeEventListener('dragleave',self);
            document.removeEventListener('drop', self);//split into drop and dropcomplete

            window.removeEventListener('wheel',self);

            window.removeEventListener('unload',self);
        }

        //endregion

        //region Utilities
        var captureElement;
        var firingCapturedEvent;
        /**
         * Activate or deactivate mouse event global capturing
         * when activated capture events from out the browser
         * @param {HTMLElment} element
         * @param {boolean} state
         * @memberof VMouseController#
         */
        function setCapture(element, state){
            if(state){
                for (var i = 0; i < mouseEventTypes.length; i++)
                    window.addEventListener(mouseEventTypes[i], self, true);
            }else{
                for (var i = 0; i < mouseEventTypes.length; i++)
                    window.removeEventListener(mouseEventTypes[i], self, true);
            }
            captureElement = state ? element :null;
        }

        /**
         * process a global captured event, retarget it to the capturing elment
         * @param event
         */
        function handleCapturedEvent(event) {
            if(!firingCapturedEvent) {
                try {
                    firingCapturedEvent = true;
                    captureElement.dispatchEvent(cloneMouseEvent(event));
                } finally {
                    firingCapturedEvent = false;
                    event.stopImmediatePropagation();
                }
            }
        }
        /**
         * Not intended to be used with subclasses of MouseEvent
         * Clones a global mouse event for retargeting it
         * @param {MouseEvent} event
         * @returns {MouseEvent}
         */
        function cloneMouseEvent(event) {
            var eventToDispatch;
            if (document.createEvent) {
                eventToDispatch = document.createEvent('MouseEvent');
                eventToDispatch.initMouseEvent(event.type,
                    event.false,
                    event.cancelable,
                    event.view,
                    event.detail,
                    event.screenX,
                    event.screenY,
                    event.clientX,
                    event.clientY,
                    event.ctrlKey,
                    event.altKey,
                    event.shiftKey,
                    event.metaKey,
                    event.button,
                    event.relatedTarget,
                    event.captured=true);
            } else {
                eventToDispatch = new MouseEvent(event.type, event);
            }

            var buttonsValue = event.buttons;
            if (eventToDispatch.buttons !== buttonsValue) {
                Object.defineProperty(eventToDispatch, 'buttons', {
                    get: function () {
                        return buttonsValue;
                    }
                });
            }

            return eventToDispatch;
        }
        /**
         * Track if we want a 'long touch' i.e. a 'hold' or a system gesture
         * Invoke the gesture handler for a 'hold'
         * used for both mouse and touch
         * @param {VirtualTarget} vt
         * @param {boolean} state
         */
        function setHoldTimer(vt, state) {
            if(state) {
                holdTimer = clearTimeout(holdTimer);
                holdTimer = setTimeout(needsGesture, fastClickThreshold, vt);
                vt.waiting = true;
            }else {
                vt.waiting = false;
                holdTimer = clearTimeout(holdTimer);
            }
        }

        /**
         * Distance between p1 nad p2
         * @param {VirtualTarget} vt
         * @returns {number}
         */
        function distance(vt) {
            return Math.round(Math.sqrt(Math.pow((vt.p1.pageX - vt.p2.pageX), 2) + Math.pow((vt.p1.pageY - vt.p2.pageY), 2)));
        }

        /**
         * calculate the midpoint between p1 and p2
         * @param {VirtualTarget} vt
         * @returns {Point}
         */
        function midpoint(vt) {
            return {x:(vt.p1.x + vt.p2.x) / 2, y:(vt.p1.y + vt.p2.y) / 2};
        }

        /**
         * Calulate angle between p1 and p2 as an absolute value
         * @param {VirtualTarget} vt
         * @returns {number}
         */
        function angle(vt) {
            var dx = vt.p2.pageX - vt.p1.pageX;
            var dy = vt.p2.pageY - vt.p1.pageY;
            var hyp = Math.sqrt(Math.pow(dx, 2) + Math.pow(dy, 2));
            return Math.round(Math.acos(Math.abs(dx) / hyp) * 57.3);
        }

        /**
         *
         * @param {VirtualTarget} vt
         * @returns {number}
         */
        function elapsed(vt) {
            return vt.p2.timeStamp - vt.p1.timeStamp;
        }
        /**
         * Convert an angle to a direction
         * @param {VirtualTarget} vt
         * @returns {string}
         */
        function direction(vt) {
            var dx = vt.p2.pageX - vt.p1.pageX;
            var dy = vt.p2.pageY - vt.p1.pageY;
            var hyp = Math.sqrt(Math.pow(dx, 2) + Math.pow(dy, 2));
            if (hyp > swipeDistanceThreshold) {
                var angle = Math.round(Math.acos(Math.abs(dx) / hyp) * 57.3);
                return (angle > 45) ? ((vt.p1.pageY < vt.p2.pageY) ? 'down' : 'up') : ((vt.p1.pageX < vt.p2.pageX) ? 'right' : 'left');
            } else return 'cancel';
        }
        /**
         * Check whether the pointer has moved past a boundary (limit) since we started tracking.
         * @param {VirtualTarget} vt
         * @returns {boolean}
         */
        function checkHasMoved(vt) {
            var boundary = vt.pointerType === POINTER_TYPE_TOUCH ? touchBoundary : mouseBoundary;
            var dx = vt.p2.pageX - vt.p1.pageX;
            var dy = vt.p2.pageY - vt.p1.pageY;
            var hyp = Math.sqrt(Math.pow(dx, 2) + Math.pow(dy, 2));
            if (hyp > boundary) {
                //cancel the 'click'
                vt.trackingClick = false;
                //and also cancel the 'wait for the hold'
                setHoldTimer(vt, false);
                return true;
            }else return false;
        }
        /**
         * Set or unsets gesture capture for a target
         * @param {HTMLElement} target
         * @param {boolean} state
         * @memberof VMouseController#
         */
        function gestureCapture(target, state){
            if(state){
                target.addEventListener('mousedown', checkMouseGestureHandler, true);
                target.addEventListener('touchstart', checkTouchGestureHandler, false);

            }else{
                target.removeEventListener('mousedown', checkMouseGestureHandler, true);
                target.removeEventListener('touchstart', checkTouchGestureHandler, false);

            }
        }

        /**
         * force gesture on on a mouse enabled device - see gestureCapture
         * @param {Event} e
         */
        function checkMouseGestureHandler(e){
            var vt = vtRoot;
            // Permitted if pointerType is mouse (no touch event has fired)
            // or if its been too long since the last touch event fired
            // on touch we want to block all mouse events unless this has happened more than 300ms after last event
            if (vt.pointerType === POINTER_TYPE_MOUSE || (((Date.now() - (vt.p2.timeStamp || 0)) > 300))) {
                //we only get here on a mousedown that the main handler will also permit - but now we know its within our capture target so we can preset to force the gesture
                vt.forceGesture=true;
            }
        }

        /**
         * force gesture handling on on a touch enabled device - see gestureCapture
         * @param {Event} e
         */
        function checkTouchGestureHandler(e){
            var vt = vtRoot;
            if (e.targetTouches.length===1) {
                //we only get here on a touch that the main handler will also permit - but now we know its within our capture target so we can preset to force the gesture
                vt.forceGesture=true;
            }
        }

        /**
         * Do we need to track gestures
         * @param {VirtualTarget} vt
         * @return {boolean} return true for gesture handling active
         */
        function needsGesture(vt){
            // cancel the timer we know we arent 'hold' ing
            setHoldTimer(vt, false);
            // ok so we either held the mouse down long enough to timeout without moving the mouse
            // or we had flagged the element as gesture only
            // so it's rare enough that it's not unreasonable to take some time to check what to do
            // first check for a draggable element
            if(vt.draggable){
                // if pointer type is mouse and the target is draggable the dnd system will have already handled this
                // so we can ignore it
                // cancel the force
                vt.forceGesture = false;
                console.info('VMOUSE: switch to drag mode')
                // cancel click tracking
                endTracking(vt, 'click');
                //we are using gestures
                return true;
            }else{
                if (vt.forceGesture) {
                    //cancel the force
                    vt.forceGesture = false;
                    if(vt.target.dataset.gestureMode){
                        //has an fast auto gesture mode.....
                        switch(vt.target.dataset.gestureMode){
                            case 'dragHandle':
                                //this should be a mousedown on a handle.....
                                //find the controller then invoke its event handler.....
                                vt.gestureConfig = {dragHandle:vt.target, handleType:vt.target.dataset.handleType, dragMode:vt.target.dataset.dragMode, autoscroll:true};
                                vt.target = vt.target.closest('.drag-target');
                                vt.currentTarget = vt.target.offsetParent;
                                //find the controller......
                                var controller = vt.target.closest('[data-action-topic]').dataset.actionTopic.split(':')[1];
                                var ctl = Widget.getController(controller);
                                vt.gestureHandler = ctl.handleEvent;
                                //if defaultPrevented then tracking is wanted the handler must set currentTarget and gestureHandler
                                vt.trackingPointer = true;
                                // cancel click tracking
                                console.info('VMOUSE: switch to drag mode')
                                endTracking(vt, 'click');
                                //if they set the gesture type to dragdrop we will simulate dd
                                //call the mouse down handler....
                                pointerDownHandler(vt);
                                //we are using gestures
                                return true;
                                break;
                            case 'drag-move':
                                break;
                            default:
                                break;

                        }
                    }else {
                        //query the target to see whether it wants a gesture at this moment - this allows the target to setup gesture handlers
                        var e = new CustomEvent("gesture", {bubbles: true, cancelable: true, detail: vt});
                        //fire off the gesture needed query event on the target
                        var cancelled = !vt.target.dispatchEvent(e);
                        //default is no gesture needed
                        if (cancelled || e.defaultPrevented) {
                            //if defaultPrevented then tracking is wanted the handler must set currentTarget and gestureHandler
                            vt.trackingPointer = true;
                            console.info('VMOUSE: switch mouse to pointer')
                            // cancel click tracking
                            endTracking(vt, 'click');
                            //if they set the gesture type to dragdrop we will simulate dd
                            //call the mouse down handler....
                            pointerDownHandler(vt);
                            //we are using gestures
                            return true;
                        } else return false;
                    }
                } else {
                    // we get here if we exceeded the click threshold and no explicit gesture was set
                    // we could be in need of a long hold or a double click - check the target
                    // or we could need an implicit system gesture
                    if (vt.target.classList.contains('dbl-click')) {
                        //need to wait some more
                        return false;
                    } else {
                        if (vt.target.classList.contains('tap-hold')) {
                            //this target has a tapHold action so trigger it now.....
                            setHoldTimer(vt, false);
                            var action = vt.target.dataset.tapHold;
                            switch (action) {
                                case 'toggle-view':
                                    Views.toggleView('', Views.topMost, '');
                                    break;
                                default:
                                    Views.publishActionEvent(vt.target, 'tap-hold', '', vt.target.id, action);
                                    break;
                            }
                            //kill all tracking
                            endTracking(vt);

                        } else {
                            // so timedout with mousedown but not gesturable or draggable or dbl-click or tap-hold
                            // does a parent need a gesture?
                            //argument is that we should always put gesture on the target?????
                            return true;
                            //for internal gestures ??????
                            vt.trackingPointer = false;
                            // cancel click tracking and hold timer
                            endTracking(vt, 'click');

                            vt.trackingGesture = true;
                            //activate gesture track pad??
                            vt.currentTarget = document.body;
                            //activate gesture track pad??
                            vt.gestureHandler = internalGestureHandler;
                            //call the mouse down handler....
                            pointerDownHandler(vt);
                        }
                        return true;
                    }
                }
            }
        }

        /**
         * Check whether the given target element is a child of a scrollable layer and if so, set a flag on it.
         * @param {VirtualTarget} vt
         */
        function updateScrollParent(vt) {
            var scrollParent = vt.scrollParent;
            var el = vt.target;
            // Attempt to discover whether the target element is contained within a scrollable layer.
            // Re-check if the target element was moved to another parent.
            if (!scrollParent || !scrollParent.contains(el)) {
                var parentElement = el;
                do {
                    if (parentElement.scrollHeight > parentElement.offsetHeight) {
                        scrollParent = parentElement;
                        break;
                    }
                    parentElement = parentElement.parentElement;
                } while (parentElement);
            }
            // Always update the scroll top tracker if possible.
            vt.scrollParent = scrollParent;
            vt.scrollTop = scrollParent ? scrollParent.scrollTop : 0;
        }

        //endregion

        //region Touch
        var hasClick = {button: true, checkbox: true, radio: true, image:true};
        var inputTypes = {INPUT: true, TEXTAREA: true, ARTICLE: true};
        var hasFastClick = {MENUITEM:true,BUTTON:true};
        /**
         * On touch start, record the position and scroll offset.
         * @param {TouchEvent} e
         * @returns {boolean}
         */
        function onTouchStart(e) {
            /**@type {VirtualTarget}*/
            var vt = vtRoot;
            //ignore 'none clickable' input controls
            if (inputTypes[e.target.tagName] && !hasClick[e.target.type] && !(vt.trackingDrag || vt.trackingPointer || vt.trackingGesture || vt.trackingClick)) {
                //always check on mousedown whether we need to close anything
                autocloseHandler(e, e.target);
                //now ignore touches on inputs
                vt.p2.timeStamp = Date.now();
                return;
            }
            // Ignore multiple touches, otherwise pinch-to-zoom is prevented
            if (e.targetTouches.length===1) {
                //setup VT for a new run
                var el = e.target;
                var touch = e.targetTouches[0];

                vt.target = el;
                vt.touchID1 = touch.identifier;
                vt.p1.pageX = touch.pageX;
                vt.p1.pageY = touch.pageY;
                vt.p1.timeStamp = touch.timeStamp||Date.now();
                vt.p2.pageX = vt.p1.pageX;
                vt.p2.pageY = vt.p1.pageY;
                vt.p2.timeStamp = vt.p1.timeStamp;
                //flag as click tracking - if we switch to pointer tracking it will turn this off
                vt.trackingClick = true;
                //force close stuff
                autocloseHandler(e, el);
                //forceGesture can get preset or the target can flag gesture test required
                if (vt.forceGesture || vt.target.classList.contains('gesture')) {
                    //force gesture test
                    vt.forceGesture = true;
                    //need to query for gesture support now
                    needsGesture(vt);
                } else {
                    //pre click setup
                    checkFastClickMode();
                    //ok no forced gesture so now we need to wait to see if there is a long tap or a system gesture
                    setHoldTimer(vt, true);
                }
                // If the target element is a child of a scrollable layer (using -webkit-overflow-scrolling: touch) and:
                // 1) the user does a fling scroll on the scrollable layer
                // 2) the user stops the fling scroll with another tap
                // then the e.target of the last 'touchend' event will be the element that was under the user's finger
                // when the fling scroll was started, causing FastClick to send a click event to that layer
                // unless a check is made to ensure that a parent layer was not scrolled before sending a synthetic click.
                // so get the current scroll parents scrollTop so we can compare it when the touchend occurs
                updateScrollParent(vt);
            }
            // cant have two touches without first having one
            // if two touches then cant be a click - must be a gesture but explicit framework gesture or implicit system gesture
            if (e.targetTouches.length===2) {
                var touch1 = e.targetTouches[0];
                var touch2 = e.targetTouches[1];
                //rebase in case there was some drift before touchdown
                vt.touchID1 = touch1.identifier;
                vt.touchID2 = touch2.identifier;

                vt.p1.pageX = touch1.pageX;
                vt.p1.pageY = touch1.pageY;
                vt.p1.timeStamp = touch1.timeStamp||Date.now();

                vt.p2.pageX = touch2.pageX;
                vt.p2.pageY = touch2.pageY;
                vt.p2.timeStamp = touch2.timeStamp||Date.now();

                vt.pinching = true;
                vt.initialDistance = vt.distance = distance(vt);
                vt.midpoint = midpoint(vt);
                if (vt.trackingDrag || vt.trackingPointer || vt.trackingGesture) {
                    // should we recheck if gesture is still valid with multiple touches
                }else{
                    // need to kill the hold and check if gesturing
                    if (vt.waiting) {
                        needsGesture(vt);
                    } else {

                    }
                }
            }
            return true;
        }
        /**
         * Update the last position.
         * @param {TouchEvent} e
         * @returns {boolean}
         */
        function onTouchMove(e) {
            /**@type {VirtualTarget}*/
            var vt = vtRoot;
            //ignore none clickable input controls
            if ((inputTypes[e.target.tagName] && !hasClick[e.target.type]) || !(vt.trackingDrag || vt.trackingPointer || vt.trackingGesture || vt.trackingClick)) {
                vt.p2.timeStamp = Date.now();
                return;
            }
            //we know its definitely not a hold
            if (vt.trackingPointer || vt.trackingGesture || vt.waiting) {
                if (e.targetTouches.length===1) {
                    var touch1 = e.targetTouches[0];
                    //make sure this is still the same finger
                    if (vt.touchID1 === touch1.identifier) {
                        //update the pointer info
                        vt.p2.timeStamp = touch1.timeStamp || Date.now();
                        vt.p2.pageX = touch1.pageX;
                        vt.p2.pageY = touch1.pageY;
                        // setting trackPointer kills the holdTimer so ignore it
                        //call the gesture handler....
                        if (vt.trackingPointer || vt.trackingGesture) {
                            pointerMoveHandler(vt);
                        } else {
                            //if we are waiting we cant be tracking pointer or gesture only clicks
                            if (vt.waiting) {
                                if (checkHasMoved(vt)) {
                                    //we moved enough to kill the hold and click
                                    console.info('VMOUSE: moved pointer...... kill click')
                                    //cancel the 'click'
                                    endTracking(vt, 'click');
                                    endTracking(vt);
                                }
                                return;
                            }
                        }
                    }
                }
                if (e.targetTouches.length===2) {
                    //we must be tracking pointer and in gesture mode
                    var touch1 = e.targetTouches[0];
                    var touch2 = e.targetTouches[1];
                    if (vt.touchID1 === touch1.identifier && vt.touchID2 === touch2.identifier) {
                        //update both pointer info
                        vt.p1.pageX = touch1.pageX;
                        vt.p1.pageY = touch1.pageY;
                        vt.p1.timeStamp = touch1.timeStamp || Date.now();

                        vt.p2.pageX = touch2.pageX;
                        vt.p2.pageY = touch2.pageY;
                        vt.p2.timeStamp = touch2.timeStamp || Date.now();

                        vt.distance = distance(vt.p1, vt.p2);
                        vt.midpoint = midpoint(vt.p1, vt.p2);
                        vt.scale = vt.distance / vt.initialDistance;
                        //call the gesture handler....
                        if (vt.trackingPointer || vt.trackingGesture) {
                            pointerMoveHandler(vt);
                        }
                    }
                }
            }
            return true;
        }
        /**
         * On touch end, determine whether to send a click event at once.
         * @param {TouchEvent} e
         * @returns {boolean}
         */
        function onTouchEnd(e) {
            var vt = vtRoot;
            //ignore none clickable input controls
            if ((inputTypes[e.target.tagName] && !hasClick[e.target.type]) || !(vt.trackingDrag || vt.trackingPointer || vt.trackingGesture || vt.trackingClick)) {
                vt.p2.timeStamp = Date.now();
                return;
            }
            vt.isDown = false;
            if (vt.trackingPointer || vt.trackingGesture) {
                //if tracking more than one touch then we could lift either one or even some other
                //but lifting one we are tracking will end everything
                var touch = e.changedTouches[0];
                if(vt.touchID1&&vt.touchID2){
                    if(vt.touchID1 === touch.identifier) {
                        //if we lifted one update one if we lifted two update two
                        vt.p1.timeStamp = touch.timeStamp || Date.now();
                        vt.p1.pageX = touch.pageX;
                        vt.p1.pageY = touch.pageY;
                        //call the mouse down handler one last time....then cleanup
                        pointerUpHandler(vt);
                        endTracking(vt);
                    }
                    if(vt.touchID2 === touch.identifier) {
                        //if we lifted one update one if we lifted two update two
                        vt.p2.timeStamp = touch.timeStamp || Date.now();
                        vt.p2.pageX = touch.pageX;
                        vt.p2.pageY = touch.pageY;
                        //call the mouse down handler one last time....then cleanup
                        pointerUpHandler(vt);
                        endTracking(vt);
                    }
                }else{
                    if(vt.touchID1 === touch.identifier) {
                        vt.p2.timeStamp = touch.timeStamp || Date.now();
                        vt.p2.pageX = touch.pageX;
                        vt.p2.pageY = touch.pageY;
                        //call the mouse down handler one last time....then cleanup
                        pointerUpHandler(vt);
                        endTracking(vt);
                    }
                }
            }
            //click synthesis is iOS only......
            if(vt.trackingClick) {
                var touch = e.changedTouches[0];
                //if we are tracking a single touch then make sure this is our touch
                if(vt.touchID1 === touch.identifier) {
                    vt.p2.timeStamp = touch.timeStamp || Date.now();
                    vt.p2.pageX = touch.pageX;
                    vt.p2.pageY = touch.pageY;
                    //stop checking for a hold
                    setHoldTimer(vt, false);
                }
            }
            return true;
        }
        /**
         * On touch cancel, stop tracking the click.
         * @param {Event} e
         * @returns {void}
         */
        function onTouchCancel(e) {
            var vt = vtRoot;
            setHoldTimer(vt, false);
            endTracking(vt)
        }

        //endregion

        /**
         * Centralised 'reset' after all tracking is completed
         * @param {VirtualTarget} vt
         * @param {string} [mode]
         */
        function endTracking(vt, mode) {
            if(mode==='click'){
                if(vt.trackingClick){
                    console.info('VMOUSE: CANCEL CLICK')
                    vt.trackingClick = false;
                    if(vt.fastClickMode) {
                        //need to restore focus back to the original ae
                        Focus.setFocusLock(false, null, true);
                    }
                }else console.info('VMOUSE: CANCEL NON CLICK')
                //no need to wait any longer - but we can try a system gesture....
                setHoldTimer(vt, false);
            }else {
                vt.trackingClick = false;
                vt.trackingPointer = false;
                vt.trackingGesture = false;
                vt.trackingDrag = false;

                vt.isDown = false;

                vt.fastClickMode = false;

                vt.activeElement = null;

                vt.gestureHandler = null;
                vt.gestureConfig = false;
                vt.gestureContext = false;

                vt.currentTarget = false;

                vt.scrollParent = null;

                vt.touchID1 = vt.touchID2 = null;
                vt.pinching = false;

                vt.needsNextClick = false;
                vt.cancelNextClick = false;

                vt.touchID1 = vt.touchID2 = null;
                //records last action time
                vt.p2.timeStamp = Date.now();
            }
        }

        //region Click
        //actions that don't force an element commit
        //an explicit close pulls focus and force a commit - then it saves if required and dirty
        var ncAction = {clear:true, cmd:true, menu:true, more:true, flip:true, expandable:true, expandAll:true, collapseAll:true, selectAll:true, unselectAll:true, toggleAll:true, tagAll:true, untagAll:true, pickColor:true, pickDay:true, "gantt-zoom":true, "toggle-class":true, 'dialog-check-all':true, 'dialog-check-none':true};
        /**
         * check whether an action should be in fastClickMode
         * A click comprises of a mouse down/up or touch start/end pair and then a click event
         * This results in focus being removed from the current active element
         * Chrome(mac) and All(pc) focus moves to the clicked element and (Safari/FF(mac)/Safari/Chrome(ios)) to the body
         * This means the activeElement on click is never the original activeElement.
         * Some actions target the original active element,
         * in other cases it is assumed that focus after the action will still be in the original element
         * But on other cases the action depends on any uncommitted changes in input control being commited before the action.
         * So in all cases focus is removed but in fastClickMode focus is restored and a non-standard controls will not trigger focus/blur so wont commit
         * So need to be careful about when we use fastClickMode.
         * explicit 'nofocus'
         * helper buttons (icon[none]), text-editor and selectlist buttons, should commit old control if cross control and should ensure focus goes to new control
         * list in dropdown, lookup, selectlist, can't get to them unless already on the control no commit and no danger...
         * pickers cant get to them unless already on control definite nf but no commit wanted
         * review buttons no danger or commit needed
         * random other buttons want focus lock but should commit first but only needed if a custom control
         *
         */
        function checkFastClickMode(){
            var vt = vtRoot;
            //menuitem or button are automaticallly fc candidates
            var fastClick = hasFastClick[vt.target.tagName];

            if(!fastClick){
                //anything that is explicitly nf
                fastClick = vt.target.classList.contains('nofocus') ;
            }
            if(!fastClick){
                //anything that is an icon with an action (helper) if helper is for
                fastClick = (vt.target.tagName === 'ICON' && vt.target.dataset.action);
            }
            if(!fastClick){
                //ignore inside a picker
                fastClick =  vt.target.closest('dialog.picker,dialog.select');
            }
            //cant fc if need tap-hold
            if(vt.target.classList.contains('tap-hold')) fastClick = false;

            if(fastClick){
                //what currently has focus
                vt.activeElement = document.activeElement;
                //do we need to force the active element to commit before processing the click
                //if ae is not in a control we dont commit
                //if ae is in a control and the target is in the same control we dont commit and its a non-standard control need to force a commit..
                //if ae is in a control and the target is in the related picker dialog/menu we dont commit.
                // how to detect/track
                // if there is a dialog/picker it is topmost - and we are anchor - we add active anchor on opening a dialog or menu
                // dropdown/select/lookup/date/color and floating menu - are anchored input-helper
                // when acting as a helper to a control it shouldnt commit -the control can decide if it needs to commit....
                var icActive = vt.activeElement.closest('.input-control');
                var icTarget = vt.target.closest('.input-control');
                var forceCommit = !( (!icActive) ||
                                     (icActive === icTarget) ||
                                     (icActive && vt.target.closest('menu.anchored.input-helper')) ||
                                     (icActive && vt.activeElement.classList.contains('anchor') && DialogManager.topMost && DialogManager.topMost.contains(vt.target))
                                  );
                //if we need fastClick but cant then revert to fl - will need to restore focus
                //if ae is in a control and that control isnt the target we need fl on the new target..
                //if ae is not in a control and the target is we need fl on the new target..
                var focusChange = (icActive && icTarget && icActive !== icTarget) || (!icActive && icTarget);

                if(forceCommit){
                    console.info('VMOUSE: FORCE COMMIT NEEDED')
                    Forms.commitActiveElement(vt.activeElement.form, vt.activeElement, focusChange );
                }
                //if we need fastClick but cant then revert to fl - will need to restore focus
                //if ae is in a control and that control isnt the target we need fl on the new target..
                //if ae is not in a control and the target is we need fl on the new target..
                if(focusChange){
                    //if clicking a helper then switch focus to new control - but focus lock will prevent it firing.....
                    vt.activeElement = icTarget.querySelector('article,input:not([hidden]),textarea')
            }
                console.info('VMOUSE: FOCUS WAS ', document.activeElement , 'WILL BE ' , vt.activeElement)
                Focus.setFocusLock(true, vt.activeElement);
            }
            vt.fastClickMode = fastClick;
        }
        /**
         * On actual clicks, determine whether this is a touch-generated click, a click action occurring
         * naturally after a delay after a touch (which needs to be cancelled to avoid duplication), or
         * an actual click which should be permitted.
         * This is the real click event
         *
         * @param {Event} e
         * @returns {boolean}
         */
        function onClick(e) {
            var vt = vtRoot;
            console.info('VMOUSE: CLICK',vt,e)
            //check for dblclick in progress
            if(vt.dblclick){
                if(e.detail===2){
                    //yes its a true double click
                }else{
                    //it cant be a double click so kill it
                    vt.dblclick=clearTimeout(vt.dblclick);
                }
            }
            // programmatic clicks all go through (pageX/Y are 0)
            // all other clicks go through unless tracking was ended (because gesture tracking was enabled)
            var permitted = (e.pageX === 0 && e.pageY === 0) || (vt.trackingClick && (vt.target === e.target));
            // If clicks are permitted, send virtual event.
            if(permitted){
                if(vt.fastClickMode){
                    //need to restore focus back to the original ae
                    Focus.setFocusLock(false, null, true);
                    //if ae is in a control and that control is not the target and its a non-standard control need to force a commit..
                    //var ic = vt.activeElement.closest('.input-control');
                    //also if the focus is going to the related picker/dialog we shouldnt commit - how to detect/track
                    //if there is a dialog/picker it is topmost - and we are anchor
                    ///var forceCommit = (ic && ic.contains(vt.target)) ? false : vt.activeElement.classList.contains('anchor') inputTypes[vt.activeElement.tagName];
                    //if(forceCommit)Forms.commitActiveElement(vt.activeElement.form, vt.activeElement);
                }
                clickHandler(e, e.target);
                //always set this after done processing a click
                vt.lastClickTime = Date.now();
            }else{
                endTracking(vt,'click');
            }
            // no longer tracking a click
            // we have to turn this off here in case the code needed to directly invoke a click
            // like file upload
            endTracking(vt);
            return permitted;
        }

        //endregion

        //region Mouse
        /**
         * Determine mouse events which should be permitted.
         * @param {MouseEvent} e
         * @returns {boolean}
         */
        function onMouse(e) {
            var vt = vtRoot;
            // ignore 'none clickable' input controls
            if (inputTypes[e.target.tagName] && !hasClick[e.target.type] && !(vt.trackingDrag || vt.trackingPointer || vt.trackingGesture || vt.trackingClick)) {
                //always check on mousedown whether we need to close anything
                if (e.type === 'mousedown') autocloseHandler(e, e.target);
                return;
            }
            switch (e.type) {
                case 'mousedown':
                    //setup VT for a new run
                    vt.target = e.target;
                    vt.isDown = true;
                    vt.p1.pageX = e.pageX;
                    vt.p1.pageY = e.pageY;
                    vt.p1.timeStamp = Date.now();
                    vt.p2.pageX = vt.p1.pageX;
                    vt.p2.pageY = vt.p1.pageY;
                    vt.p2.timeStamp = vt.p1.timeStamp;
                    //flag as click tracking - if we switch to pointer tracking it will turn this off
                    vt.trackingClick = true;
                    //always check on mousedown whether we need to close anything
                    autocloseHandler(e, e.target);
                    //ok now need to see if we need gesture support and find the right controller
                    if (vt.forceGesture || vt.target.classList.contains('gesture')) {
                        vt.forceGesture = true;
                        needsGesture(vt);
                    } else {
                        //pre click setup
                        checkFastClickMode();
                        //if you hold for longer than limit this will become a long hold regardless of what you are clicking
                        setHoldTimer(vt, true);
                    }
                    console.info('VMOUSE: DOWN',vt,e)
                    break;
                case 'mouseover':
                    //pointerEnterHandler(e, e.target);
                    break;
                case 'mouseout':
                    //pointerLeaveHandler(e, e.target);
                    break;
                case 'mousemove':
                    vt.p2.timeStamp = Date.now();
                    vt.p2.pageX = e.pageX;
                    vt.p2.pageY = e.pageY;
                    if (vt.trackingPointer || vt.trackingGesture) {
                        pointerMoveHandler(vt);
                    } else if (vt.trackingClick) {
                        if (checkHasMoved(vt)) {
                            //we moved enough to kill the hold and click
                            console.info('VMOUSE: pointer moved.......... cancel click')
                            //cancel the 'click'
                            endTracking(vt, 'click');
                            endTracking(vt);
                        }

                    }//else we arent tracking anything so why bother checking
                    break;
                case 'mouseup':
                    console.info('VMOUSE: UP',vt,e)
                    vt.isDown = false;
                    if (vt.trackingPointer || vt.trackingGesture) {
                        vt.p2.timeStamp = Date.now();
                        vt.p2.pageX = e.pageX;
                        vt.p2.pageY = e.pageY;
                        pointerUpHandler(vt);
                        endTracking(vt);
                    }else {
                        //were done so no need to keep looking for a 'hold'
                        setHoldTimer(vt, false);
                        if(vt.trackingClick)console.info('VMOUSE: CLICK COMING')
                        if(vt.trackingClick && vt.target !== e.target)console.info('VMOUSE: CLICK NOTTTTTTTTTTTTTT COMING')

                        //there will be a click to follow.....
                    }
                    break;
                case 'dragstart':
                    vt.isDown = true;
                    if(e.target.nodeType===1 && e.target.closest('[draggable]')){
                        vtRoot.trackingDrag = true;
                        vtRoot.target = e.target;
                        vtRoot.isDown = true;
                        vtRoot.trackingDrag = true;
                        handleDragStart(e);

                    }else{
                        e.preventDefault();
                        e.stopPropagation();
                    }
                    break;
                case 'dragend':
                    vt.isDown = false;
                    handleDragEnd(e);
                    endTracking(vt);
                    break;
                case 'drag':
                    handleDrag(e);
                    break;
                case 'dragenter':
                    handleDragEnter(e);
                    break;
                case 'dragover':
                    handleDragOver(e);
                    break;
                case 'dragleave':
                    handleDragLeave(e);
                    break;
                case 'drop':
                    vt.isDown = false;
                    handleDrop(e);
                    endTracking(vt);
                    break;
                default:
                    break;
            }
        }

        //endregion

//endregion

//region Handlers
        /**
         * Gets called by real events as well as synthetics - part of the mousedown or touchstart handling
         * Do we need to close something before continuing with processing this action
         * @param {Event} e the mouse or touch event
         * @param {HTMLElement} target the object clicked or touched
         */
        function autocloseHandler(e, target){
            if(!e.defaultPrevented) {
                // we tapped or clicked something
                // must be inside the DC(Dialog) or MC(Menu) or PC(Page) or the AppHeader or in Body
                // first check to see if we are in the dialog or the page containers
                // - if the mdo is inside a DIALOG check whether there is a nested dialog (select or picker)
                //      you can't get a DIALOG open without clicking on something or tab in/out of an input
                //      so no need to check grid editor, panels or pages - everything that should be closed is already closed
                // - if the mdo is inside the page container it might do any or all of the following (order is specific and relevant)
                //   - close a dialog
                //   - close a grid/list editor
                //   - closing a panel or page
                // closing something never directly blocks the new action (even though that may cause an error)
                // exception is modals they do abort the action
                // during closing if there is a form validation failure then the click needs to be killed off and focus returns to the form...
                // the close is then queued for later although the click action will be lost in this case
                // start at target and go for the body - same as click handler - track the breadcrumbs
                var cur = target;
                var prev1;
                var root = document.body;
                var pc = window.pageContainer;
                var dc = window.dialogContainer;
                var mc = window.menuContainer;
                //we can check the mc without looping
                if(mc.children.length){
                    if(mc.contains(target)){
                        //only way we can be in the mc is if there is an active menu and we clicked it....
                        //there can only be one active menu at a time
                        //so let the clickhandler deal with it but no need to check anything else we dont want to close anything
                        return;
                    }else{
                        //clicked outside the menu but a menu was active so close it but see what else needs closing
                        Menus.closeMenu(mc.firstElementChild);
                    }
                }
                for ( ; cur !== root; cur = cur.parentNode || root ) {
                    if(cur===mc) {
                        //only way we can be in the mc is if there is an active menu and we clicked it....
                        //there can only be one active menu at a time
                        //so let the clickhandler deal with it but no need to check anything else we dont want to close anything
                        break;
                    }else if(cur===dc){
                        //MDO inside a dialog or menu - prev1 is the dialog or menu
                        DialogManager.autocloseHandler(e, target, prev1);
                        break;
                    }else if(cur===pc){
                        //MDO inside pc - prev1 is page

                        // checks for any open picker or select
                        // note an open dialog may fail to close due to validation issues
                        // forms and pages will then not close if their is an open dialog
                        DialogManager.autocloseHandler(e, target, prev1);
                        // check whether we have any active embedded forms (list or grid)
                        Forms.autocloseHandler(e, target, prev1);
                        //ok now lets say that we had a mousedown outside a picker and outside the page/panel
                        //it needs to first close the picker as we have above and then close the page
                        //page close could force save prompts
                        //if a page cant close it will be queued for later
                        if(target===pc){
                            //can only happen if u click empty space on right of a base page
                        }else {
                            // check for any page or panel to close
                            // even if we had a failed validation we can queue the page close
                            Views.autocloseHandler(e, target, prev1);
                        }
                        break;
                    }else {
                        //keep on going
                        prev1 = cur;
                    }
                }
                // if we reached root and the prev is the app header then check for dialogs
                // in some cases if the base page is narrow but the screen is large then the 'body' is exposed
                // so u can directly click on body whic means prev1 etc will all be null...
                if (cur === root && prev1 && prev1.tagName === 'HEADER') {
                    DialogManager.autocloseHandler(e, target, prev1);
                }
            }
        }
        /**
         * translate the coordinates from a global context to a local context.
         * offsetX/Y allows for scroll and is offset in the client area
         *  x/y arent effected by scroll and are the physical offset form parent
         * @param {VirtualTarget} vt
         * @param {HTMLElement} [currentTarget] the local context element effectively the offset parent for this calculation
         * @returns {VirtualTarget} for chaining
         * @memberof VMouseController#
         */
        function globalToLocal( vt, currentTarget ){
            // Get the position of the context element.
            var target = (currentTarget || vt.currentTarget || vt.target);
            var rect = target.getBoundingClientRect();
            if(!vt.p1.offsetX && !vt.p1.offsetY){
                //first time through set the p1 position...
                vt.p1.offsetX = Math.floor(vt.p1.pageX - rect.left + target.scrollLeft);
                vt.p1.offsetY = Math.floor(vt.p1.pageY - rect.top + target.scrollTop);
            }
            vt.p2.offsetX = Math.floor(vt.p2.pageX - rect.left + target.scrollLeft);
            vt.p2.offsetY = Math.floor(vt.p2.pageY - rect.top + target.scrollTop);
            // Return the X/Y in the local context.
            vt.p1.x = Math.floor(vt.p1.pageX - rect.left);
            vt.p1.y = Math.floor(vt.p1.pageY - rect.top);
            vt.p2.x = Math.floor(vt.p2.pageX - rect.left);
            vt.p2.y = Math.floor(vt.p2.pageY - rect.top);

            return vt;
        }

        /**
         * Get called by real event as well as synth
         * @param {VirtualTarget} vt
         */
        function pointerCancelHandler(vt){
            vt.type='cancel';
            globalToLocal(vt);
            vt.gestureHandler(vt);
        }
        /**
         * Get called by real event as well as synth
         * @param {VirtualTarget} vt
         */
        function pointerDownHandler(vt){
            vt.type='down';
            vt.active=true;
            vt.p1.offsetX = vt.p1.offsetY =0;
            if(vt.gestureConfig){
                if(vt.gestureConfig.autoscroll){
                    //listen for scroll event and trigger move event during scroll
                    vt.currentTarget.addEventListener('scroll',scrollEvent);
                }
            }
            //localize vt to current gesture target (not same as event target)
            globalToLocal(vt);
            //used in markup global pos of container topLeft
            vt.xOffset = vt.p1.x - vt.target.offsetLeft;
            vt.yOffset = vt.p1.y - vt.target.offsetTop;
            vt.xInc = 0;
            vt.yInc = 0;
            vt.gestureHandler(vt);
        }
        /**
         * Get called by real event as well as synth
         * @param {VirtualTarget} vt
         */
        function pointerMoveHandler(vt){
            vt.type='move';
            //localize vt to current gesture target (not same as event target)
            globalToLocal(vt);

            if(vt.gestureConfig){
                vt.xInc = vt.p2.offsetX - vt.p1.offsetX - vt.xInc;
                vt.yInc = vt.p2.offsetY - vt.p1.offsetY - vt.yInc;
                vt.gestureHandler(vt);
                vt.xInc = vt.p2.offsetX-vt.p1.offsetX;
                vt.yInc = vt.p2.offsetY-vt.p1.offsetY;
                if(vt.gestureConfig.autoscroll){
                    checkOverflow(vt)
                }
            }else{
                vt.xInc = vt.p2.x - vt.p1.x - vt.xInc;
                vt.yInc = vt.p2.y - vt.p1.y - vt.yInc;
                vt.gestureHandler(vt);
                vt.xInc = vt.p2.x-vt.p1.x;
                vt.yInc = vt.p2.y-vt.p1.y;
            }
        }
        /**
         * Get called by real event as well as synth
         * @param {VirtualTarget} vt
         */
        function pointerUpHandler(vt){
            vt.type='up';
            //localize vt to current gesture target (not same as event target)
            globalToLocal(vt);
            vt.xInc = 0;
            vt.yInc = 0;
            vt.gestureHandler(vt);
            vt.active=false;
            if(vt.gestureConfig){
                if(vt.gestureConfig.autoscroll){
                    vt.currentTarget.removeEventListener('scroll', scrollEvent);
                }
            }
        }

        /**
         * check whether the mouse is at the top or bottom of the screen and we need to force a scroll
         * @param {VirtualTarget} vt
         * @param vt
         */
        function checkOverflow(vt){
            //skip this for now we cant block the chrome autoscrolll but check Safari/FF/IE
            if(vt.gestureConfig.autoscroll){
                var dir = vt.p2.y < 150 ? -1 : vt.p2.y > (vt.currentTarget.scrollHeight - 150) ? 1 : 0
                if (dir) {
                    vt.gestureConfig.scrollStop = false;
                    scrollTarget(vt, dir)
                } else vt.gestureConfig.scrollStop = true;
            }
        }

        /**
         * Programmatic scrolling
         * @param {VirtualTarget} vt
         * @param {number} step
         */
        function scrollTarget(vt, step) {
            if(vt.active) {

                vt.currentTarget.scrollTop += step;
                if (!vt.gestureConfig.scrollStop) {
                    setTimeout(function () {scrollTarget(vt, step)}, 20);
                }
            }
        }

        /**
         * Autoscrolling while gesturing.....
         * @param {Event} e
         */
        function scrollEvent(e){
            var vt = vtRoot;
            vt.p2.timeStamp = Date.now();
            if (vt.trackingPointer || vt.trackingGesture) {
                pointerMoveHandler(vt);
            }
        }
        /**
         * Handler for system implicit gestures
         * @param {VirtualTarget} vt
         */
        function internalGestureHandler(vt) {
            switch (vt.type) {
                case 'down':
                    var result = Views.gestureHandler(direction(vt), vt.type, vt);
                    break;
                case 'move':
                    var scrollTop = vt.scrollTop;
                    updateScrollParent(vt);
                    var result = Views.gestureHandler(direction(vt), (vt.scrollTop === scrollTop) ? vt.type : 'cancel', vt);
                    break;
                case 'up':
                    var scrollTop = vt.scrollTop;
                    updateScrollParent(vt);
                    var result = Views.gestureHandler(direction(vt), (vt.scrollTop === scrollTop) ? vt.type : 'cancel', vt);
                    break;
            }
        }
        /**
         * @typedef {Object} ClickEventCtx
         * @property {boolean} stopPropagation
         * @property {boolean} defaultPrevented
         * @property {HTMLElement} target
         * @property {HTMLElement} currentTarget
         * @property {HTMLElement} prev1
         * @property {HTMLElement} prev2
         * @property {HTMLElement} prev3
         * @property {Array} path
         * @property {Object} details
         * @property {Event} e
         * @property {View} view
         */
        /**
         * Simulate the browser event path if calling on a different target than the event
         * @param {HTMLElement} node
         */
        function getPath(node){
            var path = [];
            while(node) {
                path.push(node);
                node = node.parentNode;
            }
            //window is not a parentNode but get pushed on in the native implementation
            path.push(window);
            //return the path
            return path;
        }

        /**
         * Triggered from timeout waiting for double click
         * @param {HTMLElement} el
         * @param {Event} e
         */
        function dblClickHandler(el, e){
            vtRoot.dblclick=0;
            Views.publishActionEvent(el, e);
        }
        //check if link begins with protocol
        var rxLinkCheck = /^(https?|sftp|ftp|file|nfs|smb|afp|git)/;
        /**
         * Handler for application all click events
         * if a sub handler returns false propagation will stop
         * calling preventDefault may occasionally have some external browser impact like stopping navigation on an 'a' anchor
         * but primarily it may be used internally to prevent another sub handler carrying out its default action
         * @param {Event} e
         * @param {HTMLElement} target e.target or vt.target
         * @returns {boolean} return value not used
         */
        function clickHandler(e, target) {
            //we are in click node and got this far - do we need to track for double click
            var vt = vtRoot;
            // double click/tap are rare - they can only exist on a div or a widget.
            // you have to set the attribute data-dblclick-action="" in order to trigger double click tracking
            // note that when double click tracking is active then all clicks are delayed....
            // get the bubble path - note native is a NodeList
            // window document html body are >always< the last 4.
            // also note that window and document are NOT elements
            // pageContainer view should be the next for views
            // dialogContainer dialog for popups
            //IE11 and Edge dont support composedPath (polyfilled) - if its not the real target we create a fake path in getPath
            var path = (e.target === target) ? e.composedPath() : getPath(target);
            var pathLen = path.length;
            var prev1, prev2, prev3;
            var view = Views.findView(target);

            /**@type ClickEventCtx*/
            var ectx = {
                currentTarget: target,
                target: target,
                prev1: null,
                prev2: null,
                prev3: null,
                details: null,
                path: path,
                e: e,
                view: view,
                stopPropagation: false,
                defaultPrevented: false
            };
            //the element we are currently processing
            var cur = target;
            //handlers need to return true to cause an event to continue to bubble up the chain
            //we bail at the view level...
            //init - check - inc
            for(var i=0;i < pathLen - 4; i++){
                cur = ectx.currentTarget = path[i];

                if (cur === view || cur.disabled || ectx.stopPropagation) {
                    //were done....... we never process the view or higher or if its disabled or if someone has blocked
                    if(cur.tagName === 'MENU') Menus.closeMenu(cur);
                    break;
                }else {

                    prev1 = ectx.prev1 = i > 0 ? path[i - 1] : null;
                    prev2 = ectx.prev2 = i > 1 ? path[i - 2] : null;
                    prev3 = ectx.prev3 = i > 2 ? path[i - 3] : null;

                    var type = cur.tagName;
                    // can occur in usr content
                    // A > requires data-action (native will correctly fire)
                    // I > requires class
                    // UL > requires ctx (very late)
                    // OL > requires data-type
                    // TABLE > requires class grid-l
                    // TD DIV H1-H6 > requires data-action
                    // BUTTON > requires data-action (late) (anything inside a button should not fire events)
                    switch (type) {
                        case 'A':
                            if (cur.dataset.action) {
                                //just an 'action' click
                                buttonClickHandler(cur, view, ectx);
                                ectx.defaultPrevented = true;
                            } else {
                                //a real 'a' link - skip our processing and let the browser handle it... don't prevent default
                                //usually only used for in page anchors
                                // except now some usr content is not being 'escaped' but contains links which have not been made safe.....
                                //if we do nothing then platinum will get nuked so lets set a blank target to force the link to pop open in a new tab
                                // if no target is set and the link has a protocol and not that it is an internal jump (same origin and page anchor)
                                if(!cur.target && rxLinkCheck.test(cur.href) && !(cur.href.includes(location.origin) && cur.href.includes('#'))){
                                    // add target
                                    cur.target='_blank';
                                    // for security on 'wild' links in usr content
                                    cur.setAttribute('rel', 'noopener noreferer');
                                }
                                ectx.stopPropagation = true;
                            }
                            break;
                        case 'INPUT':
                        case 'ARTICLE':
                        case 'TEXTAREA':
                            inputHandler(cur, view, ectx);
                            break;
                        case 'IMG':
                            if(cur.classList.contains('image-preview')){
                                PictureViewer.showViewer(cur);
                                ectx.stopPropagation = true;
                            }
                            break;
                        case 'ICON':
                            //if isnt a btn ignore it
                            if (cur.classList.contains('btn')) {
                                buttonClickHandler(cur, view, ectx);
                            } else if (cur.classList.contains('helper')) {
                                helperClickHandler(cur, view, ectx);
                            } else if (cur.hasAttribute('data-action')) {
                                buttonClickHandler(cur, view, ectx);
                            } else if (cur.classList.contains('helper-colorpicker')) {
                                helperClickHandler(cur, view, ectx);
                            }
                            break;
                        case 'MENUITEM':
                        case 'BUTTON':
                            buttonClickHandler(cur, view, ectx);
                            break;
                        case 'LI':
                            break;
                        case 'TASKCARD':
                            if(!ectx.defaultPrevented && !ectx.stopPropagation) ectx.stopPropagation = !!TaskBoard.clickHandler(ectx, view, cur.parentElement.parentElement.parentElement, cur.parentElement.parentElement, cur);
                            if(!ectx.defaultPrevented && !ectx.stopPropagation) buttonClickHandler(cur, view, ectx);
                            break;
                        case 'TASKBIN':
                            if(!ectx.defaultPrevented && !ectx.stopPropagation) ectx.stopPropagation = !!TaskBoard.clickHandler(ectx, view, cur.parentElement, cur);
                            if(!ectx.defaultPrevented && !ectx.stopPropagation) buttonClickHandler(cur, view, ectx);
                            break;
                        case 'TASKBOARD':
                            if(!ectx.defaultPrevented && !ectx.stopPropagation) ectx.stopPropagation = !!TaskBoard.clickHandler(ectx, view, cur);
                            if(!ectx.defaultPrevented && !ectx.stopPropagation) buttonClickHandler(cur, view, ectx);
                            break;
                        case 'LISTVIEW':
                        case 'UL':
                            listClickHandler(ectx);
                            break;
                        case 'OL':
                            // Dropdown's, Lookup's and SelectList's have a data-index attribute indicating index in list
                            // check type of list control - use currentTarget
                            // as it points at where the index attr is so parentEl points at where the class is
                            var type = cur.dataset.type;
                            switch (type) {
                                case 'listview-lookup':
                                    ectx.stopPropagation = !Lookup.selectHandler(ectx, view, cur, prev1, prev2, prev3);
                                    break;
                                case 'listview-dropdown':
                                    ectx.stopPropagation = !Dropdown.selectHandler(ectx, view, cur, prev1, prev2, prev3);
                                    break;
                                case 'listview-selectlist':
                                    ectx.stopPropagation = !SelectList.selectHandler(ectx, view, cur, prev1, prev2, prev3);
                                    break;
                                default:
                                //just a basic LISTVIEW
                                //listClickHandler(ectx);
                            }
                            break;
                        case 'CMVIEWSTACK':
                            //viewstack do nothing....
                            ectx.defaultPrevented = true;
                            ectx.stopPropagation = true;
                            break;
                        case 'MENU':
                            //preventing default doesnt block a close only a stop propagation will prevent this being triggered
                            //which is what we do for submenu actions
                            if(cur!==target) Menus.closeMenu(cur);
                            ectx.defaultPrevented = true;
                            ectx.stopPropagation = true;
                            break;
                        case 'HEADER':
                            headerClickHandler(ectx);
                            break;
                        case 'FOOTER':
                            footerClickHandler(cur, view, ectx);
                            break;
                        case 'XBALLOON':
                            //TODO:does this really exist
                            if (ectx.defaultPrevented) {
                                //some lower elment triggered an action - now check dismiss
                                var dismiss = $.getFirstElementAttrValue(ectx.target, 'data-dismiss', cur, 'popup');
                                //if none then no close if default (popup) then just popup close otherwise standard close
                                switch (dismiss) {
                                    case 'none':
                                        break;
                                    case 'popup':
                                        Views.dismiss(dismiss, cur, false, ectx);
                                        break;
                                    default:
                                        Views.close(cur.parentElement);
                                        break;
                                }
                            }
                            //stop this propagating further
                            ectx.stopPropagation = true;
                            break;
                        case 'FORM':
                            //stop clicks bubbling out of a form
                            ectx.stopPropagation = true;
                            break;
                        case 'DATAGRID':
                            ectx.stopPropagation = !DataGrid.clickHandler(ectx, view, cur, prev1, prev2, prev3);
                            break;
                        case 'TABLE':
                            if (cur.classList.contains('grid-l')) {
                                ectx.stopPropagation = !Lookup.selectHandler(ectx, view, cur, prev1, prev2, prev3);
                            }

                            break;
                        case 'DIALOG':
                            var id = cur.id;//local vars get hoisted...
                            ///if(id==='alert'||id==='model')propagate = Notify.clickHandler(ectx, view, cur, prev1, prev2, prev3);
                            //route picker click back to handler
                            if (id === 'timepicker') TimePicker.clickHandler(ectx, view, cur, prev1, prev2, prev3);
                            if (id === 'datepicker') DatePicker.clickHandler(ectx, view, cur, prev1, prev2, prev3);
                            if (id === 'colorpicker') ColorPicker.clickHandler(ectx, view, cur, prev1, prev2, prev3);
                            //why do we let clicks continue to bubble up from here?
                            ectx.stopPropagation = true;
                            break;
                        case 'PROGRESS':
                            var ln = cur.dataset.bindLoad;
                            if (ln) {
                                var ds = view.view.module.model.get(ln, 'list');
                                //cancel the load
                                ds.loading('ready');
                            } else {
                                if (cur.classList.contains('data-action')) {
                                    if (!ectx.defaultPrevented) {
                                        // div with action class
                                        // any element can now have an action, button actions are handled above this is only for non button elements
                                        ectx.stopPropagation = !Views.publishActionEvent(cur, ectx);
                                        //per gg any action now automatically causes propagation to stop with no dev control over this
                                        //but if this is nested inside a list we might still want to allow some actions like 'dismiss'
                                        //and with the 'dev option' being removed we have to try our best and make a guess at what is wanted
                                        ectx.defaultPrevented = true;
                                    }

                                }

                            }
                            ectx.stopPropagation = true;
                            break;
                        case 'NOTIFY-MSG':
                            //the default action is just to remove the notification
                            //but you can add more in the msg body as links or buttons and they will trigger
                            //you can also set overrides on the notify tag
                            var action = cur.dataset.action;
                            var collapsed = false;
                            if(action.includes('notify:expandable')){
                                collapsed = !cur.classList.toggle('collapsed');
                                ectx.defaultPrevented = true;
                                ectx.stopPropagation = true;
                            }

                            if(!collapsed && action.includes('notify:remove')){
                                ectx.defaultPrevented = true;
                                ectx.stopPropagation = true;
                                Views.fireActionEvent(cur, 'close');
                                if(cur.parentElement)cur.parentElement.removeChild(cur);
                            }
                            break;
                        case 'GRID-CELL':
                        case 'TD':
                        case 'DIV':
                        case 'H1':
                        case 'H2':
                        case 'H3':
                        case 'H4':
                        case 'H5':
                        case 'H6':
                            if (cur.classList.contains('data-action')) {
                                if (!ectx.defaultPrevented) {
                                    //we need to check for double click
                                    if(cur.dataset.dblclickAction){
                                        if(vt.dblclick){
                                            //yes we have a real double click
                                            vt.dblclick=clearTimeout(vt.dblclick);
                                            // div or header with action class
                                            // any element can now have an action, button actions are handled above this is only for non button elements
                                            ectx.stopPropagation = !Views.publishActionEvent(cur, ectx, '', '', cur.dataset.dblclickAction);
                                            // per gg any action now automatically causes propagation to stop with no dev control over this
                                            // but if this is nested inside a list we might still want to allow some actions like 'dismiss'
                                            // and with the 'dev option' being removed we have to try our best and make a guess at what is wanted
                                            ectx.defaultPrevented = true;
                                        }else{
                                            vt.dblclick = setTimeout(dblClickHandler, fastClickThreshold + 100, cur, ectx);
                                        }
                                    }else {
                                        // div or header with action class
                                        // any element can now have an action, button actions are handled above this is only for non button elements
                                        ectx.stopPropagation = !Views.publishActionEvent(cur, ectx);
                                        // per gg any action now automatically causes propagation to stop with no dev control over this
                                        // but if this is nested inside a list we might still want to allow some actions like 'dismiss'
                                        // and with the 'dev option' being removed we have to try our best and make a guess at what is wanted
                                        ectx.defaultPrevented = true;
                                    }
                                }
                            }
                            break;
                        case 'svg':
                            if(cur.classList.contains('ct-chart-bar')||cur.classList.contains('ct-chart-pie')||cur.classList.contains('ct-chart-line')) {
                                var chartElement = ectx.target.closest('[*|meta]');
                                var meta = chartElement.getAttributeNodeNS(Chartist.namespaces.ct, 'meta');
                                var value = chartElement.getAttributeNodeNS(Chartist.namespaces.ct, 'value');
                                var mv = meta.value;
                                var vv = value.value;
                                if(mv){
                                    var  d = Chartist.deserialize(mv);
                                }
                                ectx.stopPropagation = !Views.fireActionEvent(cur, 'chart-click', false, false, d);
                            }
                            break;
                        default:
                            if(cur.classList.contains('clickable')) {
                                ectx.stopPropagation = !Views.publishActionEvent(cur, ectx);
                                ectx.defaultPrevented = true;
                            }
                            break;
                    }
                }
            }
            //we handle propagation in our loop we don't need browser help here
            //ectx.stopImmediatePropagation();ectx.stopPropagation();
            //if defaultPrevented then ....
        }

        // target is what was clicked on
        // currentTarget is the target of the delegate selector
        //[FORM]------------------------------------------------------------------------------------------------------->
        var reDotAll=/\..*/;
        var controls = {
            lookup: 'Lookup',
            dropdown: 'Dropdown',
            selectlist: 'SelectList',
            time: 'TimePicker',
            date: 'DatePicker',
            color: 'ColorPicker'
        };
        /**
         * TODO:CLEANUP
         * @param {HTMLElement} el
         * @param {View} view
         * @param {ClickEventCtx} e
         */
        function inputHandler(el, view, e){
            var type = el.getAttribute('type');
            switch(type){
                case 'button':
                    //a button thats really a lookup control...
                    //find the real type and the input control (in case we clicked the helper button????)
                    type = el.dataset.type.replace(reDotAll,'');
                    window[controls[type]].showDropDown(el.parentElement.querySelector('input'));
                    break;
                case 'dropdown':
                    // dropdowns don't allow user 'input' and click just opens the dd display
                    Dropdown.showDropDown(el);
                    break;
                case 'radio':
                case 'checkbox':
                    if(el.dataset.action){
                        // checkboxes can specify an 'alt' value - a value if not checked
                        // if an action is specified (usually when in a toolbar) we set the key to the value or alt value
                        // so the later action handler can fire with the right key
                        el.dataset.actionKey = el.checked ? el.value : el.getAttribute('alt');
                    }
                    el.focus();
                    break;
            }
            e.stopPropagation = true;
        }

        /**
         * Input helper buttons (the dropdown buttons inside input controls)
         * we pass it to the appropriate controller and then kill the event.
         * the control is constructed so that the helper is placed just after the related input
         * @param {HTMLElement} el
         * @param {View} view
         * @param {ClickEventCtx} e
         */
        function helperClickHandler(el, view, e){
            var stopPropagation = true;
            var defaultPrevented = true;
            //find the input control
            var input = el.closest('.input-control').querySelector('input');
            // the 'type' on desktop is 'lookup' but on 'mobile' it is button
            // 'data-type' should always be 'lookup'
            var type = input.dataset.type;
            switch (type) {
                case 'cmtime':
                case 'time':
                    TimePicker.showDropDown(input);
                    break;
                case 'cmdate':
                case 'date':
                    DatePicker.showDropDown(input);
                    break;
                case 'cmcolor':
                case 'color':
                    ColorPicker.showDropDown(input);
                    break;
                case 'lookup':
                    Lookup.showDropDown(input);
                    break;
                case 'dropdown':
                    Dropdown.showDropDown(input);
                    break;
                case 'selectlist':
                    SelectList.showDropDown(input);
                    break;
                case 'search':
                    Search.showOptionPanel(input);
                    break;
                default:
                    //unknown or generic helper just pass it on
                    stopPropagation = false;
                    defaultPrevented = false;
                    break;
            }
            //kill it if we consumed it, it is of no further use
            e.defaultPrevented = defaultPrevented || e.defaultPrevented;
            e.stopPropagation = stopPropagation || e.stopPropagation;
        }
        //[TYPE]------------------------------------------------------------------------------------------------------->
        /**
         * bind delegate for buttons - ntg lower than a button
         * @param {HTMLElement} el
         * @param {HTMLElement} view
         * @param {ClickEventCtx} e
         */
        function buttonClickHandler (el, view, e) {
            // events are actually caught at the body level and don't 'bubble' using the OS
            // as such calling stopPropagation or stopImmediatePropagation on the event has no effect whatsoever
            //
            // we handle bubbling internally - returning false from this routine stops the event bubbling to other handlers
            // 'preventDefault' causes defaultPrevented to be set but allows the event to bubble to other handlers that will respect that
            // 'stopPropagation' prevent bubbling to other handlers
            // generally you want to allow propagation in order to check for a dismiss
            // but preventDefault to stop other handlers carrying out there own action
            // el should always be the button element (in theory target could be an image inside the button)

            // we allow multiple actions - space separated
            var actions = (el.dataset.action|| (el.hasAttribute('command') ? 'command' : '') ||'none').split(' ');
            var defaultPrevented = true;
            var stopPropagation = false;
            //'behavior' handler
            // if is marked as 'close' behavior process last
            var hasClose = el.classList.contains('close');
            // if the button has a 'toggle' behavior we toggle the state first
            var hasToggle = el.classList.contains('toggle');
            if(hasToggle){
                Buttons.toggle(el, undefined, true);
            }
            //used to toggle a style on a given target.  See button controller and toggleStyle method for more details.
            var hasToggleStyle = el.classList.contains('toggle-style');
            if (hasToggleStyle){
                Buttons.toggleStyle(el, view);
            }
            // the button could be in the main view or a panel or a dialog
            // need to be careful of any 'findView' call to make sure it returns the right view
            for(var i=0;i<actions.length;i++) {
                var aid = actions[i];
                switch (aid) {
                    case 'pagex':
                    case 'page':
                        //bind delegate for page buttons
                        // the target is a 4 optional part value aid:key:io:mode
                        //  aid: this is the actionID if not set looks for the same property on a parent element
                        //  key: some pages want us to pass in a key others don't
                        //       in some cases the current record key is used in other cases a fixed key is appropriate
                        //       by default we pass the current key but u can set the param to any value including empty
                        //   io: the init object i.e. {someprop:'greg',someotherprop:9}
                        // mode: set mode to 'base' to open a base page otherwise empty opens a child page or unique opens if not already open
                        // the io (3rd) param can contain a simple value or an object - if it is an object then it contains ':' which will get misinterpreted....
                        var tgt = ((el.dataset.target.replace(':','!!').replace(':','!!').replace(':base','!!base').replace(':unique','!!unique')) + '!!!!!!!!').split('!!');
                        //menu(pagex) uses the key for other things, button(page) looks for key on the el
                        var paid = tgt[0];
                        var k = aid==='pagex' ? $.stringToValueEx(tgt[1] || '') : $.stringToValueEx(tgt[1] || el.dataset.key) || 0;
                        var io = tgt[2] ? $.stringToValueEx(tgt[2]) : null;
                        var mode = tgt[3];

                        if (k === 0 && el.dataset.blockZeroKey === 'true') {
                            //do nothing
                        }else{
                            Views.openPage(new Action(paid, k, io), mode);
                        }
                        break;
                    case 'menu'://nc
                        //bind delegate for a popup button that opens a menu
                        var pel = el.closest('[data-view]');
                        Views.openMenu(pel, el.dataset.target, el);
                        break;
                    case 'popup':
                        //bind delegate for a generic popup button
                        Views.openPopup(view, el.dataset.target, el);
                        break;
                    case 'panel':
                        //if we are in a panel or popup get the main view...
                        if (view.view.type !== 'page')view = view.view.owner.el;
                        Views.openPanel(view, el.dataset.target, el.getAttribute('rel') || el);
                        break;
                    case 'dialog-modal':
                    case 'dialog':
                        var dialog = DialogManager.getDialog(el.dataset.target);
                        if(dialog.open){
                            dialog.close(el.hasAttribute('anchor') ? el : undefined);
                        }else{
                            dialog.show(el.hasAttribute('anchor') ? el : undefined);
                        }
                        break;
                    case 'more'://nc
                        //by the time we get here toggle will have clicked the state over to the one we want
                        var state = Forms.setDisplayGroupState(el.form||view, el.dataset.target, !!(+el.value));
                        var tgt = state ? el.dataset.focusMore : el.dataset.focusLess;
                        if (tgt) Focus.focusLater(view.querySelector('article[name="' + tgt + '"]') || view.querySelector('[name="' + tgt + '"]'));
                        Buttons.updateState(el, Views.findView(view).view.name);
                        break;
                    case 'flip':
                        var cl = el.parentElement.closest('.flip').querySelector('.flip-open');  //el.parentElement.parentElement.querySelector('.flip-open');
                        var op = el.parentElement.closest('.flip').querySelector('.flip-close');  //el.parentElement.parentElement.querySelector('.flip-open');
                        cl.classList.toggle('flip-open', false);
                        cl.classList.toggle('flip-close', true);

                        op.classList.toggle('flip-close', false);
                        op.classList.toggle('flip-open', true);
                        break;
                    case 'expandable':
                        var tgt = el.dataset.target.split(':');
                        var type = tgt[0];
                        var target = tgt[1];
                        switch(type) {
                            case 'root':
                                target = el.closest('.root');
                                break
                            case 'prev':
                                target = el.previousElementSibling;
                                break;
                            case 'next':
                                target = el.nextElementSibling;
                                break;
                            case 'header':
                                target = el.closest('header').nextElementSibling;
                                break;
                            case 'footer':
                                target = el.closest('footer').previousElementSibling;
                                break;
                            case 'parent':
                                target = el.parentElement;
                                if (target.parentElement.tagName === 'PAGE') {
                                } else {
                                    Array.from(target.parentElement.querySelectorAll('.expanded')).forEach(function (item) {
                                        if (item !== target) {
                                            //toggle it off
                                            item.classList.remove('expanded');
                                            //assuming first child is the button
                                            item.firstChild.classList.remove('active');
                                        }
                                    });
                                }
                                break;
                            default:
                                //was next section but we allow both section/ul/ol - so for now leave it open
                                target = (target) ? view.querySelector('#' + target) : el.closest('header').nextElementSibling;
                                break;
                        }
                        Buttons.toggle(el, target.classList.toggle('expanded'), false);
                        break;
                    case 'confirm'://stop
                        //bind delegate for standard confirm notifications
                        //pick up message/cancelable from tag
                        //convert returns into actions
                        var p = Notify.confirm(el.dataset.message, !!el.dataset.cancelable);
                        p.always(function (retval) {
                            el.setAttribute('data-action-key', retval);
                            Views.publishActionEvent(el, e);
                        });
                        stopPropagation = true;
                        break;
                    case 'download-filelist':
                        var href = el.dataset.link.split('.');
                        var m = Views.findView(view).view.module.model;
                        var list = m.get(href[0],'list');
                        var url = list[0][href[1]];
                        Views.openWindow(url, 'direct');
                        break;
                    case 'download':
                    case 'link':
                        //bind delegate for link buttons - if you set link type to 'direct' it does a fast direct download
                        var href = el.dataset.link;
                        var linktype = el.dataset.linkType || (aid === 'download' ? 'direct' : '_blank');
                        Views.openWindow(href, linktype);
                        break;
                    case 'print'://this
                        if(el.dataset.printOptions) {
                            //bind delegate for print buttons
                            //get the data from the button that was clicked...
                            Print.print(eval('(' + el.dataset.printOptions + ')'));
                        }
                        defaultPrevented = true;
                        break;
                    case 'copyErrLog':
                        // I tried a direct call to 'DBManager.exportStore('Loggr').done(function (data)' to
                        // pull the data directly but background framework processes ran between the call
                        // and data arrival which caused 'document.execCommand('copy')' (which does the actual
                        // copy to clipboard) to fail to execute so the copy did not happen.
                        var errMsg = '';
                        var arrError = [] = e.target.value.split("|");
                        if (arrError.length > 1) {
                            for (var i = 0; i < arrError.length - 1; i++) {
                                var arrFields = [] = arrError[i].split("-");
                                for (var j = 0; j < arrFields.length; j++) {
                                    switch (j) {
                                        case 0:
                                            errMsg += arrFields[j];    // timestamp
                                            break;
                                        case 1:
                                            errMsg += '\nMethod: ' + arrFields[j];   // methodName
                                            break;
                                        case 2:
                                            errMsg += '\nFile: ' + arrFields[j];    // fileName
                                            break;
                                        case 3:
                                            errMsg += '\nError: ' + arrFields[j];    // message
                                            break;
                                        case 4:
                                            errMsg += '\nLine: ' + arrFields[j];   // lineNumber
                                            break;
                                        case 5:
                                            errMsg += '\nColumn: : ' + arrFields[j] + '\n\n';  // columnNumber
                                            break;
                                    }
                                }
                                Clipboard.copyText(u.obs(errMsg, ''));
                                Notify.alert('The Error Log has been copied to your clipboard');
                            }
                        }
                        break;
                    case 'menu-submenu':
                        // open a submenu - we could be in a popup menu or in a toolbar menu - the submenu could be embedded or a floater
                        // floater call openMenu, embeds use css on toggle select
                        // In a popup we would then dismiss by default on the click - so prevent default just in case
                        var mi = el.closest('menuitem');
                        if(mi.dataset.cmdGrp){
                            //if a floater call open menu
                            Menus.openMenu(mi);
                        }else {
                            //find any previously selected mi and deselect it
                            var prev = mi.parentElement.querySelector('menuitem.selected');
                            if (prev && prev !== mi) prev.classList.toggle('selected');
                            var state = mi.classList.toggle('selected');
                        }
                        stopPropagation = true;
                        defaultPrevented = true;
                        break
                    case 'tab'://stop
                        //bind delegate for tabs
                        ListViews.showTab(view, el.dataset.target);
                        stopPropagation = true;
                        break;
                    case 'dialog-close':
                        //must be parented by a dialog - close it
                        var dialog = el.closest('DIALOG');
                        dialog.close(el.value);
                        break;
                    case 'shortcut-clear':
                        Key.setUserKey(el.getAttribute('data-key'), false);
                        break;
                    case 'shortcut-save':
                        Key.setUserKey(el.getAttribute('data-key'), true);
                        break;
                    case 'navScroller'://stop
                        //bind delegate for tabs
                        Nav.scrollNav(el.parentElement, el.dataset.dir);
                        stopPropagation = true;
                        break;
                    case 'attachment':
                        //clicked on a button to upload, need to pass a ref to the button
                        var rid = el.dataset.rid || (el.dataset.rid = $.guid++);
                        el.classList.add(rid);
                        FileUpload.showPicker(view, el.dataset.entity, el.dataset.fileSource, !(el.dataset.multiple==='false'), el.dataset.accept, rid);
                        break;
                    case 'file-retry':
                        var key = +el.dataset.key;
                        FileUpload.retry(key);
                        break;
                    case 'file-delete':
                        var key = +el.dataset.key;
                        Notify.confirm('Are you sure you want to delete this attachment').done(function (val) {
                            FileUpload.cancel(key);
                        });
                        break;
                    case 'clear':
                        //find the related input control
                        var input = el.parentElement.querySelector('input');
                        if (input) {
                            input.value = '';
                            input.focus();
                        }
                        break;
                    case 'command':
                        var cmd = el.getAttribute('command');
                        switch (cmd) {
                            case 'fontName':
                                //update the display...
                                el.closest('dialog').info.anchor.innerHTML = el.innerHTML;
                                TextEditor.execCommand(cmd, false, el.getAttribute('face'));
                                break;
                            case 'fontSize':
                                TextEditor.execCommand(cmd, false, el.getAttribute('size'));
                                break;
                        }
                        break;
                    case 'cmd':
                        TextEditor.execCommand(el.dataset.cmd, false, el.value);
                        defaultPrevented = true;
                        break;
                    case 'expandAll'://not used
                    case 'collapseAll'://not used
                    case 'selectAll'://lookup lists
                    case 'unselectAll'://lookup lists
                    case 'toggleAll'://in wjadmin this is used to expand/collapse everywhere else it select/unselect
                    case 'tagAll':
                    case 'untagAll':
                        //the expandAll/collapseAll action would go on a button or icons to close or open all nodes
                        //the selectAll/unselectAll action would go on a button or icons to check or uncheck all nodes
                        //toggleAll can be used in either mode and is often used on a toggle button
                        var lv;
                        var state = $.stringToValue(el.value);
                        var target = el.dataset.target;
                        if (target === 'header') {
                            var header = el.closest('HEADER');
                            lv = header.nextSibling;
                        } else if (target === 'footer') {
                            var footer = el.closest('FOOTER');
                            lv = footer.previousSibling;
                        } else {
                            lv = view.view.el.querySelector('#' + target);
                        }
                        // work out the action state
                        // if this was toggle button the button toggle code will have set the value into its correct state before we get here
                        var t = aid === 'toggleAll' ? state : (aid === 'expandAll' || aid === 'selectAll' || aid === 'tagAll');

                        // in tag mode (select/unselect) we usually only update the UI not the backing data
                        // ie you would use it to toggle checkboxes in a form checkbox list
                        // but optionally it can be made to toggle the data behind the list
                        // find the target list action to see if it's in 'tag' mode
                        var mode = (aid === 'tagAll' || aid === 'untagAll') ? 'tag' : lv.dataset.listAction || 'none';
                        if(mode.indexOf('tag')!==-1){
                            var m = view.view.module.model;
                            var n = lv.dataset.source;
                            var ds = m[n];
                            t = ds.toggleField(t);//assumption that we will handle via the control and a data event
                        } else ListViews.toggleList(null, lv, t, el.dataset.mode || 'select');

                        //stop this propagating to some other list higher up if we are nested...
                        stopPropagation = true;
                        break;
                    case "addNewRow"://stop-noblock
                        //clicked on the insert row button...
                        var gid = el.dataset.target;
                        var grid = gid ? view.view.grids[gid] : el.closest('datagrid');
                        DataGrid.addNewRow(grid);
                        stopPropagation = true;
                        break;
                    case 'save-data':
                        //any active picker/lookup is already closed...
                        //if there is an active balloon with a form then save it and close
                        //if there is a grid or list with an active form save and close it
                        //or if there is a form then save it...
                        //find all forms on this page - find which ds they are bound to and save each one.....
                        //we should ensure that all active forms are closed before we trigger the save
                        Forms.autocloseHandler(e, el, view);
                        //find the related list and call GMI forcing the save event on each
                        var lists = el.dataset.bindSave.split(',');
                        for (var i = 0; i < lists.length; i++)view.view.module.model[lists[i]].getModifiedItems(true);
                        break;
                    case 'itemInsert:before':
                    case 'itemInsert:after':
                    case 'itemInsert:start':
                    case 'itemInsert:end':
                    case 'itemInsert':
                        var ds = view.view.module.model[el.dataset.source];
                        //set data-rel-key to key of the item the pos is related to
                        ds.addItem(null, aid==='itemInsert' ? '' : aid.split(':').pop(), el.dataset.relKey);
                        //get the panel
                        var pid = el.dataset.target;
                        if (pid){
                            //if we are in a panel or popup get the main view...
                            if (view.view.type !== 'page')view = view.view.owner.el;
                            Views.openPanel(view, pid, el.getAttribute('rel') || el);
                        }
                        break;
                    case 'form-delete':
                        var form = el.form || view.querySelector('form');
                        Forms.formDelete(form);
                        break;
                    case 'form-state':
                        //not a real button so we need to find the form
                        Forms.setFormState((el.tagName === 'ICON') ? $.findFirstElementByTagName(el, 'FORM') : el.form, el.dataset.formState);
                        break;
                    case 'pickColor':
                        ColorPicker.showDropDown(el);
                        break;
                    case 'pickDay':
                        DatePicker.showDropDown(el);
                        break;
                    case 'next':
                        var list = el.dataset.target||el.dataset.bindNext;
                        Views.findView(el).view.module.model[list].next();
                        break;
                    case 'prev':
                        var list = el.dataset.target||el.dataset.bindPrev;
                        Views.findView(el).view.module.model[list].prev();
                        break;
                    case 'template':
                        //switches the active template on the target
                        var lv = view.querySelector('#'+el.dataset.target);
                        //set the new template into the lv.....
                        lv.ctx.tmpl = el.dataset.tmpl;
                        break;
                    case 'togglefilter':
                        // toggle the current filter on a DS
                        // normally toggle filter would be set on a toggle button
                        // and so the button text would change to reflect the state
                        // but now we want to have two buttons like tabs and have them switch 'active'
                        var dsName = el.dataset.source;
                        var ds = Views.findView(el).view.module.model[dsName];
                        var state = el.dataset.state;
                        //if had state then its a tab rather than toggle
                        if(state){
                            state = ds.toggleFilter(state==='true');
                            //there are two buttons this on is now active and the other isnt.....
                            //either prevSibling or nextSibiling
                            var lastActive = el.parentElement.querySelector('[data-action*="togglefilter"].active');
                            if(lastActive)lastActive.classList.remove('active');
                            el.classList.add('active');
                        }else{
                            state = ds.toggleFilter();
                            el.value = state ? '1' : '0';
                        }
                        break;
                    case 'search':
                        // could be a straight search button or a form search/save button
                        var ctl = el.previousElementSibling;
                        if (ctl.classList.contains('search')) {
                            // a std button adjacent to a search input box
                            var inp = ctl.querySelector('input');
                            // get search 'key' value from the input box
                            var key = inp.value;
                            //update the button 'key' value - used by the showOptionsPanel
                            el.setAttribute('data-key', key);
                            //the real input control?
                            var rel = el.closest('.input-control').querySelector('input');
                            Search.showOptionPanel(rel);
                        }else {
                            //does this exist ?????
                            var rel = el.closest('.input-control').querySelector('input');
                            Search.showOptionPanel(rel);
                        }
                        break;
                    case 'feedback-close':
                        var uid;
                        if (el.parentElement.getAttribute('id')){
                            uid = Views.findView(view).view.name + '-' + el.parentElement.getAttribute('id');
                        }else{
                            uid = Views.findView(view).view.name;
                        }
                        Feedback.close(el, uid);
                        break;
                    case 'skipIntro':
                        System.saveSession({ skipIntro: 1 });
                        System.user.skipIntro = 1;
                        DBManager.saveCachedObject('user', System.user);
                        var dialog = el.closest('DIALOG');
                        dialog.close(el.value);
                        break;
                    case 'markSysMessageRead':
                        Remote.remoteDataCall("WJAPP.Framework.Desktop", 'markReadSystemMessages');
                        var dialog = el.closest('DIALOG');
                        dialog.close(el.value);
                        break;
                    case 'showIntro':
                        Notify.showDialog({m:Template.getTemplate('root.Desktop.intro'), type:'curtain', size:'full'});
                        var dialog = el.closest('DIALOG');
                        dialog.close(el.value);
                        break;
                    case 'startDemo':
                        var dialog = el.closest('DIALOG');
                        dialog.close(el.value);
                        if (WalkMeAPI) {
                            WalkMeAPI.startWalkthruById(270188);
                        }
                        break;
                    case 'search-options':
                        // bind delegate for search buttons
                        // find the related input control
                        var rel = el.parentElement.getElementsByTagName('input').item(0);
                        Search.showOptionPanel(rel);
                        break;
                    case 'gantt-zoom':
                        var level=el.value;
                        var target = view.querySelector('#'+el.dataset.target);
                        Gantt.zoom(level, target.ctx);
                        break;
                    case 'range-actual':
                        var range = el.parentElement.children[1];
                        range.value = parseInt(range.dataset.reset);
                        range.dispatchEvent(new CustomEvent('change', {bubbles: true}));
                        stopPropagation = true;
                        break;
                    case 'range-inc':
                        var range = el.parentElement.children[1];
                        range.value = parseInt(range.value) + parseInt(range.dataset.step);
                        range.dispatchEvent(new CustomEvent('change', {bubbles: true}));
                        stopPropagation = true;
                        break;
                    case 'range-dec':
                        var range = el.parentElement.children[1];
                        range.value -= parseInt(range.dataset.step);
                        range.dispatchEvent(new CustomEvent('change', {bubbles: true}));
                        stopPropagation = true;
                        break;
                    case 'cancel'://stop
                        // no form target so we have to find one....
                        // normally we can assume that the parent view (even an panel) contains a form
                        // but panels in an accordion are not views so we will end up at the root an then locate the wrong form
                        //in those case the dev should add the form attribute
                        var form = el.form || view.view.forms[el.dataset.form];
                        if (!form) {
                            // didnt declare a form so this is multiform use
                            // find which one it applies to - if its just one on the page then they shud specify it...
                            // so we assume thats not the case
                            // if they wanted a custom action and save all at ones they can specify that
                            // so assume it is in a viewstack
                            // find closest page or panel then query select a viewstack then find active tab and find form
                            var tab = el.closest('panel, page').querySelector('tab.active,tab.open');
                            form = (tab || view).querySelector('form');
                        }

                        if (form) Forms.cancel(form);
                        stopPropagation = true;
                        break;
                    case 'submit':
                    case 'save'://stop
                        var form = el.form || view.view.forms[el.dataset.form];
                        if (form) {
                            //submit or save with the default form action.
                            Forms.save(form, '', '', true);
                        }else{
                            // didnt declare a form so this is multiform use
                            // find which one it applies to - if its just one on the page then they shud specify it...
                            // so we assume thats not the case
                            // if they wanted a custom action and save all at ones they can specify that
                            // so assume it is in a viewstack
                            // find closest page or panel then query select a viewstack then find active tab and find form
                            var tab = el.closest('panel, page').querySelector('tab.active,tab.open');
                            if(el.dataset.form==='all'){
                                Forms.saveAll(tab, view);
                            }else {
                                form = (tab || view).querySelector('form');
                                //submit or save with the default form action.
                                if (form) Forms.save(form, '', '', true);
                            }
                        }
                        stopPropagation = true;
                        break;
                    case 'none':
                        //button should have a data-action or data-list-action
                        if(el.dataset.listAction){
                            //ignore it but allow it to propagate and don't prevent default
                            defaultPrevented = stopPropagation = false;
                        }else{
                            //dont ignore it .. kill everything
                            defaultPrevented = stopPropagation = false;
                        }
                        break;
                    case 'toggle-class':
                        var state = false;
                        switch(el.dataset.target){
                            case 'grandparent':
                                //used by texteditor
                                state = el.parentElement.parentElement.classList.toggle(el.value);
                                break;
                            case 'parent':
                                state = el.parentElement.classList.toggle(el.value);
                                break;
                            default:
                                var nodes = view.querySelectorAll(el.dataset.target);
                                if(nodes) {
                                    for (var j = 0, len = nodes.length; j < len; j++) {
                                        state = nodes[j].classList.toggle(el.value);
                                    }
                                    //state = el.closest(el.dataset.target).classList.toggle(el.value);
                                }
                                break;
                        }
                        if(el.classList.contains('active-on-toggle'))el.classList.toggle('active', state);
                        break;
                    case 'tag-click':
                        Views.publishActionEvent(el, e, 'dialog:selectlist', '', aid);
                        break;
                    default:
                        //unknown aid
                        if(aid.indexOf('submit:')===0) {
                            //find the target form - either it has a form prop  or it has an attribute or data attribute or get the first form in the view
                            var form = Forms.getTargetForm(el, view);
                            //find the action
                            var types = aid.split(':');

                            if (types[1] === 'custom') {
                                //no submit just commit active element and fire custom action
                                Forms.commitActiveElement(form);
                                defaultPrevented = stopPropagation = Views.publishActionEvent(el, e, '', el.id, types[2]);
                            } else {
                                //std - action is the submit action
                                Forms.save(form, types[1]);
                            }
                            //should we stopPropagation ? Calling save should handle close and what else is there higher up
                        }else if(aid.indexOf('date:')===0) {
                            //datepicker specific click
                            defaultPrevented = stopPropagation = DatePicker.clickHandler(e, view, aid);
                        }else if(aid.indexOf('grid:')===0) {
                            //grid specific click
                            defaultPrevented = stopPropagation = DataGrid.clickHandler(e, view, aid);
                        }else{
                            // it has to be a custom action
                            // return true to stopPropagation and preventDefault
                            // or false to allow default..
                            // by default returns true - does anyone ever return false...
                            defaultPrevented = stopPropagation = Views.publishActionEvent(el, e, '', el.id, aid);
                        }//stop
                        stopPropagation=false;
                        break;
                }
            }

            //close is now a behavior rather than action
            if(hasClose){
                //for standalone 'close' buttons  -  if it's in a list just set the attr DO NOT SET THE CLOSE CLASS
                //close buttons close what they are in
                Views.dismiss(el.dataset.dismiss||'auto', el.dataset.target||el, false, e);
            }
            //if touch - set RAF task for active on - set timeout/RAF Task for active off
            if ($.support.touch) {
                System.addRenderTask(setButtonActive, el, true);
            }
            // per gg any action now automatically causes propagation to stop with no dev control over this
            // but this wold kill the dismiss action on popup menu's - so really we want to prevent the default
            // action but allow propagation
            e.defaultPrevented = defaultPrevented || e.defaultPrevented;
            e.stopPropagation = stopPropagation || e.stopPropagation;
        }

        /**
         * Used in mobile to allow buttons to reflect there depresssed state
         * @param el
         * @param state
         */
        function setButtonActive(el, state) {
            el.classList.toggle('button-active', state);
            if (state) System.addRenderNextTask(setButtonActive, el, false);
        }
        //[LIST RELATED]----------------------------------------------------------------------------------------------->
        /**
         * target is what was clicked on - section is hdr/ftr/body(ie div or ul)
         * Could click inside a ul in user content - dont select unless there is a key
         * Could click in a HDList (attachments), it has a key but don't select
         * We need a way to determine whether we stop at this node or bubble
         * Basically if we have a selectAction, or an expandAction or a listAction then we stop unless explicitly  falggged to do so
         * /@ param {Event} e
         * /@ param {HTMLElement} view
         * /@ param {HTMLElement} target
         * /@ param {HTMLOListElement|HTMLUListElement} ol
         * /@ param {HTMLLIElement} li
         * /@ param {HTMLElement} section
         * @param {ClickEventCtx} ectx
         */
        function listClickHandler (ectx){
            /** @type {HTMLOListElement|HTMLUListElement} */
            var ol = ectx.currentTarget;
            /** @type {HTMLLIElement} */
            var li = ectx.prev1;
            /** @type {HTMLElement} */
            var section = ectx.prev2;
            /** @type {Event} */
            var e = ectx.e;
            /** @type {HTMLElement} */
            var view = ectx.view;
            /** @type {HTMLElement} */
            var target = ectx.target;
            if(li) {
                // BASIC RULES
                // preventDefault blocks the list action but allows select,
                // anything from a footer or header blocks the dismiss
                // if has the list-group-hdr class this blocks everything
                // ul are always nested in ol unless in 'ignore' mode
                // set preventDefault to preventDefault, set stopPropagation to stop propagation
                var defaultPrevented = ectx.defaultPrevented;
                var stopPropagation = false;
                // could be ol/ul
                var tagName = ol.tagName;
                //the actual bound element (LISTVIEW)
                //a ul should always be parented by an li - most likely a child of an ol but could be nested deeper
                var root = tagName === 'LISTVIEW' ? ol : ol.closest('LISTVIEW');
                //if we are not in an LISTVIEW or we are in an LISTVIEW that is user content then skip but still bubble
                if(!root || !root.ctx) return false;
                //the dismiss action - if just expanding/collapsing or toggling or clicked a button it shouldn't trigger a dismiss but other actions will
                var dismiss = defaultPrevented ? false : $.getFirstElementAttrValue(target, 'data-dismiss', ol, '');
                //data - model/ds/key/item
                var dsName = root.dataset.source;
                var model, ds, nodeKey, item;
                //it is possible on static jumplists to have a fixed key even if there is no ds
                //its very likely its on the root LISTVIEW but I guess there is no reason you couldn't also put it on an LI
                //its up to the dev to ensure the chosen listAction doesn't require an item
                var keyNode = $.findFirstElementByAttr(target, 'data-key', ol);
                var key = keyNode ? keyNode.dataset.key : '';

                //sometimes the key is not the pk so we fallback to the nodeKey
                nodeKey = li.dataset.nodeKey || key;

                //get the ds and item
                if (dsName) {
                    model = Views.findView(root).view.module.model;
                    ds = model.get(dsName, 'list');
                    //thing is this - if we are an HD source then the key is prb a valid record in the ds but if not an HD then the key is in a child of the top level item
                    item = nodeKey ? ds.findByKey(nodeKey) : null;
                }
                // BEHAVIOR
                // select mode options: multipart value
                // 1 sync-selected node/row/both/none
                // 1.1 sync-node > row - true/false selecting the html node (by user) selects the ds row
                // 1.2 sync-row > node - true/false selecting the ds row selects the html node
                //     both/none enable/disable both behaviors
                // 2 select-elements path/node
                //     select-path - select all nodes from the leaf node to the root node
                //     select-node - selects just the current leaf node
                // 3 close-behavior force/allow/both/check/toggle/radio/accordion/both
                // 3.1 close-force opening a node forces other nodes to close (like accordion or radio buttons)
                // 3.2 close-allow can an open node be explicitly closed
                //     both/none enable/disable both behaviors
                // 4 deselect behavior
                // 5 notify notify/silent
                //     notify itemSelect event fires to notify the module
                //     silent no event fires
                // 6 action-special save
                //     action-save the old row is saved when closed
                //
                // presets
                // auto = node:path:allow:notify:none
                // save = node:node:both:notify:save
                // none = none:path:allow:silent:none
                // check if we need to select this node
                var ctx = root.ctx;

                // header body footer - selected unselected - opened closed - action
                // if header/body/footer always select and only exception header already open - close and deselect
                // so really we are looking for the edge case isHeader+opened+selected node+
                var allowClose = ctx.options.allowClose;
                var allowDeselect = ctx.options.allowDeselect;

                var isHeader   = !!section && section.tagName === 'HEADER';
                var isFooter   = !!section && section.tagName === 'FOOTER';
                var isContent  = !isHeader && !isFooter;
                var isTopLevel = li.parentElement === root;
                var isButton   = (target.tagName === 'BUTTON' || target.tagName === 'ICON');

                var isBranch   = li.firstChild.tagName==='HEADER';
                // is this particular node expanded or collapsed
                // well if explicitly expanded or collapsed then we have the answer but if neither is present then....
                // only branch nodes can expand or collapse
                // and we define a 'branch' to be anything with a header as the first child (RULE)
                // you dont need to mark a node if it is in its default state
                // and we only support one 'default' node state - that is 'expanded'
                // so if it isn't explicitly collapsed we assume it is expanded by default (RULE)
                // so explicit trumps everything, otherwise non-branch is always collapsed, and branch is by default expanded
                var isExpanded = li.classList.contains('collapsed') ? false : li.classList.contains('expanded') ? true : isBranch;

                var isSelected = li.classList.contains('selected') && li.classList.contains('node');

                var isLocked   = li.classList.contains('locked');

                var expandAction,selectAction;
                //if the list is marked none selectable then skip it, this is more important for flagging up HDList's
                if(root.dataset.selectMode==='none' || li.dataset.selectMode==="none" || !nodeKey){
                    // determine if we want to close or open the node
                    expandAction = isHeader && isExpanded && allowClose && !isButton && !isLocked ? 'close' : isHeader && !isExpanded && !isButton && !isLocked ? 'open' : 'none';
                    selectAction='none';
                }else{
                    // determine if we want to close or open the node
                    expandAction =
                        isHeader && isExpanded && isSelected && allowClose && !isButton && !isLocked ? 'close' :
                            isHeader && isExpanded && !isSelected && allowClose && !isButton && !isLocked ? 'close' :
                                isHeader && !isExpanded && !isButton && !isLocked ? 'open' :
                                    'none';
                    // if a header is opened it selects if closed it deslects
                    selectAction =
                        expandAction === 'open' && !isSelected ? 'select' :
                            expandAction === 'close' && !isSelected ? 'select' :
                                expandAction === 'close' && isSelected && allowDeselect ? 'deselect' :
                                    expandAction === 'none' && !isSelected ? 'select' :
                                        !isHeader && !isSelected ? 'select' :
                                            !isHeader && isSelected && allowDeselect ? 'deselect' :
                                                'none';
                }
                //exception - if its a header and selected and can deselect and cant close and top level
                if(isHeader && isExpanded && isSelected && !allowClose && !isButton && allowDeselect && isTopLevel){
                    selectAction = 'deselect';
                }
                if(isHeader && isExpanded && isButton && target.dataset.listAction==='close'){
                    //expandAction = 'close';
                    ListViews.toggleListItem(view, li, 'force-close', item, root);
                    stopPropagation = defaultPrevented = true;
                }
                // if a header is locked it can't close but it still deselects
                // can you open a node that has no children ?
                // well thats the problem you can but you cant
                //the way they are being constructed they do have toggle on none branch nodes
                //but a none branch node then could still open/close/select/deselect
                if(isLocked || !isBranch)expandAction='none';
                //DATA SELECT
                switch (selectAction) {
                    case 'select':
                        // if this IS NOT the currently selected node - see if we need/must select it
                        // 'selected', 'open' and 'tagged' are different things!!!!!
                        // we need root, key, item, mode
                        // check for no select option
                        if (ctx.options.syncNode) {
                            // MUST select the node and the underlying data
                            // so we need to sync the bound data list...
                            // if there is an editable form and it is set to autosave then it will validate and save before returning
                            // there is probably a key type mismatch (alpha/numeric) so soft
                            if (nodeKey == (ds.selectedKey = nodeKey)) {
                                //get the key type right....
                                nodeKey = ds.selectedKey;
                                if (ctx.options.selectNodes) {
                                    //we selected a new node so notify module
                                    if (ctx.options.notify)Views.fireActionEvent(root, 'itemSelect', false, false, item);
                                    if (!ctx.options.syncRow) {
                                        //if sync  row is enabled then selecting the db will have triggered the update
                                        //remove any previous selection
                                        var si = root.querySelectorAll('li.selected');
                                        for (var i = 0, len = si.length; i < len; i++) {
                                            si[i].classList.remove('selected', 'root', 'node');
                                        }
                                        if (ctx.options.selectNodes === 'node') {
                                            li.classList.add('selected', 'node');
                                        } else {
                                            $.selectPath(e.target, root, 'LI', true);
                                        }
                                    }
                                }
                            } else {
                                //we failed to sync the ds probably due to a form validation failure blocking the move to a new dataitem
                                //so abort
                                expandAction='none';
                                stopPropagation = defaultPrevented = false;
                            }
                        } else {
                            //MUST select the node but not the underlying data
                            if (ctx.options.selectNodes) {
                                // so someone decides to create a list with no ds i.e. manually building each item and no selected sync,
                                // but they still want it to show 'selected' - if they didn't they would have set data-select-mode="none" wouldn't they
                                // but no ds means no data..... we only have the key
                                if (ctx.options.notify)Views.fireActionEvent(root, 'itemSelect', false, false, key);
                                //clearly can't have syncRow set as there is no ds...
                                //remove any previous selection
                                var si = root.querySelectorAll('li.selected');
                                for (var i = 0, len = si.length; i < len; i++) {
                                    si[i].classList.remove('selected', 'root', 'node');
                                }
                                if (ctx.options.selectNodes === 'node') {
                                    li.classList.add('selected', 'node');
                                } else {
                                    $.selectPath(e.target, root, 'LI', true);
                                }
                            }
                        }
                        break;
                    case 'deselect':
                        // if this IS the currently selected node - see if we can deselect it
                        // it can only be deselected if explicitly allowed and is a branch i.e. it has no children
                        // becuase children would imply open/close behavior rather than a deselect....
                        // but if this is heirarchical should it select parent
                        //if select mode is path and its tree,group or groupTree then
                        var parent = window.DataManager.getParent(ds, ds.selectedItem);
                        //select the parent
                        ds.selectedItem = parent;
                        // set the selected key to invalid
                        ///ds.selectedKey = 0;
                        // key goes undefined but should it be 0???
                        if (!ds.selectedKey) {
                            //we set the selected node to null so notify module and pass it the old node for reference
                            if (ctx.options.notify)Views.fireActionEvent(root, 'itemSelect', false, false, null, item);
                            if (!ctx.options.syncRow) {
                                // if sync  row is enabled then selecting the db will have triggered a UI update
                                // but if not we must manually force it and remove any previous selection
                                var si = root.querySelectorAll('li.selected');
                                for (var i = 0, len = si.length; i < len; i++) {
                                    si[i].classList.remove('selected', 'root', 'node');
                                }
                            }
                        } else {
                            // we failed to sync the ds probably due to a form validation failure
                            // blocking the move to a new dataitem so just abort
                            expandAction='none';
                            stopPropagation = defaultPrevented = false;
                        }
                        break;
                    case 'none':
                    default:
                        break;
                }

                //NODE EXPAND
                if(expandAction==='open'&&!isExpanded){
                    ListViews.toggleListItem(view, li, true, item, root);
                    stopPropagation = defaultPrevented = true;
                }
                if(expandAction==='close'&&isExpanded){
                    ListViews.toggleListItem(view, li, false, item, root);
                    stopPropagation = defaultPrevented = true;
                }
                //NODE ACTION
                if (defaultPrevented) {
                    //buttons call default prevented but dont stop propagation so we can select
                    stopPropagation = defaultPrevented = true;
                }
                // if section is a header that is collapsible
                // and not defaultPrevented
                // and target is not a button then it's  a toggle
                //ok fire the action
                var actionNode = $.findFirstElementByAttr(target, 'data-list-action', ol);
                var action = actionNode ? actionNode.dataset.listAction : '';
                if (!stopPropagation && !defaultPrevented && action) {
                    //select if not selected?
                    //ACTIONS
                    // various list specific actions.... could be on the target Element or ListItem or List....
                    // you SHOULD place a default listAction on the 'ol' but this can be overridden on any sub-element
                    // if the default action has been prevented then ignore all list actions as they are inherently 'the default action'
                    // the action to perform
                    var actions = (action || 'noop').split(' ');
                    for (var i = 0; i < actions.length; i++) {
                        var action = actions[i];
                        //now do the primary action
                        switch (action) {
                            case 'file':
                                if (target.classList.contains('file-upload-retry')) {
                                    FileUpload.retry(key);
                                } else {
                                    Notify.confirm('Are you sure you want to delete this attachment').done(function (val) {
                                        FileUpload.cancel(key);
                                    });
                                }
                                break;
                            case 'itemDelete'://currently only used in deliverables
                                Notify.confirm('Are you sure you want to delete this item?').done(function(val){
                                    ds.removeItem(item);
                                });
                                break;
                            case 'itemInsert:start':
                            case 'itemInsert:end':
                            case 'itemInsert:before':
                            case 'itemInsert:after':
                            case 'itemInsert'://currently only used in deliverables
                                ds.addItem(null, action === 'itemInsert' ? '' : action.split(':').pop(), item);
                                //get the panel
                                var pid = $.getFirstElementAttrValue(target, 'data-target', ol);
                                if (pid) {
                                    //if we are in a panel or popup get the main view...
                                    if (view.view.type !== 'page')view = view.view.owner.el;
                                    Views.openPanel(view, pid, target.getAttribute('rel') || target);
                                }
                                break;
                            case 'itemEdit'://currently only used in deliverables
                                //tell the ds we want to edit the current record
                                ///ds.setEditing(true);
                                //if there is a panel specified open it
                                var pid = $.getFirstElementAttrValue(target, 'data-target', ol);
                                if (pid) {
                                    //if we are in a panel or popup get the main view...
                                    if (view.view.type !== 'page')view = view.view.owner.el;
                                    Views.openPanel(view, pid, target.getAttribute('rel') || target);
                                } else {
                                    //TODO: no target so must mean???
                                }
                                break;
                            case 'tag':
                                ds.toggleField(undefined, key);
                                break;
                            case 'page':
                                // the target is a 4 optional part value aid:key:io:mode
                                //  aid: this is the actionID if not set looks for the same property on a parent element
                                //  key: some pages want us to pass in a key others don't
                                //       in some cases the current record key is used in other cases a fixed key is appropriate
                                //       by default we pass the current key but u can set the param to any value including empty
                                //   io: the init object i.e. {someprop:'greg',someotherprop:9}
                                // mode: normally 'page' opens a child page but for the hub we want too replace the base page
                                //       set mode to 'base' to open a base page otherwise empty open a child page or 'unique' opens only if not already open
                                // the io (3rd) param can contain a simple value or an object - if it is an object then it contains ':' which will get misinterpreted....
                                var tgt = (($.getFirstElementAttrValue(target, 'data-target', ol).replace(':', '!!').replace(':', '!!').replace(':base', '!!base').replace(':unique', '!!unique').replace(':replace', '!!replace')) + '!!!!!!!!').split('!!');
                                var aid = tgt[0] ? tgt[0] : $.getFirstElementAttrValue(target.parentElement, 'data-target', ol);
                                var k = tgt[1] ? tgt[1] : key;
                                var io = tgt[2] ? $.stringToValueEx(tgt[2]) : null;
                                var mode = tgt[3];
                                Views.openPage(new Action(aid, k, io), mode);
                                break;
                            case 'panel':
                                var view = Views.findView(ol);
                                //if we are in a panel or popup get the main view...
                                if (view.view.type !== 'page')view = view.view.owner.el;
                                var pid = $.getFirstElementAttrValue(target, 'data-target', ol);
                                Views.openPanel(view, pid, target.getAttribute('rel') || target);
                                break;
                            case 'download':
                            case 'link':
                                var tgtNode = $.findFirstElementByAttr(target, 'data-link', ol);
                                var href = tgtNode.dataset.link;
                                var type = $.getFirstElementAttrValue(target, 'data-link-type', ol);
                                //if the href is expecting a param or if href came of the list it needs a key so add it
                                if (tgtNode === root || tgtNode === ol || href[href.length - 1] === '=')href += key;
                                Views.openWindow(href, type);
                                break;
                            case 'downloadAttachment':
                                // the key will be the file AttachmentKey in the attachments list
                                // but remember the ds is actually the root ds
                                // so the key may not be valid if the attachment list is in a nested child array
                                if (item && item.Path && item.FileName) {
                                    //if item and FileName are set then item points to the attachment or file item
                                    u.downloadAttachment(key, System.webdavInit, item.Path, item.FileName);
                                } else {
                                    // otherwise attachment list is a child array of a property of the current parent item
                                    var parentKey = $.getFirstElementAttrValue(ol, 'data-key', root);
                                    var parentItem = ds.findByKey(parentKey);
                                    var keyPath = $.getFirstElementAttrValue(ol, 'data-key-path');
                                    var p = keyPath.split(':');
                                    var childList = parentItem[p[0]]
                                    var index = $.getFirstElementAttrValue(target, 'data-index');
                                    var childItem = childList[index];
                                    u.downloadAttachment(key, System.webdavInit, childItem.Path, childItem.FileName);
                                }
                                break;
                            case 'itemToggle':
                                //equiv of toggle the header - but this can only ever be to close the expanded node (because a header would open it)
                                ListViews.toggleListItem(view, li, false, root, item);
                                break;
                            case 'tab':
                                //bind delegate for tabs
                                ListViews.showTab(view, $.getFirstElementAttrValue(target, 'data-target'));
                                break;
                            case 'select':
                            case 'none':
                            case 'noop':
                                break;
                            default:
                                //generic/unknown just triggers an action event
                                var ae = {
                                    target: e.target,
                                    currentTarget: ol,
                                    list: ol,
                                    listitem: li,
                                    detail: item,
                                    ctrlKey: e.ctrlKey || e.metaKey
                                };
                                Views.publishActionEvent(ae.target, ae, '', ol.dataset.id||ol.id||root.dataset.id||root.id, action, key);
                                break;
                        }
                        //stop this propagating to some other list higher up if we are nested...
                        stopPropagation = true;
                    }
                }//end of non blocked ACTIONS
                //LIST DISMISS
                //default prevented generally blocks the dismiss
                if (dismiss) {
                    if (isHeader||isFooter) {
                        // the header and footer clicks always block the default action - regardless
                    }else {
                        // bind delegate for close buttons
                        // dismiss -but what?? --
                        // form has its own handler for this -
                        // but a list close is multilevel
                        // it can close/end edit in a sub list etc
                        // if force was set remove it when passing to handler
                        Views.dismiss(dismiss.replace('-force', ''), ol, false, e);
                    }
                }
                // you can explicitly mark the ol/ul to stop or continue by setting data-propagate="true/false"
                // if this is not an LISTVIEW but it has a selectAction then dont propagate
                // if this is not an LISTVIEW and there is no selectAction but there is an expandAction then dont propagate.
                // if this is an LISTVIEW there is nothing higher so we can abort but UL/MENU could be nested.
                if(ol.dataset.propagate){
                    stopPropagation = stopPropagation || (ol.dataset.propagate==='false');
                }else{
                    if(!stopPropagation){
                        stopPropagation = ol.tagName === 'LISTVIEW';
                        if(!stopPropagation) {
                            stopPropagation = (selectAction && selectAction !=='none');
                            if(!stopPropagation) {
                                stopPropagation = (expandAction && expandAction !=='none');
                            }
                        }
                    }
                }
                if(stopPropagation) {
                    //stop this propagating to some other list higher up if we are nested...
                    ectx.stopPropagation = true;
                    ectx.defaultPrevented = true;
                }
            }else{
                // some lists are using 'margin' i.e. space between the LI's rather than using padding
                // this means it is possible to click into the 'space' in between the listitems
                // so we just have to ignore those clicks.......
                // if you don't like it don't put margins in your list!!!!!
                ectx.defaultPrevented = true;
                ectx.stopPropagation = true;
            }
        }


        /**
         *
         * @param {ClickEventCtx} e
         */
        function headerClickHandler(e) {
            var el = e.currentTarget;
            if(!e.defaultPrevented) {
                if (el.parentElement.tagName==='LI') {
                    if(el.parentElement.classList.contains('list-group-hdr')){
                        //completely ignore inline group headers - they don't cause a select
                        e.defaultPrevented = true;
                        e.stopPropagation = true;
                    }
                    //let the listClickHandler deal with this
                }else{
                    if (el.classList.contains('collapsible')){
                        SectionHeader.toggleHeader(el, e);
                        e.stopPropagation = true;
                        e.defaultPrevented = true;
                    }
                }
            }
            if (el.parentElement === document.body && e.target.tagName === 'BUTTON') {
                Views.dismiss('footer', Views.topMost);
            }
        }

        /**
         * @param {HTMLElement} el
         * @param {HTMLElement} view
         * @param {ClickEventCtx} e
         */
        function footerClickHandler(el, view, e) {
            if(el.classList.contains('page-footer')){
                el.classList.toggle('active');
                e.defaultPrevented = true;
                e.stopPropagation = true;
            }
        }
//endregion

//region DragDrop
        /**
         * NOTE: FireFox - During the dragstart, drag and dragend events, the mouse position (pageX, clientX, screenX) is set to 0 (zero).
         * "Won't Fix" - Workaround add dragover to doc and steal coordinates
         *
         * Generic DragDrop relates to DataBound List or Table elements (Files have there own special handler)
         * data-drag-mode is set on the drag container (the UL or TABLE or DATAGRID element)
         * the drag configuration is determined by the dragMode and is one of a set of predefined configs.
         * draggable="true" is set on the element that you allow to be dragged
         * See DragSrc for full details on the properties of the dragMode config object
         *
         * row-add-remove-self
         * drag a row and drop it in the same container
         * moves the row and inserts it before or after the drop target
         * row-add-hide-dest
         * drag a row and drop it on another container
         * adds the row to the new container before or after the drop target
         * hides or remove the old row
         *
         * Handler for the drag start event
         * @param {Event} e
         */
        function handleDragStart(e) {
            var vt = vtRoot;
            var el = e.target;
            if(el.nodeType===1) {
                //make sure that the target is an HTMLElement (could be just text) and is flagged as draggable (this event will always fire on the draggable target)
                if (el.getAttribute('draggable')) {
                    //make sure we are not trying to drag a focused input control or nested draggables
                    if (el.classList.contains('draggable-prevent') || inputTypes[document.activeElement.tagName]) {
                        e.preventDefault();
                        e.stopImmediatePropagation();
                    } else {
                        var cfg = CreateDragConfig();
                        cfg.vt = vtRoot;
                        cfg.isDragHandle = el.classList.contains('drag-handle');
                        if(cfg.isDragHandle){
                            // generally the draggable element is the draggable target but sometimes we want to start a drag on a handle
                            // and so we need to find the real parent target - you need to identify the target node if it is not the actual draggable element
                            cfg.dragHandle = el;
                            cfg.dragTarget = el.closest('.drag-target');
                            //explicitly marked or implicitly the config node
                            cfg.dragContainer = el.closest('.drag-container')||cfg.dragTarget.parentElement;//List or Table Container
                            cfg.dragMode = el.closest('[data-drag-mode]').dataset.dragMode;
                        }else{
                            // generally the draggable element is the draggable target but sometimes we want to start a drag on a handle
                            // and so we need to find the real parent target - you need to identify the target node if it is not the actual draggable element
                            cfg.dragHandle = el;
                            cfg.dragTarget = el;
                            //explicitly marked or implicitly the config node
                            cfg.dragContainer = el.parentElement.closest('.drag-container, [data-drag-mode]');//List or Table Container
                            cfg.dragMode = el.closest('[data-drag-mode]').dataset.dragMode;
                            if(cfg.dragContainer.dataset.dragConfig){
                                var config = $.stringToValueEx(cfg.dragContainer.dataset.dragConfig);
                                Object.assign(cfg, config);
                            }
                        }
                        cfg.dragImage = cfg.dragTarget;
                        cfg.offsetX = e.offsetX;
                        cfg.offsetY = e.offsetY;
                        cfg.dropAsChild = !!cfg.dragContainer.dataset.dropAsChild;
                        cfg.getLocation = getLocation;
                        // RJE: dropzones as well as other cfg changes are supposed to be added by the dev in the dragstart event
                        // this is fired through the actionTopicHandler below
                        // we only drag databound nodes
                        cfg.model = Views.findView(el).view.module.model;
                        cfg.dragSource = cfg.model.get(cfg.dragContainer.dataset.source, 'list', true);
                        cfg.dragKey = $.closestValue(el, 'key');
                        if (cfg.dragSource && cfg.dragKey != null) {//<< intentional != to cover null and undef
                            cfg.dragItem = cfg.dragSource.findByKey(cfg.dragKey);
                        }

                        cfg.dataTransfer = e.dataTransfer;
                        cfg.ctrlKey = Key.ctrlKey;
                        //resize/move/drop are now flagged for removal
                        switch (cfg.dragMode) {
                            case 'calendar':
                                //set topic to allow messages to be redirected to controller
                                cfg.topic = 'widget:';
                                //limit to drop on widget only
                                cfg.dropItemTypes = {WIDGET: 'TD'};
                                cfg.calDragTarget = e.target;
                                cfg.calDragLayerY = e.layerY;
                                break;
                            case 'drag-resize':
                                var rect = cfg.dragTargetRect = new Rectangle(cfg.dragTarget.getBoundingClientRect());
                                cfg.dragSourceRect = new Rectangle(rect);
                                cfg.dropTargetRect = new Rectangle(rect);
                                cfg.showDragImage = 'none';
                                cfg.dragImage = dragMarkerImage;//el.querySelector('.border');
                                cfg.dragImage.classList.add('drag-image', cfg.dragMode);
                                cfg.dragImage.style.top = rect.top+'px';
                                cfg.dragImage.style.left = rect.left+'px';
                                cfg.dragImage.style.width = rect.width+'px';
                                cfg.dragImage.style.height = rect.height+'px';
                                document.body.appendChild(cfg.dragImage);
                                cfg.showDropImage = true;
                                cfg.dropMarkerImage = dropMarkerImage;//el.querySelector('.border');
                                cfg.dropMarkerImage.classList.add('drop-marker');
                                document.body.appendChild(cfg.dropMarkerImage);
                                cfg.location='resize';
                                //get a default topic for event firing
                                cfg.topic = $.closestValue(el,'drag-action-topic');
                                break;

                            case 'drag-move-board':
                                //board use before-after for siblings but 'on'/'child' for a parent - child
                                //within a board only cards and bins are valid dragTargets and dropTargets
                                //a card is not a valid dropTarget for a bin
                                cfg.showDragImage = 'target';
                                cfg.showDropMarker = false;
                                cfg.location='before-after';
                                cfg.topic = $.closestValue(el,'drag-action-topic');
                                cfg.dragTarget.classList.add('draggable-ghost');
                                cfg.dragMove = true;
                                break;
                            case 'drag-move':
                                var rect = cfg.dragTargetRect = new Rectangle(cfg.dragTarget.getBoundingClientRect());
                                cfg.dropTargetRect = new Rectangle(rect);
                                cfg.showDragImage = 'target';
                                cfg.dragImage = dummyDragImage;//el.querySelector('.border');
                                cfg.dragImage.classList.add('drag-image', cfg.dragMode);
                                cfg.dragImage.style.top = rect.top+'px';
                                cfg.dragImage.style.left = rect.left+'px';
                                cfg.dragImage.style.width = rect.width+'px';
                                cfg.dragImage.style.height = rect.height+'px';
                                document.body.appendChild(cfg.dragImage);
                                cfg.showDropImage = true;
                                cfg.dropMarkerImage = dropMarkerImage;//el.querySelector('.border');
                                cfg.dropMarkerImage.classList.add('drop-marker');
                                document.body.appendChild(cfg.dropMarkerImage);
                                cfg.location='before-on-after';
                                cfg.topic = $.closestValue(el,'drag-action-topic');
                                break;
                            case 'row':
                            default:
                                cfg.showDragImage = 'none';
                                break;
                        }
                        globalToLocal(vtRoot, cfg.dragContainer);
                        //fire the event through the initial container..... the dev can reconfig the operation
                        var result = Views.fireActionEvent(cfg.dragContainer, e.type, false, true, cfg);

                        cfg.dataTransfer.effectAllowed = cfg.allowed;
                        if($.browser.browser==='msie'){
                            cfg.dataTransfer.setData('text', 'cfg');
                        }else{
                            cfg.dataTransfer.setData('application/x-dragsrc', cfg);
                        }

                        switch(cfg.showDragImage){
                            case 'none':
                                //update the drag image - dev can choose show it
                                if($.browser.browser!=='msie')cfg.dataTransfer.setDragImage(dummyDragImage, 0, 0);
                                break;
                            case 'target':
                                //update the drag image - dev can choose show it
                                //if drag by image then the mouse position needs adjusting for the image offset???
                                if($.browser.browser!=='msie')cfg.dataTransfer.setDragImage(cfg.dragTarget, cfg.offsetX, cfg.offsetY);
                                break;
                            case 'default':
                            case 'custom':
                            default:
                                //update the drag image - dev can choose show it
                                if($.browser.browser!=='msie')cfg.dataTransfer.setDragImage(cfg.dragImage, cfg.offsetX, cfg.offsetY);
                                break;
                        }

                        // check for dev dropzone config changes
                        // dev was supposed to pop the dz array directly
                        // but as a convenience we add simple string selector handling
                        // ie a dev supplied selector like an array of ids or all flagged dz's
                        // #mylist1,#mylist2
                        // [dropzone]
                        if (cfg.dropzones) {
                            var view = Views.findView(cfg.dragContainer);
                            var rootView = view.view.el;
                            cfg.dropzones = $.NodeListToArray(rootView.querySelectorAll(cfg.dropzones));
                        } else cfg.dropzones = [cfg.dragContainer];

                        // this/e.target is the source node.
                        if (result) {
                            e.target.classList.add('dragging');
                            vt.draggable = true;
                            vt.trackingDrag = true;
                            vt.dragConfig = cfg;
                            e.stopImmediatePropagation();
                            e.stopPropagation();
                        } else {
                            endTracking(vt);
                            e.stopImmediatePropagation();
                            e.stopPropagation();
                            e.preventDefault();
                        }
                    }
                }
            }
        }

        /**
         * Handler for the drag cancel event
         * really a drag without a drop also called from drop because of an issue where FF doesnt always call it
         * @param {Event} e
         */
        function handleDragEnd(e) {
            var vt = vtRoot;
            var el = e.target;
            var cfg = vt.dragConfig;
            if(cfg) {
                if(!cfg.finalized) {
                //fire the event through the initial container..... the dev can handle operation by returning false
                    //if there hasn't been a drop then this is a cancel event, excpet in drag-move there is no drop and no cancel
                    if(cfg.dragMode === 'drag-move-board'){
                        cfg.dropTarget = e.target;
                    }
                    var result = Views.fireActionEvent(cfg.dragContainer, cfg.dropped || cfg.moved ? 'dragend' : 'dragcancel', false, true, cfg);
                if(result) {
                    //cleanup here
                    if (cfg.dropContainer) {
                        cfg.dropContainer.classList.remove('dz-over');
                        }
                        if (cfg.dropTarget) {
                            cfg.dropTarget.classList.remove('dz-dropTarget', 'dz-after', 'dz-before', 'dz-highlight');
                        }
                        if (cfg.dragTarget) {
                            cfg.dragTarget.classList.remove('dragging','draggable-ghost');
                    }
                }
                    // Stops some browsers from redirecting on a drop (for dragging links or images).
                    //remove the vt markers
                    vt.draggable = false;
                    vt.trackingDrag = false;
                    vt.dragConfig = null;
                    dragMarkerImage.remove();
                    dropMarkerImage.remove();
                    e.stopPropagation();
                    e.preventDefault();
                    cfg.finalized = true;
                }
            }
        }

        /**
         * Handler for drag event (dragmove)
         * @param {Event} e
         */
        function handleDrag(e){
            var vt = vtRoot;
            var el = e.target;
            var cfg = vt.dragConfig;
            if (cfg) {
                //if dropTarget is a grid-row and we are in top half of first row scroll up...
                //if dropTarget is a grid-row and we are in bottom half of last row scroll down...
                if (cfg.dropTarget && cfg.dropTarget.tagName === 'GRID-ROW') {
                    var gs = cfg.dropTarget.closest('GRID-SECTION');
                    if (gs) {
                        //auto scroll
                        var pt = $.globalToLocal(gs, patchDragPosition ? dragMouseX : e.clientX, patchDragPosition ? dragMouseY : e.clientY);
                        var gsRect = gs.getBoundingClientRect();
                        var dg = gs.closest('DATAGRID');
                        if (pt.y > gsRect.height - 45) {
                            DataGrid.scrollRow(dg.ctx, 'down');
                        } else if (pt.y <= 0 + 45) {
                            DataGrid.scrollRow(dg.ctx, 'up');
                        }
                    }
                }
            }
        }
        /**
         * Handler for the drag enter event
         * For grids linked to grouped(Group) or hdd(Tree) data we really still use a flat markup structure.
         * However for lists we use a real hierarchical structure which complicates things a lot
         * so when we dragenter the dt.contains() will be misleading....
         * so..
         * is the new target a valid DT Tag for this DZ and is it different from the old DT
         * yes DT changed no DT didnt change
         * during a drag operation you MUST NOT rebuild the DOM, but you can manipulate it....
         * @param {Event} e
         */
        function handleDragEnter(e) {
            var vt = vtRoot;
            var el = e.target;
            var cfg = vt.dragConfig;
            if (cfg) {
                if(cfg.dragMode==='drag-resize'){

                }else {
                    globalToLocal(vtRoot, cfg.dragContainer);
                    if (cfg.dropContainer && cfg.dropContainer.contains(el)) {
                        // still within the same DZ so just find the new DT
                        // make sure we don't drop on ourselves and if in a grid we don't drop on the header or footer row
                        // this is done through the dropItemsType selector
                        // its valid to drop on a UL it just means append as a child of the parent item
                        // generally we are searching for valid 'drop' tags for the current drop container
                        var dt = el.closest(cfg.dropItemTypes[cfg.dropContainer.tagName] || '[dropzone]');
                        if (dt) {
                            //make sure we are not dropping on ourselves
                            if (!cfg.dragMove && cfg.dragTarget === dt) {
                                if(cfg.dropTarget){
                                    cfg.dropTarget.classList.remove('dz-dropTarget', 'dz-after', 'dz-before', 'dz-highlight');
                                    cfg.dropTarget = null;
                                }
                                cfg.dropLocation = null;
                                //cant drop here........
                            }else{
                                if (cfg.dragMode === 'calendar') {
                                    cfg.calDragOverPageX = e.pageX;
                                    cfg.calDragOverPageY = e.pageY;
                                }
                                if (dt === cfg.dropTarget) {
                                    //same DT no need to verify or reset
                                    handleDragOver(e);
                                }else{
                                //DT changed - clear old DT and find the new DT
                                    var dxx = cfg.dropTarget;
                                    if (cfg.dropTarget) {
                                        cfg.dropTarget.classList.remove('dz-dropTarget', 'dz-after', 'dz-before', 'dz-highlight');
                                        cfg.dropTarget = null;
                                    }
                                    cfg.dropLocation = null;
                                    //if its an explicitly blocked combo then ignore
                                    //from the potential dropTarget find an explict dropzone attr (can be set false) or assume true
                                    if (cfg.canDrop(el, dt, cfg)) {
                                        //check drop dir......
                                        cfg.dropDir = $.getFirstElementAttrValue(dt,'data-drop-dir', cfg.dropContainer,'vertical');
                                        cfg.dropTarget = dt;
                                        cfg.dropTarget.classList.add('dz-dropTarget');
                                        //now fake a dragOver
                                        handleDragOver(e);
                                    }else{
                                        //cant drop here........
                                    }
                                }
                                    }
                                } else {
                            //cant drop here........
                        }
                    } else {
                        //no longer in old DZ - clear the old DZ and the old DT
                        if (cfg.dropContainer) {
                            //clear the old DZ
                            cfg.dropContainer.classList.remove('dz-over');
                            cfg.dropContainer = null;
                            //clear the old DT
                            if (cfg.dropTarget) {
                                cfg.dropTarget.classList.remove('dz-after', 'dz-before', 'dz-highlight');
                                cfg.dropDir='';
                                cfg.dropTarget = null;
                                cfg.dropLocation = null;
                            }
                        }
                        //now find the find new DZ
                        // drop zones are high level structures - containers like DG/T/LV/TB/UL
                        // drop targets are elements within the dz
                        for (var i = 0, len = cfg.dropzones.length; i < len; i++) {
                            var dz = cfg.dropzones[i];
                            if (dz.contains(el)) {
                                //check whether this is an excluded dz
                                //new DZ
                                cfg.dropContainer = dz;
                                cfg.dropContainer.classList.add('dz-over');
                                if (cfg.dragMode === 'calendar') {
                                    cfg.calDragOverPageX = e.pageX;
                                    cfg.calDragOverPageY = e.pageY;
                                }
                                //fire the drag enter event
                                Views.fireActionEvent(cfg.dragContainer, e.type, false, true, cfg);
                                var dt = el.closest(cfg.dropItemTypes[cfg.dropContainer.tagName] || '[dropzone]');
                                if (dt) {
                                    //new DT - but is it really valid
                                    if (cfg.dragMove || cfg.dragTarget !== dt) {
                                        if (cfg.canDrop(el, dt, cfg)) {
                                            //check drop dir......
                                            cfg.dropDir = $.getFirstElementAttrValue(dt,'data-drop-dir', cfg.dropContainer,'vertical');
                                            cfg.dropTarget = dt;
                                            cfg.dropTarget.classList.add('dz-dropTarget');
                                            //fake a dragOver
                                            handleDragOver(e);
                                        }else{
                                            //cant drop here........
                                        }
                                    }
                                }else{
                                    //cant find a dt here
                                }
                                //break on first positive
                                break;

                            }
                        }
                    }
                }
            }
        }

        /**
         *
         * @param {HTMLElement} el
         * @param {HTMLElement} dt
         * @param {DragConfig} cfg
         * @return {boolean}
         */
        function checkCanDrop(el, dt, cfg) {
            return $.getFirstElementAttrValue(el, 'dropzone', cfg.dropContainer, true);
        }
        /**
         * returns the valid drop locations points
         * this function can be overridden by the dev during drag init
         * @param cfg
         * @returns {string}
         */
        function getLocation(cfg){
            //general list related element can only be dropped 'on'
            return (cfg.dragMode === 'highlight') || (cfg.dropTarget.tagName==='UL') || (cfg.dropTarget.tagName==='TASKBIN' && cfg.dragTarget.tagName!=='TASKBIN') ? 'on' : cfg.location;
        }
        /**
         * Handler for the drag over a dropzone event
         * Fires every 150-550ms
         * First we drag-enter a dropTarget, then we drag-over and then drag-leave or drop
         * so at this point we have a valid pair of drag/drop targets
         * now we want to determine the relative drop point
         * @param {Event} e
         */
        function handleDragOver(e) {
            var vt = vtRoot;
            var el = e.target;
            var cfg = vt.dragConfig;
            if (cfg) {
                //track dims change
                vt.p2.lastPageX = vt.p2.pageX;
                vt.p2.lastPageY = vt.p2.pageY;
                if (cfg.dragMode === 'drag-resize') {
                    //ignore
                } else {
                    //only fire if something changed
                    globalToLocal(vtRoot, cfg.dragContainer);

                    // beware the event can trigger on a child of the DT
                    //if what we are over is the target or is contained by the target...
                    if (cfg.dropTarget && (cfg.dropTarget === el || cfg.dropTarget.contains(el))) {
                        //process the drag action
                        var pt = $.globalToLocal(cfg.dropTarget, patchDragPosition ? dragMouseX : e.clientX, patchDragPosition ? dragMouseY : e.clientY);
                        var xrp,yrp;
                        if(cfg.dragMove){
                            //is the rect to the left or right of the dropTarget
                            var y = ((-cfg.offsetY) + pt.y);
                            var x = ((-cfg.offsetX) + pt.x);
                            if(x < 0 && (x + cfg.dragTarget.offsetWidth) < cfg.dropTarget.offsetWidth){
                                xrp =  (x + cfg.dragTarget.offsetWidth) / cfg.dropTarget.offsetWidth;
                            }else{
                                xrp =  x / cfg.dropTarget.offsetWidth;
                            }
                            if(y < 0 && (y + cfg.dragTarget.offsetHeight) < cfg.dropTarget.offsetHeight){
                                yrp =  (y + cfg.dragTarget.offsetHeight) / cfg.dropTarget.offsetHeight;
                            }else{
                                yrp =  y / cfg.dropTarget.offsetHeight;
                            }
                            doDragEffect(cfg, cfg.getLocation(cfg), xrp, yrp);
                        }else{
                            yrp = pt.y / cfg.dropTarget.offsetHeight;
                            xrp = pt.x / cfg.dropTarget.offsetWidth;
                            doDragEffect(cfg, cfg.getLocation(cfg), xrp, yrp);
                        }
                    } else {
                        cfg.dropLocation = '';
                    }
                }
                //regardless of whether the mouse moved we need to reset the dropEffect
                if(cfg.dropLocation){
                    e.preventDefault(); // Necessary. Allows us to drop.
                    e.dataTransfer.dropEffect = cfg.performed;  // See the section on the DataTransfer object.
                }
            }
        }

        /**
         *
         * @param {Object} cfg
         * @param {string} location
         * @param {number} xrp
         * @param {number} yrp
         */
        function doDragEffect(cfg, location, xrp, yrp){
            var rp = cfg.dropDir === 'horizontal' ? xrp : yrp;
                        switch (location) {
                            case 'before-after'://ie dropAsChild is false
                                if (rp > .5) {
                                    //over bottom half of drop target element
                                    cfg.dropLocation = 'after';
                                    cfg.dropTarget.classList.remove('dz-highlight', 'dz-before', 'dz-after');
                                    cfg.dropTarget.classList.add('dz-after');
                                } else {
                                    //over top half of drop target element
                                    cfg.dropLocation = 'before';
                                    cfg.dropTarget.classList.remove('dz-highlight', 'dz-before', 'dz-after');
                                    cfg.dropTarget.classList.add('dz-before');
                                }
                                break;
                            case 'on':
                                cfg.dropLocation = 'on';
                                cfg.dropTarget.classList.remove('dz-highlight', 'dz-before', 'dz-after');
                                cfg.dropTarget.classList.add('dz-highlight');
                                break;
                            case 'before-on'://ie dropAsChild is true
                                if (rp > .5) {
                                    //over bottom half of drop target element
                                    cfg.dropLocation = 'on';
                                    cfg.dropTarget.classList.remove('dz-highlight', 'dz-before', 'dz-after');
                                    cfg.dropTarget.classList.add('dz-highlight');
                                } else {
                                    //over top half of drop target element
                                    cfg.dropLocation = 'before';
                                    cfg.dropTarget.classList.remove('dz-highlight', 'dz-before', 'dz-after');
                                    cfg.dropTarget.classList.add('dz-before');
                                }
                                break;
                            case 'before-on-after':
                                if (rp > .75) {
                                    //over bottom half of drop target element
                                    cfg.dropLocation = 'after';
                                    cfg.dropTarget.classList.remove('dz-highlight', 'dz-before', 'dz-after');
                                    cfg.dropTarget.classList.add('dz-after');
                                    if (cfg.showDropMarker) positionDropMarker(cfg, cfg.dropMarkerImage, cfg.dropLocation);
                                } else if (rp < .25) {
                                    //over top half of drop target element
                                    cfg.dropLocation = 'before';
                                    cfg.dropTarget.classList.remove('dz-highlight', 'dz-before', 'dz-after');
                                    cfg.dropTarget.classList.add('dz-before');
                                    if (cfg.showDropMarker) positionDropMarker(cfg, cfg.dropMarkerImage, cfg.dropLocation);
                                } else {
                                    //over middle half of drop target element
                                    cfg.dropLocation = 'on';
                                    cfg.dropTarget.classList.remove('dz-highlight', 'dz-before', 'dz-after');
                                    cfg.dropTarget.classList.add('dz-highlight');
                                    if (cfg.showDropMarker) positionDropMarker(cfg, cfg.dropMarkerImage, cfg.dropLocation);
                                }
                                break;
                            default:
                    cfg.dropLocation = '';
                                break;
                        }
            if(cfg.dragMode==='drag-move-board') {
                //extra processing for board...
                var dragTarget = cfg.dragTarget;
                var dropTarget = cfg.dropTarget;
                //only move if 'after'
                var moved = false;
                if(dragTarget === dropTarget){

                    } else {
                    switch (cfg.dropLocation) {
                        case 'after': {
                            //dont do it if it is already there
                            if (dragTarget.previousElementSibling !== dropTarget) {
                                dropTarget.after(dragTarget);
                                moved = true;
                            }
                            break;
                        }
                        case 'before': {
                            //dont do it if it is already there
                            if (dragTarget.nextElementSibling !== dropTarget) {
                                dropTarget.before(dragTarget);
                                moved = true;
                            }
                            break;
                        }
                        case 'on': {
                            var childContainer = (dropTarget.dataset.dropLocation) ? dropTarget.querySelector(dropTarget.dataset.dropLocation) : dropTarget;
                            if (dragTarget.parentElement !== childContainer) {
                                if(cfg.dragMoveOnAppend){
                                    childContainer.appendChild(dragTarget);
                                }else{
                                    childContainer.prepend(dragTarget);
                        }
                                moved = true;
                            }
                            break;
                    }
                }
                    if (moved) {
                        cfg.moved = true;
                        //populate for event
                        cfg.dropSource = cfg.model.get(cfg.dropContainer.dataset.source, 'list', true);
                        cfg.dropKey = cfg.dropTarget.dataset.key;
                        cfg.dropItem = (cfg.dropSource && cfg.dropKey) ? cfg.dropSource.findByKey(cfg.dropKey) : null;
                        Views.fireActionEvent(cfg.dragContainer, 'dragmove', false, true, cfg);
                }
            }
        }
        }
        /**
         *
         * @param cfg
         */
        function adjustResizeRect(cfg){
            //push the target rect back to the source rect then increment to current size
            var rect = cfg.dragTargetRect.setRect(cfg.dragSourceRect);
            var xInc = cfg.vt.p2.pageX - cfg.vt.p1.pageX;
            var yInc = cfg.vt.p2.pageY - cfg.vt.p1.pageY;

            switch(cfg.dragHandle.dataset.handleType){
                case 'e':
                    rect.width += xInc;
                    break;
                case 's':
                    rect.height += yInc;
                    break;
                case 'se':
                    rect.width += xInc;
                    rect.height += yInc;
                    break;
                case 'w':
                    rect.x += xInc;
                    rect.width += -xInc;
                    break;
                case 'sw':
                    rect.x += xInc;
                    rect.width += -xInc;
                    rect.height += yInc;
                    break;
                case 'n':
                    rect.height += -yInc;
                    rect.x += xInc;
                    break;
                case 'ne':
                    rect.width += xInc;
                    rect.height += -yInc;
                    break;
                case 'nw':
                    rect.height += -yInc;
                    rect.x += xInc;
                    rect.width += -xInc;
                    break;
                default:
                    break;
            }
            cfg.dropTargetRect.setRect(cfg.dragTargetRect);
        }
        /**
         *
         * @param {DragConfig} cfg
         * @param {HTMLElement} dropMarker
         * @param {string} position
         */
        function positionDropMarker(cfg, dropMarker, position){
            switch(position){
                case 'left':
                case 'before':
                    var rect = cfg.dropTarget.getBoundingClientRect();
                    dropMarker.style.left = rect.left+'px';
                    dropMarker.style.top = (rect.top-10)+'px';
                    dropMarker.style.height = 10+'px';
                    dropMarker.style.width = rect.width+'px';
                    break;
                case 'on':
                    var rect = cfg.dropTarget.getBoundingClientRect();
                    dropMarker.style.left = rect.left+'px';
                    dropMarker.style.top = rect.top+'px';
                    dropMarker.style.height = rect.height+'px';
                    dropMarker.style.width = rect.width+'px';
                    break;
                case 'right':
                case 'after':
                    var rect = cfg.dropTarget.getBoundingClientRect();
                    dropMarker.style.left = rect.left+'px';
                    dropMarker.style.top = rect.bottom+'px';
                    dropMarker.style.height = 10+'px';
                    dropMarker.style.width = rect.width+'px';
                    break;
            }
        }
        /**
         * Handler for the drag leave event
         * With ToDos we now support have to support nested dropTarget
         * @param {Event} e
         */
        function handleDragLeave(e) {
            var vt = vtRoot;
            var el = e.target;
            var cfg = vt.dragConfig;
            if(cfg) {
                if(cfg.dragMode==='drag-resize'){

                }else {
                    globalToLocal(vtRoot, cfg.dragContainer);
                    if (cfg.dropTarget && cfg.dropTarget.contains(el)) {
                        //just moving around within the same DT
                    } else if (cfg.dropTarget === el) {
                        //leaving the DT
                        cfg.dropTarget.classList.remove('dz-dropTarget', 'dz-after', 'dz-before', 'dz-highlight');
                        cfg.dropTarget = null;
                    } else if (cfg.dropContainer === el && !el.contains(e.relatedTarget)) {
                        Views.fireActionEvent(cfg.dragContainer, e.type, false, true, cfg);
                        //leaving the DZ and not for some child....
                        if (cfg.dropTarget) {
                            cfg.dropTarget.classList.remove('dz-dropTarget', 'dz-after', 'dz-before', 'dz-highlight');
                            cfg.dropTarget = null;
                        }
                        cfg.dropContainer.classList.remove('dz-over');
                        cfg.dropContainer = null;
                    }
                }
            }
        }

        /**
         * Callback handler for the drop event
         * @param {Event} e
         */
        function handleDrop(e) {
            var vt = vtRoot;
            var el = e.target;
            var cfg = vt.dragConfig;
            if(cfg) {
                globalToLocal(vtRoot, cfg.dragContainer);
                if(cfg.dragMove){
                    //drag-move only uses drag-end
                    e.preventDefault();
                    e.stopImmediatePropagation();
                    e.stopPropagation();
                }else if(cfg.dragMode==='drag-resize'){
                    Views.fireActionEvent(cfg.dragContainer, e.type, false, true, cfg);
                    Views.fireActionEvent(cfg.dragContainer, 'dropcomplete', false, true, cfg);
                }else {
                    //if we dropped then the op is over regardless of what got dropped on
                    if (cfg.dropContainer) {
                        cfg.dropContainer.classList.remove('dz-over');
                        if (cfg.dropTarget) {
                            cfg.dropTarget.classList.remove('dz-after', 'dz-before', 'dz-highlight');

                            cfg.dropSource = cfg.model.get(cfg.dropContainer.dataset.source, 'list', true);
                            cfg.dropKey = cfg.dropTarget.dataset.key;
                            cfg.dropItem = (cfg.dropSource && cfg.dropKey) ? cfg.dropSource.findByKey(cfg.dropKey) : null;

                            cfg.dropPoint = $.globalToLocal(cfg.dropContainer, vt.p2.pageX, vt.p2.pageY);
                            //check with module - if they return false or prevent default we don't do anything
                            // If you don't want to update the DS then set the drag config operation 'performed' property to 'none' or return false on the drop event.
                            // The drag config is passed as the 'detail' property of the drop event.
                            var result = Views.fireActionEvent(cfg.dragContainer, e.type, false, true, cfg);
                            // this/e.target is the source node.
                            if (result) {
                                if (cfg.dragMode === 'calendar') {
                                    cfg.calDropTarget = e.target;
                                    cfg.calLayerY = e.layerY;
                                } else {
                                    var item;
                                    switch(cfg.performed){
                                        case 'move':
                                            item = cfg.dragItem;
                                            var dropLocation = cfg.dropLocation==='on' ? 'child' : cfg.dropLocation;
                                            //now there is a problem if the targets are LI but in different LISTVIEW or UL/LISTVIEW
                                            //this is way to simplistic really we should let the LV or DG handle this directly as there maybe style or attribute changes needed...
                                            //TODO:after the drop of the new list code remove this and let LV and DG handle directly....
                                            var node = cfg.dragTarget;
                                            node.classList.remove('dragging');
                                            if (cfg.modifyDOM) cfg.dropTarget.parentElement.insertBefore(node, cfg.dropLocation === 'before' ? cfg.dropTarget : cfg.dropTarget.nextElementSibling);
                                            if (cfg.dropSource !== cfg.dragSource) {
                                                cfg.dropSource.addItem(item, dropLocation, cfg.dropItem);
                                                //if we are moving the item to another DS then we need to manually remove from old DS
                                                cfg.dragSource.removeItem(cfg.dragItem);
                                            } else {
                                                //set old parent for use by validation routines after the move
                                                var parent = DataManager.getParent(cfg.dragSource, item);
                                                if (parent !== DataManager.getParent(cfg.dragSource, cfg.dropItem)) {
                                                    cfg.oldParent = parent;
                                                }

                                                //if we are moving the item to within the DS then DS will handle everything
                                                cfg.dragSource.addItem(item, dropLocation, cfg.dropItem);
                                            }
                                            break;
                                        case 'copy':
                                            var dropLocation = cfg.dropLocation==='on' ? 'child' : cfg.dropLocation;
                                            item = Object.assign({}, cfg.dragItem);
                                            var node = cfg.dragTarget.cloneNode(true);
                                            node.classList.remove('dragging');
                                            if (cfg.modifyDOM) cfg.dropTarget.parentElement.insertBefore(node, cfg.dropLocation === 'before' ? cfg.dropTarget : cfg.dropTarget.nextElementSibling);
                                            cfg.dragSource.addItem(item, dropLocation, cfg.dropItem);
                                            break;
                                        case 'none':
                                        default:
                                            break;
                                    }
                                }
                                Views.fireActionEvent(cfg.dragContainer, 'dropcomplete', false, true, cfg);
                            }
                        }
                    }
                }
                //flag as dropped
                cfg.dropped = true;
                //in FF the dragend doesnt fire if we remove the DOMElement so we force it here
                handleDragEnd(e);
            }
        }

//endregion

//region MouseWheel
        // Reasonable defaults
        var PIXEL_STEP  = 10;
        var LINE_HEIGHT = 40;
        var PAGE_HEIGHT = 800;

        var DOM_DELTA_PIXEL	= 0x00;//The delta values are specified in pixels.
        var DOM_DELTA_LINE	= 0x01;//The delta values are specified in lines.
        var DOM_DELTA_PAGE	= 0x02;//The delta values are specified in pages.
        /**
         * Normalize the wheel and scrollpad cross device and browser
         * @param {Event} event
         * @returns {{spinX: number, spinY: number, pixelX: number, pixelY: number}}
         * @memberof VMouseController#
         */
        function normalizeWheel(event){
            var sX = 0, sY = 0;       // spinX, spinY
            var pX = 0, pY = 0;       // pixelX, pixelY

            // Legacy
            if ('detail'      in event) { sY =  event.detail; }
            if ('wheelDelta'  in event) { sY = -event.wheelDelta  / 120; }
            if ('wheelDeltaY' in event) { sY = -event.wheelDeltaY / 120; }
            if ('wheelDeltaX' in event) { sX = -event.wheelDeltaX / 120; }

            // side scrolling on FF with DOMMouseScroll
            if ( 'axis' in event && event.axis === event.HORIZONTAL_AXIS ) {
                sX = sY;
                sY = 0;
            }

            pX = sX * PIXEL_STEP;
            pY = sY * PIXEL_STEP;

            if ('deltaY' in event) { pY = event.deltaY; }
            if ('deltaX' in event) { pX = event.deltaX; }

            if ((pX || pY) && event.deltaMode) {
                if (event.deltaMode == DOM_DELTA_LINE) {          // delta in LINE units
                    pX *= LINE_HEIGHT;
                    pY *= LINE_HEIGHT;
                } else {                                         // delta in PAGE units
                    pX *= PAGE_HEIGHT;
                    pY *= PAGE_HEIGHT;
                }
            }

            // Fall-back if spin cannot be determined
            if (pX && !sX) { sX = (pX < 1) ? -1 : 1; }
            if (pY && !sY) { sY = (pY < 1) ? -1 : 1; }
            //generally delta is in lines but it can be pixels on some platforms
            return { deltaX: sX, deltaY: sY, spinX: sX, spinY: sY, pixelX: pX, pixelY: pY };
        }
        //the last wheel event
        var wheelEvent;
        var wheelEventTimer;
        /**
         * Track scrolling when popup is open
         * If the UI has not been clicked in it will not have focus and will not trigger any kbd events
         * So therefore Key.ctrlKey will not registers any changes in its state,
         * however the UI will still register wheel events if the mouse simply hovers over the UI
         * we must therefore check e.ctrlKey as wel as the stored kbd state
         * passive listner only used to make sure any open popups get closed on a scroll
         * @param {Event} e
         */
        function mouseWheelHandler(e){
            //wheel scroll ...if it's from inside the page container we need to close any popup
            if (!wheelEvent || e.target !== wheelEvent.target) {
                //we optimise so we don't keep repeatedly checking the same object
                //but rem to kill this on any view close...
                wheelEvent = e;
                wheelEventTimer = clearTimeout(wheelEventTimer);
                wheelEventTimer = setTimeout(clearWheelTarget, 2000);
                if (window.pageContainer.contains(e.target)) {
                    //close any open popup....
                    Views.notify.fire('wheel', e);
                }
            }

        }

        /**
         * remove ref to old event otherwise target wont gc
         */
        function clearWheelTarget(){
            // remove any old wheel event as it contains a reference to an element
            wheelEvent=null;
        }

//endregion

        /**
         * Handle the window unload event
         * @param {Event} e
         */
        function unloadHandler(e) {
            //ok we are definitely going to shut down so lets tidy up - we only do this becuase DevTools has problems and it confuses people
            detachHandlers();
            self.handleEvent = $.noop;
        }

        /**@var Used to hold a 1x1 dummy image to hide the default drag image*/
        var dummyDragImage;
        var dragMarkerImage;
        var dropMarkerImage;

        /**
         *
         */
        function initDragDrop(){
            //dummy drag image - creates a transparent 1x1 gif
            dragMarkerImage = document.createElement('span');
            dropMarkerImage = document.createElement('span');
            dummyDragImage = document.createElement('img');
            dummyDragImage.src = 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7';
            if($.browser.browser === 'firefox'){
                patchDragPosition = true;
                document.addEventListener('dragover', gobalDragOverHandler);
            }
        }
        var patchDragPosition =false;
        var dragMouseX=0;
        var dragMouseY=0;
        /**
         *
         * @param e
         */
        function gobalDragOverHandler(e){
            dragMouseX = e.clientX;
            dragMouseY = e.clientY;
        }
        /**
         * init the controller
         * prior to being active the controller intentionally swallows all events
         */
        function init(){
            //detect if we are running on a mac
            isMac = ($.browser.os === 'mac');

            self.handleEvent = handleEvent;
            attachHandlers();
            initDragDrop();
        }
        init();
        /**@type VirtualTarget the VMouse target object */
        Object.defineProperty(self, 'vtarget', {get: function () {return vtRoot;}});
        self.globalToLocal = globalToLocal;
        self.gestureCapture = gestureCapture;
        self.normalizeWheel = normalizeWheel;
        self.setCapture = setCapture;
        self.getListenerOptions=getListenerOptions;
        return self;
    };
    /**@type VMouseController*/
    window.VMouse = new VMouseController();
//endregion


})(jQuery);